#ifndef __BCKeyMan__
#define __BCKeyMan__

#include <windows.h>
#include "..\CertificateLib\SMemory\SBlocks.h"
#include "..\CertificateLib\crtLibErrs.h"
#include "..\Includes\UnicodeFuncs.h"

extern HINSTANCE hMainInst;

typedef class cPKA * HPKA; // Handle to public/private key algorithm

#define DllExport __declspec( dllexport )

extern "C"
{

//
//	Function:
//		GenerateKeyPair
//	1. Generates key pair
//	2. Saves private key to file pfx_pdu format with .p12 extension
//		Adds the file to BCStorage.
//	3. Saves public key to certificate file with .cer extension.
//		The PK may be signed by its private key
//			Note, private key has to be added to BCStorage because the function takes
//			secret key attributes by ska local key ID
//		Or any other specified secret from BCStorage ( or may be from system storage WinNT )
//		Adds the file to BCStorage.
//	4. Fills localKeyID of the key been generated.
//		The generated key may be extracted from BCDB by GetPKBlockByID
//
//	If OK returns x509e_ok
//
DllExport int GenerateKeyPair(
		HINSTANCE hInst,
		HWND hWnd,
		BYTE *localKeyID = NULL,
		int *pLocalKeyIDBuffSize = NULL
	);

//
//	Key file format.
//
enum type_kff { kffNone = 0, kffX509 = 1, kffPGP = 2, kffPGPAscii = 3, kffPEM = 4 };

//
//	Functions required for extracting of public/secret key blocks from BCDB
//	GetPKBlockByID
//	BrowseForPKBlock
//	1. The functions does not decrypt secret key and does not create key handles,
//	it allows to avoid interceptions.
//	2. It uses cPlainTextBlock the class uses corret memory allocation,
//	so it's buffers may be properly allocated and deallocated in different modules
//	like Applicaton and dynamic loaded library
//
//	GetPKBlockByID
//	Tries to find key with specified ID in BCDB
//	Uses ( .cer ) files for public  or ( .p12 ) files for secret key
//	Choose public ( in local.pkr ) or secret ( in local.skr ) key from BCDB
//
//	If specified key is found, returns true and fills ptbKey buffer 
//
//	Notes: 
//
//	The packet is not parsed.
//	Usable function to extract corresponded key from the package
//
//		Public kffX509 - x509_certificate::parse	to get x509_publicKeyInfo as value tbsCertificate.subjectPKInfo 
//													and then pka_createAlgorithmX509 to create key handle
//		Secret kffX509 - P12_PFX_PDU_parse			to get x509_privateKeyInfo in ptbInfo.buffer 
//													and then ska_createAlgorithmX509 to create keu handle
//
//		Public kffPGP  - PKT_CreatePublicKeyAlgFromPacket to create public key handle
//
//		Secret kffPGP  - PKT_CreateSecretKeyAlgFromPacket to create secret key handle
//						or pgp_DecryptPublicKeyEncryptedSessionKeyFromPacket
//
DllExport bool GetPKBlockByIDT(
		HINSTANCE hInst,
		HWND hWnd,
		BYTE *keyID,
		int keyIDSize,
		bool secret,
		cPlainTextBlock *ptbKey,// [OUT] key in x509 or PGP format
		type_kff *pKeyPktFormat,// [OUT] Format of key packet
		//cPlainTextBlock *ptbUserID
		cWCStr *wcsUserID
	);

//
//	BCDB_EnumKeys
//		Walks BCDB and fills wcsText
//			the text contains list of key info
//			keyID, userName
//	Sample:
//  123456789, <qqq@qqqq.qqq>
//
//	Note, the key id may be used for extracting key from BCDB by GetPKBlockByID
//

//
//	05.08.2013
//	It is not right to reallocate cWCStr at dll and free at aplication.
//	Otherwise have to implement allocation with with heep.
//	Exclude overwritting all cWCStr implementation we juct change this function only
//
//DllExport bool BCDB_EnumKeysT( cWCStr *wcsText, bool secret );
//DllExport bool BCDB_EnumPGPKeysT( cWCStr *wcsText, bool secret );
//DllExport bool BCDB_EnumX509KeysT( cWCStr *wcsText, bool secret );

//
// All this functions return
//	- ppe_ok if ok,
//  - ppe_bufferTooSmall if buffer too small
//  - or other error.
//
DllExport int BCDB_EnumKeysBf( wchar_t *wc, size_t *pLen, bool secret );
DllExport int BCDB_EnumPGPKeysBf( wchar_t *wc, size_t *pLen, bool secret );
DllExport int BCDB_EnumX509KeysBf( wchar_t *wc, size_t *pLen, bool secret );

//DllExport bool BCDB_EnumKeys( cPlainTextBlock *ptbText, bool secret );
//DllExport bool BCDB_EnumPGPKeys( cPlainTextBlock *ptbText, bool secret );
//DllExport bool BCDB_EnumX509Keys( cPlainTextBlock *ptbText, bool secret );

//
//	Choose public key or subkey for encryption
//
//	1. Parses key packet with subkeys.
//	2. Retures ppe_ok if the key has no subkeys and it may be used for encryption
//	( RSA or ElGamal algorithm, never DSA ).
//	3. If subkeys present 
//		3.1 Uses cDlgBrowsePKUser without key ring values
//		3.2 Allows to choose directly the key or it's subkeys
//	4. It does not allows to choose a key what may be used for signature only.
//	Returns true if key was chosen otherwise returns false
//
DllExport bool ChooseKeyOrSubkeyT(
		HINSTANCE hInst,
		HWND hWnd,
		BYTE *keyPkt, int keyPktSize,	// Common key packet
		cPlainTextBlock *ptbKey,		// Choosen direct key or it's subkey
		cWCStr *wcsUserID
		//cPlainTextBlock *ptbUserID		// Key ID of choosen packet
	);

//
//	BrowseForPKBlock
//		Uses OpenFile to choose a public ( *.cer ) or secret ( *.p12 ) key file from BCDB
//		or from PGP key ring by means of BrowseKeyRingForUser ( *.pkr, *.skr, *.gpg )
//
//	See same Notes in GetPKBlockByID description.
//
DllExport bool BrowseForPKBlockT(
		HINSTANCE hInst,
		HWND hWnd,
		const wchar_t *initialPath, // if initialPath == NULL it runs in BCDB
		bool secret,
		cPlainTextBlock *ptbKey,	// [OUT] key in x509 or PGP format
		type_kff *pKeyPktFormat,	// [OUT] Format of key packet
		cWCStr *wcsUserID,			// [OUT] Pgp subkey does not have an user ID ( master key has only )
									// but we have to ask password for secret key with a name.
									// The BrowseKeyRingForUser generates here the subkey name as 
									// as "Subkey of ( master key user ID )" or  key user ID if the key has it
		cPTB *entirePacket = NULL
	);

//
//	BrowseForSecretKey
//	Uses OpenFile to choose the secret key file specified by keyID.
//	It may be as '.p12' secret key file as well as '.pgp, .skr, .asc' secret key ring file.
//
DllExport bool BrowseForSecretKey(
		BYTE *keyID,
		int keyIDSize,
		cPlainTextBlock *ptbKey,// [OUT] key in x509 or PGP format
		type_kff *pKeyPktFormat, // [OUT] Format of key packet
		cWCStr *wcsUserID // = NULL
	);

//
//	BrowseForPublicKey
//		Uses OpenFile to choose a public ( *.cer ) or secret ( *.p12 ) key file from BCDB
//		or from PGP key ring by means of BrowseKeyRingForUser ( *.pkr, *.skr, *.gpg )
//		password is prompted for secret key decryption
//
//DllExport bool BrowseForPublicKey(
//		HINSTANCE hInst,
//		HWND hWnd,
//		const char *initialPath, // if initialPath == NULL it runs in BCDB
//		bool secret,
//		HPKA *pHPKA
//	);

DllExport bool BrowseForPublicKeyT(
		HINSTANCE hInst,
		HWND hWnd,
		const wchar_t *initialPath,
		bool secret,
		HPKA *pHPKA
	);

//
//	BrowseKeyRingForUser
//		Visual interfase for choosing a public or secret key from PGP key ring ( *.pkr, *.skr, *.gpg )
//		It allows to dynamically change key file path, so krOut contains the new path
//		ptbUserID is 
//			UserID for PGP key packet,
//			Friendly name for P12 secret key packet,
//			Subject RDN for certificate
//
//		If both ptbKey and ptbKeyBlock are NULL the function is used for viewing and deleting a key
//		OK button in corresponded dialogue is changed as Delete and does not close the dialog.
//
//DllExport int BrowseKeyRingForUser(
//		HINSTANCE hInst,
//		HWND hWnd,
//		cPlainTextBlock *ptbKey, // [OUT] - pure key packet in sPublicKey or sSecretKey format as declared in Packets.h
//		const char *keyRingPath, // [IN ]
//		char *krOut = NULL,		 // [OUT] if krOut != NULL it is to be MAX_PATH size buffer.
//		bool secret = false,
//		cPlainTextBlock *ptbUserID = NULL,
//				// [OUT] Pgp subkey does not have an user ID ( master key has only )
//				// but we have to ask password for secret key with a name.
//				// The BrowseKeyRingForUser generates here the subkey name 
//				// as "Subkey of ( master key user ID )" or  key user ID if the key has it
//		cPlainTextBlock *ptbKeyBlock = NULL
//				// Entire key packet - master key with subpackets including subkeys packets
//	);

DllExport int BrowseKeyRingForUserT(
		HINSTANCE hInst,
		HWND hWnd,
		cPlainTextBlock *ptbKey, // [OUT] - pure key packet in sPublicKey or sSecretKey format as declared in Packets.h
		const wchar_t *keyRingPath, // [IN ]
		wchar_t *krOut = NULL,		 // [OUT] if krOut != NULL it is to be MAX_PATH size buffer.
		bool secret = false,
		cWCStr *wcsUserID = NULL,
		//cPlainTextBlock *ptbUserID = NULL,
				// [OUT] Pgp subkey does not have an user ID ( master key has only )
				// but we have to ask password for secret key with a name.
				// The BrowseKeyRingForUser generates here the subkey name 
				// as "Subkey of ( master key user ID )" or  key user ID if the key has it
		cPlainTextBlock *ptbKeyBlock = NULL
				// Entire key packet - master key with subpackets including subkeys packets
	);

//
//	SearchPKUser
//	  Searchs PGP uer on PGP lookup sites
//		1. Opens window with search options
//		2. Fills view window by search results.
//		3. Saves a selected key to local key ring.
//		4. Fills ptbKey block by a selected key.
//		
//	Returns ppe_ok if ptbKey is successfully filled, or ppe_keyNotFound. 
//	Otherwise returns a error code.
//
DllExport int SearchPKUserT(
		HINSTANCE hInst,
		HWND hWnd,
		cPlainTextBlock *ptbKey = NULL, // If NULL the dialog is used for saving the key to BCDB.
		BYTE *keyID = NULL, // If key id specified the key ID filed is filled and then locked.
		cWCStr *wcsUserID = NULL,
		//cPlainTextBlock *ptbUserID = NULL,// Chosen key block may not contain user ID,
										  // f.e. if it is a subkey. So it is useful to
										  // fill the value here.
		cPlainTextBlock *ptbLastSavedKey = NULL // If user saves key to BCDB it is copied to the ptb.
	);

//
//	SearchPKUserByID
//		Queries specified key on server. 
//		Fills ptbKey block if key found and returns ppe_ok.
//		Otherwise returns a error code.
//
int SearchPKUserByID(
		HINSTANCE hInst,
		HWND hWnd,
		cPlainTextBlock *ptbKey,
		BYTE *keyID
	);

//
//	SearchPKUserIDs
//		Queries list of users with specified info. 
//		Fills ptbUser block with query result.
//
int SearchPKUserIDs(
		HINSTANCE hInst,
		HWND hWnd,
		cPlainTextBlock *ptbUsers,
		int urlNum,
		char *userInfo,
		bool *pCancelled
	);

//
//	AddPublicKeyToStorage
//		Adds PGP public key to local key ring
//		Writes x509 certificate file to BCDB
//
DllExport bool AddPublicKeyToStorage( HWND hWnd, BYTE *pkt, int pktSize, type_kff kff );
DllExport bool AddKeyToStorage( HWND hWnd, BYTE *pkt, int pktSize, type_kff kff, bool secret = false );
DllExport bool AddKeyToStorageEx( HWND hWnd, BYTE *pkt, int pktSize, type_kff kff, bool secret = false, wchar_t *path = NULL );

//
//	BrowseKeyFiles
//		Browses key file in BCDB ( default if pathBuff is empty ) or an other directory
//		Uses detailed view of p12 or cer files.
//		Returns true if key file is selected and fills pathBuff by file path name.
//
DllExport bool BrowseKeyFiles(
		HWND hWnd,
		wchar_t *pathBuff, int buffSize,
		const wchar_t *filters,
		const wchar_t *boldCaptionText,
		const wchar_t *notes = L"",
		const wchar_t *OKBtnText = L"&OK",
		DWORD dwFlags = 0
	);

//
//	GetBCBDPath
//

enum { GetDataBasePath_Flag_Folder = 0, GetDataBasePath_Flag_PublicKeyRing = 1, GetDataBasePath_Flag_SecretKeyRing = 2 };

DllExport bool GetDataBasePath( wchar_t *buff, int buffLen, DWORD dwFalgs = GetDataBasePath_Flag_Folder );


//
//	Public recipient list functions
//

class cPublicRecipientList;

//
//	UpdatePublicRecipientList
//		Add all public key encounterred to new allocated cPublicRecipientList
//			1. Walks all certificate files.
//			2. Gets all public key from local key ring
//		Returns the list;
//
//	Note, all allocation and deallocations are performed in the dll.
//
DllExport cPublicRecipientList *UpdatePublicRecipientList( cPublicRecipientList *prl = NULL );
DllExport void FreePublicRecipientList( cPublicRecipientList *prl );

//
//	ViewLocalDatabase
//		1. Shows and manages keys in local data base.
//		2. If paramters ptbKey and pKFF are specified it allows to choose
//			key from local database.
//
//		View the decription for cDlgViewLocalDatabase::GetKeyOnly function
//

enum { vld_flag_viewSecret = 1, vld_flag_viewPublic = 2 };

DllExport bool ViewLocalDatabaseT(
		HWND hWnd,
		DWORD viewFlags = vld_flag_viewSecret | vld_flag_viewPublic,
		cPlainTextBlock *ptbKey = NULL,
		type_kff *pKFF = NULL,
		cWCStr *wcsUserID = NULL
		//cPlainTextBlock			// [OUT] Pgp subkey does not have an user ID ( master key has only )
		//	*ptbUserID = NULL	// but we have to ask password for secret key with a name.
								// The BrowseKeyRingForUser generates here the subkey name as 
								// as "Subkey of ( master key user ID )" or  key user ID if the key has it

	);

//
//	UpdateModuleTranslation
//		Use this function to reinitialise the module traslation if it has been changed, but the dll is not reloaded.
//		The function automatically starts when the dll loads.
//
DllExport bool UpdateModuleTranslation();

} // #extern "C"

//
// Predefined registry values for store information 
//	by readFromRegistry and writeToRegistry functions.
//
extern HKEY hKeyRoot;//  = HKEY_CURRENT_USER;
extern wchar_t *szBCKeyManKey; // = "Software\\Jetico\\BCKeyMan";

// Realised at PKSearch.cpp
extern bool GetPGPServicesList( HINSTANCE hInst );
extern bool FreePGPServicesList();

#endif // __BCKeyMan__