#include "PKParse.h"
#include "..\CertificateLib\x509pkts\base64funcs.h"
#include "..\CertificateLib\PGPPkts\Packets.h"
#include "..\CertificateLib\PGPPkts\PGPPkts.h"

//
//	ExtractDataAndVerifyPGPPublicKeyBlock
//		1. Parses key packet ( the packet has to be already decoded )
//		2. Verifies the key. It searches issuer key ring for the public key if the 
//			key is not self signed.
//		3. Generates specified values.
//
int ExtractDataAndVerifyPGPPublicKeyBlock(
		BYTE *pkt,	int pktSize,		// Public Key Packet with header.
		BYTE *issuerKR, int issuerKRSize,	// Key Ring where issuer can be found

		HPKA *pPka,	// Pointer to Key handle, if it is not NULL the key will be generated 
							// and has to be freed later.

		// It is set as pointer to position of user ID string in the pkt and it's size.
		BYTE **pUserID,	int *pUserIDSize,

		//
		// Both keys ID are 8 bytes size buffers, 
		//	if not NULLs the buffers are filled by subject/issuer generated key ID.
		//
		BYTE *keyID,		
		BYTE *issuerKeyID,

		bool extractOnly,

		BYTE *pAlgID
	)
{
	BYTE tag;
	int dataSize;
	int hdrSize;
	bool partial = 0;

	int res;
	int recCnt = 1;

	int parsed = 0;

	//
	//	Parses public key packet.
	//
	
	res = pgp_PacketHeader_Parse( &tag, &hdrSize, &dataSize, &partial, pkt, pktSize );
	if ( res != ppe_ok ) return res;

	if ( tag != pkt_PublicKey )
		return ppe_invalidBlockType;

	sPublicKey *pk = (sPublicKey *)(pkt + parsed + hdrSize);
	int pkSize = dataSize;

	BYTE algID = -1;

	int publicMpiSize;

	res = PKT_PublicKey_Parse( pk, pkSize, &publicMpiSize, &algID );
	if ( res != ppe_ok ) return res;

	if ( pAlgID )
		*pAlgID = algID;


	//
	//	Parses user ID
	//	

	while( true )
	{
		parsed += dataSize + hdrSize;
		res = pgp_PacketHeader_Parse( &tag, &hdrSize, &dataSize, &partial, pkt + parsed, pktSize - parsed );
		if ( res != ppe_ok ) return res;

		if ( tag == pkt_UserID )
			break;
		else
		if ( ( tag == pkt_Trust ) || 
			 ( tag == pkt_Signature ) ) // Version 1.05.2.2 - 16.08.2016
			continue;
		else
			return ppe_invalidBlockType;
	}

	BYTE *userID = (pkt + parsed + hdrSize);
	int  userIDSize = dataSize;

	if ( pUserID != NULL ) *pUserID = userID;
	if ( pUserIDSize != NULL ) *pUserIDSize = userIDSize;


	//
	//	Parses and verifies signature...
	//

	while( true )
	{
		parsed += dataSize + hdrSize;
		res = pgp_PacketHeader_Parse( &tag, &hdrSize, &dataSize, &partial, pkt + parsed, pktSize - parsed );
		if ( res != ppe_ok ) return res;

		if ( tag == pkt_Signature )
			break;
		else
		if ( tag == pkt_Trust )
			continue;
		else
			return ppe_invalidBlockType;
	}

	BYTE *sig = pkt + parsed + hdrSize;
	int sigSize = dataSize;

	if ( keyID )
		PKT_GetPublicKeyIDInternal( pk, pkSize, keyID );

	//
	//	Create key handle.
	//
	if ( pPka )
	{
		res = pka_createAlgorithm( pPka, algID, pk->mpi, publicMpiSize );
		if ( res != ppe_ok ) return res;
	}

	if ( extractOnly )
		return res;

	//
	//	Tries to verify signatute if the key is self signed.
	//
	res = pgp_VerifyKey(
		(BYTE *)pk, pkSize,
		userID, userIDSize,
		sig, sigSize,
		issuerKR, issuerKRSize,
		issuerKeyID );

	return res;
}

int DecodePGPKeyBlock( BYTE *msg, int msgSize, cPlainTextBlock *ptbKey, int *pParsed, const char **pHeaders, bool secret )
{
	//
	//	Finds armored body.
	//
	const char *szStart	= "-----BEGIN PGP PUBLIC KEY BLOCK-----";
	const char *szEnd	= "-----END PGP PUBLIC KEY BLOCK-----";

	if ( secret )
	{
		szStart = "-----BEGIN PGP PRIVATE KEY BLOCK-----";
		szEnd	= "-----END PGP PRIVATE KEY BLOCK-----";
	}

	const BYTE *keyBody;
	int keyBodySize;

	int parsed = 0;

	if ( pParsed )
		parsed = *pParsed;

	int res = findEncodedBlock( &keyBody, &keyBodySize, msg + parsed, msgSize - parsed, szStart, szEnd, pHeaders);

	if ( res != ppe_ok )
		return res;

	parsed = (int)( keyBody - msg ) + keyBodySize + strlen( szEnd );

	if ( pParsed )
		*pParsed = parsed;

	//
	//	Decodes the buffer.
	//

	int size = 0;
	res = base64Decode( NULL, &size, keyBody, keyBodySize );

	if ( res != ppe_bufferTooSmall )
		return res;

	if ( ptbKey->Allocate( size ) == NULL )
		return ppe_memoryAllocation;

	res = base64Decode( ptbKey->buffer, &size, keyBody, keyBodySize );

	return res;

}

int DecodePGPPublicKeyBlock( BYTE *msg, int msgSize, cPlainTextBlock *ptbKey, int *pParsed, const char **pHeaders )
{
	return DecodePGPKeyBlock( msg, msgSize, ptbKey, pParsed, pHeaders, false );
}

