#include "BCKeyMan.h"
#include "BCDB_Functions.h"
#include "PKParse.h"

#include "..\BCArchive_V2\Frame\cPublicRecipient.h"

#include "..\CertificateLib\PGPPkts\PGPPkts.h"
#include "..\CertificateLib\PGPPkts\Packets.h"

#include "..\Includes\FileView.h"
#include "..\Includes\UnicodeFuncs.h"

static bool WalkCertificateAndAddToList( wchar_t *path, cPublicRecipientList *prl )
{
	int keyIDSize;
	BYTE *keyID;
	wchar_t eMail[ 1024 ];
	wchar_t commonName[ 1024 ];

	int len = wcslen( path );
	if ( path[ len - 1 ] != L'\\')
	{
		path[ len++ ] = L'\\';
		path[ len ] = L'\0';
	}

	wcsncpy( path + len, L"*.cer", MAX_PATH - len );

	WIN32_FIND_DATAW  findData;
	memset( &findData, 0, sizeof( findData ) );

	HANDLE hFind = FindFirstFileT( path, &findData );
	if ( hFind == INVALID_HANDLE_VALUE )
		return false;

	do
	{
		wcsncpy( path + len, findData.cFileName, MAX_PATH - len );

		if ( ( findData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ) != 0 )
			continue;

		cFileView cfv( path );
		cPlainTextBlock ptb;

		x509_certificate crt;
		if ( crt.parse( cfv.m_ptr, (int)cfv.m_size.QuadPart ) != x509e_ok )
			continue;

		int res = crt.tbsCertificate.getSubjectKeyID( &keyID, &keyIDSize );
		if ( res == x509e_extensionIsNotPresent )
		{
			keyIDSize = 0;
			res = crt.tbsCertificate.generateSubjectKeyID( NULL, &keyIDSize );
			if ( res != pkae_bufferSizeTooSmall )
				continue;

			if ( ptb.Allocate( keyIDSize ) == NULL )
				continue;

			res = crt.tbsCertificate.generateSubjectKeyID( ptb.buffer, &keyIDSize );
			if ( res != x509e_ok )
				continue;

			keyID = ptb.buffer;
		}

		const char *szCNValue = crt.tbsCertificate.subjectName.getAttributeValue( sNameAttribute::getOid( sNameAttribute::id_at_commonName ) );
		const char *szEMailValue = crt.tbsCertificate.subjectName.getAttributeValue( sNameAttribute::getOid( sNameAttribute::pkcs9_emailAddress ) );

		if ( szCNValue != NULL )
			mbstowcs( commonName, szCNValue, maxLen( commonName ) );
		else
			commonName[0] = L'\0';

		if ( szEMailValue != NULL )
			mbstowcs( eMail, szEMailValue, maxLen( eMail ) );
		else
			eMail[0]= L'\0';

		if ( ( eMail[0] != L'\0' ) && ( commonName[0] != L'\0' ) )
			prl->Add( commonName, eMail, keyID, keyIDSize );
	}
	while( FindNextFileT( hFind, &findData ) != 0 );

	FindClose( hFind );

	return true;
}

static bool checkEncryptionAllowed( sPublicKey *keyPkt, int keyPktSize )
{
	BYTE algID = keyPkt->getAlgID();

	switch ( algID )
	{
		case pgppkaRSA		:
		case pgppkaRSA_e	:
		case pgppkaRSA_s	:
		case pgppkaElgamal_e:
		case pgppkaElgamal	:
			return true;
		case pgppkaDSA		:
			return false;
		default : return false;
	}
}

//
//	checkEncryptionAllowed
//		walks key and all subkeys,
//		returns first encryption ready subkey.
//
static cpgp_keyInfo *checkEncryptionAllowed( cpgp_keyInfo *cKeyList )
{
	for( ; cKeyList; cKeyList = cKeyList->next )
	{
		if ( checkEncryptionAllowed( ( sPublicKey * )cKeyList->keyPkt, cKeyList->keyPktSize ) )
			return cKeyList;
	}

	return NULL;
}

//
//	WalkKeyRingAndAddToList
//		1. Enumirates all master key and subkeys in specified key ring
//		2. If master key is encryption key or has encryption subkey the key is added to list
//			2.1 It gets userID from master key only
//			2.2 It gets keyID form master if it is encryption key or from first subkey.
//
static bool WalkKeyRingAndAddToList( BYTE *keyPkt, int keyPktSize, cPublicRecipientList *prl )
{
	cPlainTextBlock keyIDTbl;

	int size = 0;

	int res = pgp_EnumKeyIDs( NULL, &size, keyPkt, keyPktSize, false );

	if ( res == ppe_bufferTooSmall )
	{
		if ( keyIDTbl.Allocate( size ) == NULL )
			return false;

		res = pgp_EnumKeyIDs( keyIDTbl.buffer, &size, keyPkt, keyPktSize, false );
	}

	if ( res != ppe_ok )
	{
		assert( false );
		return false;
	}

	int keyQ = size / pgp_keyIDSize;
	

	for ( int i = 0; i < keyQ; i++ )
	{
		cpgp_keyInfo cKey, *pSubkey = NULL;
		
		BYTE *keyIDPos = keyIDTbl.buffer + i * pgp_keyIDSize;

		if ( ( pgp_FindKeyByIDEx( &cKey, keyIDPos, keyPkt, keyPktSize, false, &pSubkey ) == ppe_ok )
			&& ( pSubkey == NULL ) ) // Skips subkeys ( they have no userID ), uses master key only
		{
			//
			//	Adds master key or first key with encryption algorithm
			//	But it takes user ID from master key only
			//
			cpgp_keyInfo *pEncKey = checkEncryptionAllowed( &cKey );
			if ( pEncKey )
			{
				//wchar_t wUserID[ 1024 ];
				//int len = cKey.userID->size;
				//if ( len >= maxLen( wUserID ) )
				//	len = maxLen( wUserID ) - 1;

				//mbstowcs( wUserID, (char *)cKey.userID->body, len );
				//wUserID[ len ] = L'\0';

				//prl->Add( NULL, wUserID, pEncKey->keyID, pgp_keyIDSize );

				cWCStr wcs;
				wcs.Set( (char *)cKey.userID->body, cKey.userID->size, CP_UTF8 );

				prl->Add( NULL, wcs.m_szBody, pEncKey->keyID, pgp_keyIDSize );
			}
		}
	}

	return true;
}
//
//	UpdatePublicRecipientList
//		Add all public key encounterred to new allocated cPublicRecipientList
//			1. Walks all certificate files.
//			2. Gets all public key from local key ring
//		Returns the list;
//
cPublicRecipientList *UpdatePublicRecipientList( cPublicRecipientList *prl /*= NULL*/ )
{
	if ( prl != NULL )
		prl->Empty();
	else
	{
		prl = new cPublicRecipientList;
		if ( prl == NULL )
			return NULL;
	}

	wchar_t path[ MAX_PATH ];
	BCDB_GetDataBasePath( path, maxLen( path ) );
	WalkCertificateAndAddToList( path, prl );

	BCDB_GetLocalKeyRingPath( path, maxLen( path ), false );
	cFileView fv( path );
	if ( !fv.m_err )
		WalkKeyRingAndAddToList( fv.m_ptr, (int)fv.m_size.QuadPart, prl );

	return prl;
}

//
//	GetPublicRecipientList
//		Returns new updated list.
//
cPublicRecipientList *GetPublicRecipientList()
{
	return UpdatePublicRecipientList( NULL );
}

void FreePublicRecipientList( cPublicRecipientList *prl )
{
	if ( prl != NULL )
		delete prl;
}

