#include <stdlib.h>
#include <memory.h>
#include <string.h>
#include "smemory.h"
#include <assert.h>
#include <stdio.h>
#include <errno.h>

#ifndef _WIN32
int strcat_s( char *d, size_t size, const char *s )
{
    if (d == NULL || s == NULL)
        return EINVAL;
    
    size_t len = strlen( s );
    if (len >= size)
        return ERANGE;
    
    size_t pos = strlen( d );
	if ( pos + len >= size )
		return ERANGE;

	strncpy( d + pos, s, size - pos );
	return 0;
}
#endif


#if defined(_windows_kernel_) || defined(__APPLE__)
#	define sMemory_Free( p ) delete p
#	define sMemory_Alloc( __size ) new char[__size]
#else
#	define sMemory_Free( p ) free( p )
#	define sMemory_Alloc( __size ) malloc( __size )
#endif

/*****************************************************************************\
* SMalloc::
* 1) Allocate memory with additional place for required size storing.
* 2) Save the size at start of allocated block.
* 3) Return pointer to normal continuos memory required size.
*
* Note: Use only SFree function to deallocate blocks allocated 
*       by SMalloc or SCalloc .
*
\*****************************************************************************/

//
//	It does not fill allocated memory by zero if signature is not compatible.
//
//#define _useSig 0xBCA
//#define sigTailSize 256 // 40

//
// The signed memory block looks like
//  | Head Signature | real size including both head and tail signatures | ... user memory body ... | Tail Sigature |
//
void *SZalloc( size_t size ) //, bool fillByZero )
{
	void *ptr = SMalloc( size );

	if ( ptr != NULL )
		memset( ptr, 0, size );

	return ptr;
}

void *SMalloc( size_t size )
{
	// reserves space for size at new block, SReAlloc uses this field.
	size_t realSize = size + sizeof(size_t); 

#ifdef _useSig
	realSize += ( 2 * sigTailSize ) * sizeof(size_t);
#endif
	size_t *ptr = (size_t *) sMemory_Alloc( realSize );

	if (ptr == 0)	return NULL;

#ifdef _useSig
	int i = 0;

	// Head signature
	for ( i = sigTailSize; i--; )
    	    *(ptr++) = _useSig; 

#endif

	*(ptr++) = realSize;
	
#ifdef _useSig

	// Tail signature
	size_t *tailSigPtr = (size_t*)(( unsigned char * )ptr + size);
	for ( i = sigTailSize; i--; )
	    tailSigPtr[i] = _useSig;
	    
#endif
	SMemCheck( ptr, 0, 0 );

	return (void *)(ptr);
}

void *SCalloc(size_t num, size_t size)
{
	return SMalloc(num * size);
}

int SMemCheck( void *memBlock, size_t **ppPtr, size_t *pSize )
{
	size_t *ptr = (size_t *)memBlock;
	size_t realSize = *( --ptr );
	
#ifdef _useSig

	size_t blockSize = realSize - (1 + 2 * sigTailSize) * sizeof( size_t );
	size_t *tailSigPtr = (size_t*)( (unsigned char*)memBlock + blockSize );
	
	int i = 0;

	for ( i = sigTailSize; i--; )
	{
	    if ( *( --ptr ) != _useSig ) 
	    {
//fprintf( stderr, "%s Head: memBlock %X, size %d, ptr %X, realSize %d, i %d, *ptr %X\n", __FUNCTION__, memBlock, blockSize, ptr, realSize, i, *ptr );
		assert( false );
	    }
	}
	
        for ( i = sigTailSize; i--; )
    	{
    	    if ( tailSigPtr[i] != _useSig )
	    {
//fprintf( stderr, "%s Tail: memBlock %X, size %d, ptr %X, realSize %d, tailSigPtr %X[%d]%X\n", __FUNCTION__, memBlock, blockSize, ptr, realSize, tailSigPtr, i, tailSigPtr[i] );
//fprintf( stderr, "tailSig[-1] %X, [0] %X, [%d] %X, [%d]%X\n", tailSigPtr[-1], tailSigPtr[0], i - 1, tailSigPtr[i-1], i, tailSigPtr[i] );
	        assert( false );
	    }
	}
#endif

	if ( ppPtr != NULL )
	    *ppPtr = ptr;
	    
	if ( pSize != NULL )
	    *pSize = realSize;
	    
	return 0;
}

/*****************************************************************************\
* SFree::
* 1) Retrieve stored size of early allocated by SMalloc or SCalloc block.
* 2) Fill the block by zero.
* 3) Deallocate memory block.
*
* Note: Use SFree function to deallocate blocks allocated only
*       by SMalloc or SCalloc.
*
\*****************************************************************************/

void SFree( void *memBlock )
{
	size_t *ptr, realSize;
	
	if ( SMemCheck( memBlock, &ptr, &realSize ) != 0  )
	{
	    assert( false );
	}
	else
	    memset( ptr, 0, realSize );

	sMemory_Free(ptr);
}

/*****************************************************************************\
* SReAlloc::
* 1) Free previosly allocated block by SFree function.
* 2) Return new block allocated by SMalloc.
*
* Note: Use blocks allocated by SMalloc or SCalloc or SReAlloc only.
*       Use SFree function only to deallocate reallocated block.
*
\*****************************************************************************/

#define MemBlockSize(ptr) (*( (size_t *)ptr - 1 ))

void *SReAlloc(void *memblock, size_t size)
{
	size_t shortSize = 0;
	void *newPtr = SMalloc(size);

	if ( newPtr == NULL )
		return NULL;

	if ( memblock != NULL )
		shortSize = MemBlockSize( memblock );

	if ( shortSize > size )
		shortSize = size;

	if ( shortSize > 0 )
		SMemCpy( newPtr, memblock, shortSize );

	if ( memblock != NULL )
		SFree(memblock);

	return newPtr;
}

int SMemCpyS( void *dest, size_t dstSize, const void *src, size_t count ) 
{
	if ( dstSize < count )
		return ERANGE;

	if ( ( dest == NULL ) || ( src == NULL ) )
		return EINVAL;

	memcpy( dest, src, count );
	return 0;
}

void *SMemCpy( void *dest, const void *src, size_t count )
{
	return memcpy(dest,src,count);
}

void *SMemSet(void *dest, int c, size_t count)
{
	return memset(dest,c,count);
}

void *SMemMove(void *dest, const void *src, size_t count)
{
	memmove(dest,src,count);
	return dest;
}

size_t SMemCmp( void *buf1 , const void *buf2 , size_t count )
{
	return memcmp( buf1 , buf2 , count );
}


/******************************************************************************************\
* This functions required to replace standard memory allocation function from GNU MP 2.0
\******************************************************************************************/


/******************************************************************************************\
* Function: void * S_allocate_function (size_t alloc_size) 
* Function: void * S_reallocate_function (void *ptr, size_t old_size, size_t new_size) 
* Function: void S_deallocate_function (void *ptr, size_t size)
*
* To replace these functions use the GMP function before first initialisation
*  mp_set_memory_functions ( S_allocate_function, S_reallocate_function, S_deallocate_function );
*
\******************************************************************************************/


void * S_allocate_function( size_t alloc_size ) 
{
	void *ptr;

	if ( SMemoryError )
		return NULL;

	ptr = SMalloc( alloc_size );

	if ( ptr == NULL )
		SMemoryError = 1;

	return ptr;
}

void * S_reallocate_function(void *memblock, size_t old_size, size_t new_size) 
{
	(void)old_size; // Unreferenced formal parameter.

	assert( MemBlockSize( memblock ) == old_size );

	return SReAlloc( memblock , new_size );
}

void S_deallocate_function(void *memblock, size_t size) 
{
	(void)size; // Unreferenced formal parameter.

	assert( MemBlockSize( memblock ) == size );

	SFree( memblock );
}

int SMemoryRedirected = 0; /* Used for detection was redirection or not */
int SMemoryError = 0;	   /* Not zero if was memory allocation error 
						    * if redirection is applied you need zeroed the flag 
							* to normal allocate memory again.
							*/
