/////////////////////////////////////////////////////////////////
//
//	Copyright (c) 2017 Jetico Inc., Finland
//	All rights reserved.
//
//	File: x509_sequence.h
//	Revision: 
//	Created:
//	Description:
//		Fills and parses sequence with elements.
//	Sample filling:
//		x509_sequence seq;
//		BYTE id[] = { 1,2,3,4,5,6,7,8 };
//		if ( !seq.addElement( asn1tag_OCTET_STRING, id, sizeof( id ) ) )
//			return seq.err;
//		int size = seq.prepareForWritting();
//
//		cPTB ptb;
//		BYTE *bf = ptb.AddBytes( NULL, size );
//		if ( seq.writeIt( ptb.buffer, ptb.bufferSize ) != x509e_ok )
//			return seq.err;
//
//	Sample reading:
//
//		x509_sequence seqRead;
//	
//		int res = seqRead.parseIt( ptb.buffer, ptb.bufferSize );
//		if ( res != x509e_ok )
//			return res;
//
//		x509_struct element;
//
//		cPTB key;
//		while( seqRead.getNextElement( element ) )
//		{
//			// Todo
//			.. element.B, element.L, ..
//		}
//
/////////////////////////////////////////////////////////////////

#pragma once

#include "x509_structs.h"
#include "../SMemory/SBlocks.h"

struct x509_sequence 
	: public x509_struct
{
	int m_err;
	BYTE *m_pkt;	// next position value for updating
	int m_pktSize;	// size left value for updating

	x509_sequence( const BYTE *pkt = NULL, int pktSize  = 0) 
	{
		parseIt( pkt, pktSize );
	}

	int parseIt( const BYTE *pkt, int pktSize )
	{
		clear();

		if ( pkt == NULL )
			return m_err = x509e_invalidPktSize;

		m_err = x509_struct::parse( pkt, pktSize, NULL, NULL, asn1tag_CONSTRUCTED_SEQUENCE );
		if ( m_err != x509e_ok )
			return m_err;

		m_pkt = B;
		m_pktSize = L;

		return m_err;
	}

	bool getFirstElement( x509_struct &element )
	{
		m_pkt = B;
		m_pktSize = L;
		return getNextElement( element );
	}

	bool getNextElement( x509_struct &element )
	{
		if ( m_pktSize == 0 )
			return false;

		m_err = element.parse( m_pkt, m_pktSize, &m_pkt, &m_pktSize );
		return m_err == x509e_ok;
	}

	cPTB m_ptbBody;

	bool addElement( BYTE tag, BYTE *body, unsigned int bodySize )
	{
		x509_struct element;

		int elementSize = element.prepareForWritting( tag, bodySize );

		if ( elementSize == 0 )
		{
			m_err = x509e_invalidPktSize;
			return false;
		}

		BYTE *ptr = m_ptbBody.AddBytes( NULL, elementSize );
		if ( ptr == NULL )
		{
			m_err = x509e_memoryAllocation;
			return false;
		}

		if ( element.writeHeader( ptr, elementSize ) == 0 )
		{
			m_err = x509e_internalError;
			return false;
		}

		m_err = element.writeBody( body, bodySize );

		return m_err == x509e_ok;
	}

	int prepareForWritting()
	{
		return x509_struct::prepareForWritting( asn1tag_CONSTRUCTED_SEQUENCE, m_ptbBody.bufferSize );
	}

	int writeIt( 
		BYTE *buff,		// Buffer for header writting
		int buffSize	// Size of allocated buffer
		)
	{
		int res = writeHeader( buff, buffSize );

		if ( res == 0 )
		{
			m_err = x509e_internalError;
			return false;
		}

		m_err = writeBody( m_ptbBody.buffer, m_ptbBody.bufferSize );

		return m_err;
	}
};
