/*
 *	Copyright 2014 Jetico, Inc. All rights reserved.
 *	cAutoThread.h
 *	Description:
 *		1. Creates thread by 'createThread' function, the thread just calls virtual 'ThreadActor' function.
 *		2. Stops thread.
 *			There are two thread types:
 *				a. Walking, like directory walking. This thread has to check m_pCancelled parameter.
 *					If *m_pCancelled becomse true this thread has to stop walking.
 *				b. Wating a notification, like wait file change notification. This thread has to use multiply objects waiting function,
 *					where first object is m_hCloseEvent member. So SetEvent will stop this wating.
 *		3. It automatically stops and close thread at this object destructor.
 *
 *	Created:	22.04.2013
 *	Author:		Alexander Pichuev.
 */

#pragma once

#include <windows.h>
#include "ODS.h"

#ifdef _ODS
#include "UnicodeFuncs.h"
#endif

//
//	Class cAutoThread
//	Usage:
//		cAutoThread at;
//
//		// 1. Creates separate thread, 
//		at.createThread()
//
//		// 2. Stops thread
//		at.stopAndCloseThread()
//
//		// 3. The abstract 'ThreadActor' function has to be released at derived class
//		virtual DWORD ThreadActor()
//		{
//			// It has to use one of or both stop factors:
//			//	m_hCloseEvent and/or m_pCancelled
//		}
//
class cAutoThread
{
//protected:
public:
	HANDLE m_hThread;

	DWORD m_waitAtClosingTimeout;

	bool m_bCancelled;

	HANDLE m_hCloseEvent;
	bool m_bUseCloseEvent;

	static DWORD WINAPI createThread_static( LPVOID lpThreadParameter )
	{
		cAutoThread *pThis = ( cAutoThread * )lpThreadParameter;

		if ( pThis == NULL )
		{
			ODS_fls( cWCStr().FormatT( L"Assertion! Error in casting object class at thread 0x%X", GetThreadId( GetCurrentThread() ) ) );
			return -1;
		}

		DWORD dwRes = pThis->ThreadActor();

		return dwRes;
	}

	DWORD waitThread()
	{
#ifdef _ODS
		if ( m_waitAtClosingTimeout == INFINITE )
		{
			for ( DWORD waitTimeout = 1000; ; waitTimeout *= 2 )
			{
				DWORD dwRes = WaitForSingleObject( m_hThread, waitTimeout );

				if ( dwRes != WAIT_TIMEOUT )
					return dwRes;

				ODS_fls( cWCStr().FormatT( L"Warning! Timeout in waiting thread 0x%04X", GetThreadId( m_hThread ) ) );
			}
		}
		else
#endif
			return WaitForSingleObject( m_hThread, m_waitAtClosingTimeout );
	}

	bool createCloseEvent()
	{
		if ( !m_bUseCloseEvent )
			return true;

		m_hCloseEvent = CreateEvent( NULL, false, false, NULL );

		if ( m_hCloseEvent != NULL )
			return true;

		ODS_fls( L"Assertion!" );
		return false;
	}

	//
	//	If ThreadActor waits m_hCloseEvent so it stops working.
	//
	bool setAndCloseEvent()
	{
		if ( m_hCloseEvent == NULL )
			return false;

		SetEvent( m_hCloseEvent );
		CloseHandle( m_hCloseEvent );
		m_hCloseEvent = NULL;

		return true;
	}
public:

	bool *m_pCancelled;

	virtual DWORD ThreadActor() = 0;

	cAutoThread( DWORD timeout = INFINITE, bool *pCancelled = NULL, bool bUseCloseEvent = false )
	{
		m_hThread = NULL;
		m_bCancelled = false;
		m_hCloseEvent = NULL;

		cAutoThread_Init( timeout, pCancelled, bUseCloseEvent );
	}

	bool cAutoThread_Init( DWORD timeout, bool *pCancelled = NULL, bool bUseCloseEvent = false )
	{
		m_waitAtClosingTimeout = timeout;
		m_bUseCloseEvent = bUseCloseEvent;

		if ( pCancelled != NULL )
			m_pCancelled = pCancelled;
		else
			m_pCancelled = &m_bCancelled;
	
		return true;
	}

	bool createThread()
	{
		if ( m_hThread != NULL )
		{
			ODS_fls( cWCStr().FormatT( L"Assertion! Thread 0x%X is active", GetThreadId( m_hThread ) ) );
			return false;
		}

		if ( !createCloseEvent() )
			return false;

		m_hThread = CreateThread( NULL, 0, createThread_static, (void*)(cAutoThread*)this, 0, NULL );

		if ( m_hThread != NULL )
			return true;

		ODS_fls( cWCStr().FormatT( L"Assertion! Error %d in creating thread", GetLastError() ) );

		return false;
	}

	~cAutoThread()
	{
		stopAndCloseThread();
	}

	bool isActiveThread()
	{
		if ( m_hThread == NULL )
			return false;

		DWORD dwRes = 0;

		if ( !GetExitCodeThread( m_hThread, &dwRes ) )
		{	ODS_fls( L"Assertion!" ); return false; }

		return ( dwRes == STILL_ACTIVE );
	}

	bool stopAndCloseThread()
	{
		setAndCloseEvent();

		if ( m_hThread == NULL )
			return true;
		
		DWORD dwRes = 0;

		if ( isActiveThread() )
		{
			//
			//	Thread function has to check this flag every time to be stopped without termination.
			//
			*m_pCancelled = true;

			ODS_fls( cWCStr().FormatT(L"Start waiting thread 0x%04X", GetThreadId( m_hThread ) ) );

			dwRes = waitThread();
			
			if ( dwRes == WAIT_TIMEOUT )
			{
				ODS_fls( cWCStr().FormatT( L"Warning! Timeout in waiting thread 0x%04X", GetThreadId( m_hThread ) ) );
				// TerminateThread( m_hThread, -1 );
			}
			else
				ODS_fls( cWCStr().FormatT( L"Stop waiting thread 0x%04X", GetThreadId( m_hThread ) ) );
		}

		CloseHandle( m_hThread );
		m_hThread = NULL;

		return dwRes != WAIT_TIMEOUT;
	}
};
