#pragma once

//
//	Saves and reads flags from registery, currently it uses 32 bit DWORD
//	31.10.2011 See cAppFlags.h the class supports by default 64 bits, and may be extended.
//
//	Get and Set function of predefined flags
//	Note, GetFlag function returns true if flag has never set before.
//
//	Usage:
//	#include "..\..\Includes\cRegFlags.h"
//
//	class cAppFlags : public cRegFlags
//	{
//	public:
//		enum { flag1 };
//		cAppFlags() : cRegFlags( L"Software\\Vendor\\AppName" ) {}
//	}
//
//	extern cAppFlags appFlags;
//	...
//
//	cAppFlags appFlags;
//	if ( appFlags.GetFlag( cAppFlags::flag1 ) ) { ... }
//

#ifndef KEY_WOW64_32KEY
#define KEY_WOW64_32KEY 0x200
#define KEY_WOW64_64KEY 0x100
#endif

enum { sveNative = 0, sve32 = KEY_WOW64_32KEY, sve64 = KEY_WOW64_64KEY };

class cRegFlags
{
	wchar_t *m_keyName;
	wchar_t *m_valueName;
	HKEY m_hRoot;

public:

	wchar_t *duplicate( const wchar_t *wc )
	{
		if ( wc == NULL ) return NULL;
		int len = (int)wcslen( wc );
		if ( len == 0 ) return NULL;

		wchar_t *wcNew = new wchar_t[ len  + 1 ];

		if ( wcNew == NULL ) return NULL;

		return wcsncpy( wcNew, wc, len + 1 );
	}

	void free( wchar_t **pwc )
	{
		if ( *pwc == NULL ) return;
		int len = (int)wcslen( *pwc );
		memset( *pwc, 0, len * sizeof( wchar_t ) );
		delete *pwc;
		*pwc = NULL;
	}

	cRegFlags()
	{
		memset( this, 0, sizeof( *this ) );
	}

	cRegFlags( const wchar_t *regKeyName, const wchar_t *regValueName = L"Flags" )
	{
		Init( regKeyName, regValueName );
	}

	void Init( const wchar_t *regKeyName, const wchar_t *regValueName = L"Flags" )
	{
		m_hRoot = HKEY_CURRENT_USER;
		m_keyName = duplicate( regKeyName );
		m_valueName = duplicate( regValueName );
	}

	~cRegFlags()
	{
		free( &m_keyName );
		free( &m_valueName );
	}

	//
	// GetFlag
	//	returns true if flag has never set before.
	//
	bool GetFlag( int flag )
	{
		assert( flag < 32 );
		DWORD dwFlags = GetRegDW( m_hRoot, m_keyName, m_valueName, true, -1 );

		DWORD mask = 1 << flag;

		return ( dwFlags & mask ) != 0;
	}

	bool SetFlag( int flag, bool set )
	{
		assert( flag < 32 );
		DWORD dwFlags = GetRegDW( m_hRoot, m_keyName, m_valueName, true, -1 );

		DWORD mask = 1 << flag;

		if ( set )
			dwFlags |= mask;
		else
			dwFlags &= ~mask;

		return SetRegDW( m_hRoot, m_keyName, m_valueName, dwFlags, true ) == TRUE;
	}

	bool SetRegDW( const wchar_t *subKey, DWORD value, BOOL useBCSection = false, int sveFlags = sveNative )
	{
		return SetRegDW( m_hRoot, m_keyName, subKey, value, useBCSection, sveNative );
	}

	static bool SetRegDW( HKEY root, const wchar_t *section, const wchar_t *subKey, DWORD value, BOOL useBCSection, int sveFlags = sveNative )
	{
		HKEY hKey;
		DWORD disp;

		if ( ( RegOpenKeyExT( root, section, 0, KEY_SET_VALUE | sveFlags, &hKey ) != ERROR_SUCCESS ) &&
			 ( RegCreateKeyExT( root, section, 0, NULL, 0, KEY_WRITE | sveFlags, NULL, &hKey, &disp ) != ERROR_SUCCESS )
			 )
				return FALSE;

		bool ok = ( RegSetValueExT( hKey, subKey, NULL, REG_DWORD, LPBYTE(&value), sizeof(DWORD) ) == ERROR_SUCCESS );
		
		RegCloseKey(hKey);

		return ok;
	}
	
	DWORD GetRegDW( const wchar_t * subKey, BOOL useBCSection = false, DWORD defValue = -1, int sveFlags = sveNative )
	{
		return GetRegDW( m_hRoot, m_keyName, subKey, useBCSection, defValue, sveFlags );
	}

	static DWORD GetRegDW( HKEY root, const wchar_t *section, const wchar_t * subKey, BOOL useBCSection, DWORD defValue, int sveFlags = sveNative )
	{
		HKEY hKey;

		if (RegOpenKeyExT( root, section, 0, KEY_READ | sveFlags, &hKey ) != ERROR_SUCCESS)
			return defValue;

		DWORD Data;
		DWORD cbData = sizeof(DWORD);

		LONG res = RegQueryValueExT( hKey, subKey, NULL, NULL, LPBYTE(&Data), &cbData );
		
		RegCloseKey(hKey);

		if ( res != ERROR_SUCCESS)
			return defValue;
		

		return Data;
	}
};