#pragma once

#include <windows.h>
#include "ODS.h"
#include "cSysCriticalSection.h"

//
//	cSemaphore
//		Class for creating, opening, wating and releasing seaphore.
//		By default it create non secure ( with default security, security parameter for CreateSemaphore is NULL )
//		bSecure parameter in creating semaphore allows to create semaphore visible for Everyone group
//		( see CreateSecureSemaphore ).
//		Note, secure semaphore has to begin from L"Global\\..."
//
class cSemaphore
{
public:

	HANDLE m_hSemaphore;
	int m_err;
	long m_prev;
	bool m_bExisted;  // createSemaphore assigns this member.

	cSemaphore( long lInitialCount = 0, long lMaximumCount = 1, const wchar_t *lpName = NULL, bool bSecure = false )
	{
		m_prev = 0;
		m_hSemaphore = NULL;
		m_bExisted = false;

		if ( lpName != NULL )
			createSemaphore( lInitialCount, lMaximumCount, lpName, bSecure );
	}

	~cSemaphore()
	{
		closeSemaphore();
	}

	bool closeSemaphore()
	{
		if ( m_hSemaphore == NULL )
			return false;

		CloseHandle( m_hSemaphore );
		m_hSemaphore = NULL;

		return true;
	}

	//
	//	It will not close delegated semaphore at destructor...
	//
	HANDLE delegateHandle()
	{
		HANDLE hDelegate = m_hSemaphore;
		m_hSemaphore = NULL;
		return hDelegate;
	}

	bool createSemaphore( long lInitialCount, long lMaximumCount, const wchar_t *lpName, bool bSecure = false );
	bool createSemaphore( const wchar_t *lpName, bool bSecure = false ) { return createSemaphore( 0, 1, lpName, bSecure ); }

	bool openSemaphore( DWORD dwDesiredAccess, bool bInheritHandle, const wchar_t *lpName );
	bool openSemaphore( const wchar_t *lpName ) { return openSemaphore( SEMAPHORE_MODIFY_STATE, false, lpName ); }

	DWORD waitForSemaphore( DWORD dwMilliseconds = INFINITE );

	static DWORD waitForSemaphore( const wchar_t *lpName, DWORD dwMilliseconds = INFINITE )
	{
		//ODSFuncS( lpName );

		DWORD res = (DWORD)-1;

		cSemaphore cs;
		if ( cs.openSemaphore( lpName ) )
		{
			res = cs.waitForSemaphore( dwMilliseconds );
			ODS_f( cWCSF().FormatT( L"Wait result %d ( dwMilliseconds = 0x%X )", res, dwMilliseconds ) );
		}

		//ODSFuncRes( res );

		return res;
	}

	static DWORD waitForSemaphoreClosed( const wchar_t *lpName, DWORD dwMilliseconds = INFINITE )
	{
		ODSFuncS( lpName );

		DWORD dwWaitTime = 0;
		DWORD dwStep = 100;

		DWORD res = WAIT_ABANDONED;

		for(;;)
		{
			cSemaphore cs;
			if ( !cs.openSemaphore( lpName ) )
			{
				res = WAIT_OBJECT_0;
				break;
			}
			
			Sleep( dwStep );
			dwWaitTime += dwStep;

			if ( dwMilliseconds != INFINITE )
			{
				if ( dwWaitTime > dwMilliseconds )
				{
					res = WAIT_TIMEOUT;
					break;
				}
			}
		}

		ODSFuncRes( res );
		return res;
	}

	bool releaseSemaphore( long lReleaseCount = 1, long *lpPreviousCount = NULL );
	static bool releaseSemaphore( const wchar_t *lpName, long lReleaseCount = 1, long *lpPreviousCount = NULL );

	static bool isActive( const wchar_t *lpName )
	{
		cSemaphore cs;
		return cs.openSemaphore( SYNCHRONIZE, false, lpName );
	}

	static DWORD waitForMultipleObjects( bool bWaitAll, DWORD timeout, HANDLE obj0, HANDLE obj1, HANDLE obj2 = NULL, HANDLE obj3 = NULL, HANDLE obj4 = NULL ) 
	{	
		DWORD count = 0;
		HANDLE hT[] = { obj0, obj1, obj2, obj3, obj4 };

		for( count = 0; count < (sizeof( hT )/sizeof(*hT)); ++count )
		{
			if ( hT[count] == NULL )
				break;
		}

		return WaitForMultipleObjects( count, hT, bWaitAll, timeout );  
	}
};

//
//	cCriticalSectionForDifferentProcesses
//		cSysCriticalSection uses CRITICAL_SECTION but it does not work for different processes.
//		The class uses named mutex.
//	Useage 1:
//	cCriticalSectionForDifferentProcesses cs( _UniqueString );
//	cs.Lock();
//	...
//	cs.Unlock();
//
//	Useage 2:
//	cCriticalSectionForDifferentProcesses cs( _UniqueString, true );
//
class cCriticalSectionForDifferentProcesses
{
	HANDLE  m_mtx;
	bool m_bAutoLock;

public:

	cCriticalSectionForDifferentProcesses( const wchar_t *name, bool bAutoLock = false )
	{
		m_bAutoLock = bAutoLock;
		m_mtx = ::CreateMutexW( NULL, FALSE, name );

		if ( m_bAutoLock )
			Lock();
	}

	~cCriticalSectionForDifferentProcesses()
	{
		if ( m_bAutoLock )
			Unlock();

		::CloseHandle(m_mtx);
	}

	void Lock()
	{
		::WaitForSingleObject(m_mtx, INFINITE);
	}

	void Unlock()
	{
		::ReleaseMutex(m_mtx);
	}

	bool    TryLock()
	{
		if (WAIT_OBJECT_0 == ::WaitForSingleObject(m_mtx, 50))
			return TRUE;
		return FALSE;
	}

	int RunDefended( cCriticalSectionActorBlock &scs, DWORD dwParam )
	{
			int res = -1;
		Lock();
			res = scs.scsActor( dwParam );
		Unlock();
			return res;
	}

};

//
//	cCriticalSection
//		Just creates named semaphore in nonsignaged state with warrancy that it is a single semaphore object in this session.
//
class cCriticalSection : public cSemaphore
{
public:
	cCriticalSection( const wchar_t *lpName, bool bSecure = false );

	static bool waitFor( const wchar_t *lpName )
	{
		return waitForSemaphore( lpName ) != -1;
	}
};

#define CRITICAL_SECTION_OBJECT( _bSecure ) cCriticalSection _CCS( _UniqueString, _bSecure	)
