/*
 *	Copyright 2018 Jetico Inc. All rights reserved.
 *	cSysCriticalSection.h
 *	System critical section object...
 *	Created:	28.09.2018
 *	Author:		Alexander Pichuev.
 */

#pragma once
#include "ODS.h"

//
//	Critical section actor block class.
//		See RunDefended functions from cCriticalSectionForDifferentProcesses or cSysCriticalSection
//
class cCriticalSectionActorBlock
{
public:
	virtual int scsActor( DWORD param ) = 0;
};

//
// cSysCriticalSection
//
// Usage inside thread function:
//    static cSysCriticalSection cs;
//    cs.Enter();
//  ....
//    cs.Leave();
//

class cSysCriticalSection_Block;

class cSysCriticalSection : public CRITICAL_SECTION
{
    friend cSysCriticalSection_Block;

    // It is used to check in debug the section using.
    bool m_bIsEnterred;
	bool m_bIsDeleted;

    const wchar_t *m_szFunction;
    int m_line;

public:
    cSysCriticalSection()
    {
        InitializeCriticalSection( this );
        m_bIsEnterred = false;
		m_bIsDeleted = false;
    }

    ~cSysCriticalSection()
    {
        DeleteCriticalSection( this );
		m_bIsDeleted = true;
    }

    bool inline isEnterred()
    {
        return m_bIsEnterred;
    }

	void Enter( const wchar_t *szFunction = NULL, int line = 0 )
	{
		//
		//	23.10.2018 - Problem was detected!
		//	Critical section is used at objects' getStatic() functions. The critical section is statoc object too.
		//	So if an object at destructor uses already deleted object so the m_bIsDeleted is true.
		//	We have to avoid calling static object at destructor of other static object.
		//
		if ( m_bIsDeleted )
			ODS_fls( L"Assertion! Critical section is deleted!" );

		EnterCriticalSection( this );
		m_bIsEnterred = true;
		m_szFunction = szFunction;
		m_line = line;
	}

	void Leave()
	{
		m_bIsEnterred = false;
		m_szFunction = NULL;
		m_line = 0;
		LeaveCriticalSection( this );
	}

    int RunDefended( cCriticalSectionActorBlock &scs, DWORD dwParam )
    {
            int res = -1;
        Enter( (_CRT_WIDE( __FUNCTION__)), __LINE__ );
            res = scs.scsActor( dwParam );
        Leave();
            return res;
    }
};

//
// cSysCriticalSection
//
// Usage inside thread function:
//    static cSysCriticalSection staticCS;
//    cSysCriticalSection_Block scBlock( &staticCS );
//

#define SYS_CRITICAL_SECTION_BLOCK( obj ) cSysCriticalSection_Block _SCSB( obj, (_CRT_WIDE( __FUNCTION__)), __LINE__ )
#define SYS_CRITICAL_SECTION( _BLOCK, obj ) cSysCriticalSection_Block _BLOCK( obj, (_CRT_WIDE( __FUNCTION__)), __LINE__ )

class cSysCriticalSection_Block
{
    cSysCriticalSection *m_object;

public:

    cSysCriticalSection_Block( cSysCriticalSection *staticObject, const wchar_t *szFunction = NULL, int line = 0 )
	{
		m_object = staticObject;

		if ( m_object->isEnterred() )
			ODS( szFunction, line, _FormatT(  L"Warning! Section is owned by %s::%d", m_object->m_szFunction, m_object->m_line ) );

		m_object->Enter( szFunction, line );
	}

	~cSysCriticalSection_Block()
    {
        m_object->Leave();
    }
};

/*
//
// cSysCriticalSection
//
// Usage inside thread function:
//	static cSysCriticalSection cs;
//	cs.Enter();
//  ....
//	cs.Leave();
//
class cSysCriticalSection : public CRITICAL_SECTION
{
public:
	cSysCriticalSection()
	{
		InitializeCriticalSection( this );
	}

	~cSysCriticalSection()
	{
		DeleteCriticalSection( this );
	}

	void Enter()
	{
		//ODSFunc;
		EnterCriticalSection( this );
	}

	void Leave()
	{
		//ODSFunc;
		LeaveCriticalSection( this );
	}

	int RunDefended( cCriticalSectionActorBlock &scs, DWORD dwParam )
	{
			int res = -1;
		Enter();
			res = scs.scsActor( dwParam );
		Leave();
			return res;
	}
};

//
// cSysCriticalSection
//
// Usage inside thread function:
//	static cSysCriticalSection staticCS;
//	cSysCriticalSection_Block scBlock( &staticCS );
//
class cSysCriticalSection_Block
{
	cSysCriticalSection *m_object;

public:
	cSysCriticalSection_Block( cSysCriticalSection *staticObject )
	{
		m_object = staticObject;
		m_object->Enter();
	}
	
	~cSysCriticalSection_Block()
	{
		m_object->Leave();
	}
};

*/