/*
 *	Copyright 2013-2015 Jetico Inc. All rights reserved.
 *	cWCS.cpp
 *	Base class for wide character string classes like cWCSF, cWCStr ..
 *	Created:	30.10.2013
 *	Author:		Alexander Pichuev.
 */

#include "cWCS.h"
#include <string.h>
#include "windefs.h"
#include <assert.h>

bool cWCS::isEmpty()
{
	if ( ( m_len == 0 ) || ( m_szBody == 0 ) )
		return true;

	for( int i = 0; i < m_len; i++ )
	{
		if ( m_szBody[i] != L'\0' )
			return false;
	}
	
	return true;
}

void cWCS::Swap( cWCS *wcs )
{
	wchar_t *szBody = m_szBody;
	int len = m_len;
	int allocatedLen = m_allocatedLen;

	m_szBody = wcs->m_szBody;
	m_len = wcs->m_len;
	m_allocatedLen = wcs->m_allocatedLen;

	wcs->m_szBody = szBody;
	wcs->m_len = len;
	wcs->m_allocatedLen = allocatedLen;
}

void cWCS::Free()
{
	if ( m_szBody != NULL )
	{
#ifndef _noClearWcs
		memset( m_szBody, 0, m_len * 2 );
#endif
		freeStr( m_szBody, &m_allocatedLen );
		m_szBody = NULL;
		m_len = 0;
	}
}

int cWCS::Add( const wchar_t *wcs, int lenAdd ) //  = NULL, = -1
{
	int lenAddAlloc;

	if ( wcs == NULL )
	{
		if ( lenAdd < 0 )
			lenAdd = 0;

		lenAddAlloc = lenAdd + 1;

		lenAdd = 0;
		wcs = L"";
	}
	else
	{
		if ( lenAdd < 0 )
			lenAdd = (int)wcslen( wcs );

		lenAddAlloc = lenAdd + 1;
	}

	int lenOld = m_len;

	wchar_t *bf = NULL;
	int allocatedLen = 0;

	if ( lenOld + lenAddAlloc < m_allocatedLen )
	{
		bf = m_szBody;
		allocatedLen = m_allocatedLen;
	}
	else
	{
		bf = allocateStr( lenOld + lenAddAlloc, &allocatedLen );

		if ( bf == NULL )
			return 0;

		memcpy_s( bf, allocatedLen * 2, m_szBody, lenOld * 2 );

		Free();
	}

	memcpy_s( (BYTE*)bf + lenOld * 2, lenAddAlloc * 2, wcs, lenAdd * 2 );
	memset( (BYTE*)bf + lenOld * 2 + lenAdd * 2, 0, ( lenAddAlloc - lenAdd ) * 2 );

	m_szBody = bf;
	m_len = lenOld + ( lenAddAlloc - 1 );
	m_allocatedLen = allocatedLen;

	return lenAddAlloc;
}

//
//	SetLength
//		If string is more then 'len' so string length is truncated.
//		If allocated buffer is less then 'len' so the buffer is extended.
//
int cWCS::SetLength( int len )
{
	if ( len == -1 )
	{
		if ( m_len == 0 )
			len = 0;
		else
			len = (int)wcslen( m_szBody );
	}

	if ( m_len < len )
	{
		if ( Add( len - m_len ) == 0 )
			return 0;
	}
	else
	if ( m_len > len )
	{
		m_len = len;
		m_szBody[ m_len ] = L'\0';
	}

	return m_len + 1;
}

//
//	Safely gets length of buffer.
//
int cWCS::GetLength()
{
	if ( m_szBody == 0 )
		return 0;

	for( int i = 0; ; i++ )
	{
		if ( i == m_len )
			return i;

		if ( m_szBody[i] == L'\0' )
			return i;
	}
}

//
//	AddLength
//		adds zero buffer specified len to string tail
//		returns pointer to added tail if success otherwise returns NULL 
//
wchar_t *cWCS::AddLength( int lenAdd )
{
	if ( lenAdd == 0 )
		return m_szBody;

	int len = 0;
	if ( m_szBody != NULL )
		len = (int)wcslen( m_szBody );

	if ( SetLength( len + lenAdd ) == 0 )
		return NULL;

	return m_szBody + len;
}

//
// The reassigment is required to exclude copy constructor.
//	Otherwise it copies m_szBody as pointer.
//
cWCS::cWCS( cWCS &wcs, bool bSwap ) // = false )
{
	m_szBody = NULL;
	m_len = 0;
	m_allocatedLen = 0;

	if ( bSwap )
		Swap( &wcs );
	else
		Add( wcs.m_szBody, wcs.m_len );
}

cWCS::cWCS( const wchar_t *wcs, int len ) //  = NULL, = -1 )
{
	m_szBody = NULL;
	m_len = 0;
	m_allocatedLen = 0;

	Add( wcs, len );
}

cWCS::cWCS( int len )
{
	m_szBody = NULL;
	m_len = 0;
	m_allocatedLen = 0;

	Add( (const wchar_t *)NULL, len );
}

// static
wchar_t *cWCS::allocateStr( int wcsLen, int *pAllocatedLen )
{
	size_t size = wcsLen * sizeof( wchar_t );

#ifdef _DEBUG
	static size_t blockSize = 1 << 12;
	size += sizeof(DWORD) * 2;
#else
	const size_t blockSize = 1 << 12;
#endif

	size = ( ( size + blockSize - 1 ) & ~( blockSize - 1) );

	void *hMem = memAllocate( size );

	if ( hMem == NULL )
	{
		*pAllocatedLen = 0;
		return NULL;
	}

#ifdef _DEBUG

	DWORD *p = (DWORD*)hMem;
	p[0] = 0xBCBA;
	p[1] = (DWORD)wcsLen;
	
	size -= sizeof(DWORD) * 2;
	hMem = (HGLOBAL)(p + 2);

#endif

	*pAllocatedLen = (int)(size / sizeof( wchar_t ));
	return (wchar_t*)hMem;
}

//static 
void cWCS::freeStr( wchar_t *wcs, int *pAllocatedLen )
{
	if ( wcs == NULL )
		return;

	if ( wcs == (wchar_t *)0xfefefefe )
	{
		//assert( false );
		return;
	}

	memset( wcs, 0, *pAllocatedLen * sizeof( wchar_t ) );
	*pAllocatedLen = 0;

#ifdef _DEBUG
	DWORD *p = (DWORD*)wcs - 2;
	wcs = (wchar_t*)p;
#endif
	memFree( wcs ); //GlobalFree( (HGLOBAL)wcs );
}

//
//	Inserts string at position.
//	If len is -1 so zero termination symbol is added.
//
const wchar_t *cWCS::InsertAt( int pos, const wchar_t *str, int len ) // = -1 )
{
	if ( len == -1 )
		len = (int)wcslen( str ) + 1; // zero termination symbol will be added.

	if ( pos + len > m_len )
	{
		SetLength( pos );
		Add( str, len );
	}
	else
		wmemcpy_s( m_szBody + pos, m_allocatedLen - pos, str, len );

	return m_szBody;
}

//static
int cWCS::cmp( const wchar_t *wcs0, const wchar_t *wcs1, int len ) // = -1 )
{
	if ( wcs0 == NULL )
	{
		if ( wcs1 == NULL )
			return 0;

		return -1000;
	}
	else
	if ( wcs1 == NULL )
		return 1000;

	return wcsncmp( wcs0, wcs1, len );
}

//static
int cWCS::cmpI( const wchar_t *wcs0, const wchar_t *wcs1, int len ) // = -1 )
{
	if ( wcs0 == NULL )
	{
		if ( wcs1 == NULL )
			return 0;

		return -1000;
	}
	else
	if ( wcs1 == NULL )
		return 1000;

	return _wcsnicmp( wcs0, wcs1, len );
}

wchar_t cWCS::getLast( unsigned int last ) // = 0 )
{
	if ( ( m_szBody == NULL ) || ( m_len == 0 ) || ( (unsigned int)m_len - 1 < last ) )
		return 0;

	return m_szBody[ m_len - 1 - last ];
}

wchar_t cWCS::getAt( int pos )
{
	if ( ( m_szBody == NULL ) || ( pos < 0 ) || ( pos >= m_len ) )
	{
		//ODS_fls( L"Assertion!" );
		return 0;
	}

	return m_szBody[ pos ];
}

wchar_t cWCS::setAt( int pos, wchar_t ch )
{
	if ( ( m_szBody == NULL ) || ( pos < 0 ) || ( pos >= m_len ) )
	{
		//ODS_fls( L"Assertion!" );
		return 0;
	}

	m_szBody[ pos ] = ch;

	return m_szBody[ pos ];
}

const wchar_t *cWCS::ChangeSymbols( wchar_t existingSymbol, wchar_t newSymbol, eEncodeType encodeType ) // = eetUndefined )
{
	for ( int i = m_len; i--; )
	{
		if ( encodeType == eetEncode )
		{
			if ( m_szBody[i] == newSymbol )
			{
				cWCS tail( m_szBody + i );
				SetLength( i + 1 );
				Add( tail.m_szBody );
				continue;
			}
		}
		else
		if ( encodeType == eetDecode ) // decode
		{
			if ( ( m_szBody[i] == existingSymbol ) && 
				 ( i != 0 ) &&
				 ( m_szBody[i-1] == existingSymbol ) )
			{
				wmemmove( m_szBody + i - 1, m_szBody + i, m_len - i + 1);
				i--;
				m_len--;
				continue;
			}
		}

		if ( m_szBody[i] == existingSymbol )
			m_szBody[i] = newSymbol;
	}

	return m_szBody;
}

const wchar_t *cWCS::ChangeNewLineToTab()
{
	for ( int i = m_len; i--; )
	{
		if ( m_szBody[i] == L'\n' )
			m_szBody[i] = L'\t';

		if ( m_szBody[i] == L'\r' )
			m_szBody[i] = L' ';
	}

	return m_szBody;
}

//
//	GetString
//		Walks new line separated strings
//		Cuts off stringQ by GetFirstString.
//
const wchar_t *cWCS::GetString( const wchar_t *str, int pos, int stringQ ) // = 1 )
{
	if ( pos != 0 )
	{
		for(;;)
		{
			if ( *str == L'\0' )
				break;

			if ( *(str++) == L'\n' )
			{
				if ( --pos == 0 )
					break;
			}
		}
	}

	Set( str );
	return GetFirstString( stringQ );
}

//
//	GetFirstString
//		Cuts off first stringQ from new line separated string list.
//
const wchar_t *cWCS::GetFirstString( int stringQ ) // = 1 )
{
	for( wchar_t *p = m_szBody; *p; p++ )
	{
		if ( *p == L'\n' )
		{
			if ( --stringQ > 0 )
				continue;

			*p = L'\0';
			break;
		}
	}

	return m_szBody;
}

const wchar_t *cWCS::SetInt( int i, int radix ) // = 10 )
{
	SetLength( 32 );
	_itow_s( i, m_szBody, 32, radix );
	m_len = (int)wcslen( m_szBody );
	return m_szBody;
}

int cWCS::AddInt( int i, int radix ) // = 10 )
{
	int pos = m_len;
	if ( Add( 32 ) == 0 )
		return 0;

	_itow_s( i, m_szBody + pos, m_len - pos, radix );

	m_len = pos + (int)wcslen( m_szBody + pos );

	return m_len;
}

const wchar_t *cWCS::MakeLower()
{
	for( int i = m_len; i--; )
		m_szBody[i] = towlower( m_szBody[i] );

	return m_szBody;
}

const wchar_t *cWCS::MakeUpper()
{
	for( int i = m_len; i--; )
		m_szBody[i] = towupper( m_szBody[i] );

	return m_szBody;
}

const wchar_t *cWCS::MakePath( const wchar_t *dir, const wchar_t *element )
{
	int len = 0;

	if ( dir == NULL )
		SetLength( 0 );
	else
	{
		len = (int)wcslen( dir );
		if ( Set( dir, len ) == 0 )
			return NULL;
	}

	if ( ( len == 0 ) || (  getAt( len-1 ) != L'\\' ) )
	{
		if ( Add( L"\\" ) == 0 )
			return NULL;
	}

	if ( element != NULL )
	{
		if ( Add( element ) == 0 )
			return NULL;
	}

	return m_szBody;
}

class cFindWord : public cWCS
{
	const wchar_t *m_ptr;
	const wchar_t *m_word;
	int m_wordLen;
	const wchar_t *m_beg;

public:
	bool findFirst( const wchar_t *str, const wchar_t *word );
	void changeCurrent( const wchar_t *newWord );
	bool findNext();
	void copyTail();
};

bool cFindWord::findFirst( const wchar_t *str, const wchar_t *word )
{
	m_ptr = str;
	m_word = word;
	m_wordLen = (int)wcslen( m_word );

	if ( m_wordLen == 0 )
		return false;

	m_beg = NULL;

	return findNext();
}

void cFindWord::changeCurrent( const wchar_t *newWord )
{
	Add( m_ptr, (int)( m_beg - m_ptr ) );
	Add( newWord );
	m_ptr =	m_beg + m_wordLen;
}

bool cFindWord::findNext()
{
	if ( m_beg == NULL )
		m_beg = m_ptr;
	else
		m_beg += m_wordLen;

	m_beg = wcsstr( m_beg, m_word );
	return ( m_beg != NULL );
}

void cFindWord::copyTail()
{
	Add( m_ptr );
}

//
//	ChangeWords
//		Sample:
//			body = "user_pref('QUOTE'browser.cache.disk.parent_directory'QUOTE'"
//			presentWord = "'QUOTE'"
//			newWord		= "\""
//		Result = "user_pref(\"browser.cache.disk.parent_directory\""
//
const wchar_t *cWCS::ChangeWords( const wchar_t *presentWord, const wchar_t *newWord )
{
	cFindWord fw;

	if ( fw.findFirst( m_szBody, presentWord ) )
	{
		do
		{
			fw.changeCurrent( newWord );
		}
		while( fw.findNext() );

		fw.copyTail();
		Swap( &fw );
	}

	return m_szBody;
}

//
//	ExtractString
//		Finds szBegin and szEnd at szSource and set this content as body.
//
bool cWCS::ExtractString( const wchar_t *szSource, const wchar_t *szBegin, const wchar_t *szEnd )
{
	const wchar_t *beg = szSource;
	int len = 0;

	if ( szBegin != NULL )
	{
		beg = wcsstr( szSource, szBegin );
		if ( beg == NULL )
			return false;
		len = (int)wcslen( szBegin );

		if ( len == 0 )
			return false;

		beg += len;
	}

	const wchar_t *end;

	if ( szEnd == NULL )
		len = -1; // Up to end of string
	else
	{
		end = wcsstr( beg, szEnd );
		if ( end == NULL )
			return false;

		len = (int)( end - beg );
	}


	Set( beg, len );
	return true;
}


//virtual
int cWCS::Add( const char *str, int lenAdd )
{
	if ( lenAdd == -1 )
		lenAdd = (int)strlen( str );

	wchar_t *pos = AddLength( lenAdd );
	if ( pos == NULL )
		return 0;

	for ( int i = 0; i < lenAdd; i++ )
	{
		pos[i] = str[i];
	}

	return lenAdd;
}

