/*
 *	Copyright 2016 Jetico, Inc. All rights reserved.
 *	Abstract class cWaitRegistryKeys allows to wait multiply registry keys simultaneously.
 *		Requires:
 *			1. Opened root key
 *			2. Table of subkeys
 *			3. Optionally name of ODS file
 *		Creates separate waiting thread,
 *			1. Creates:
 *				- stop smaphore
 *				- in loop creates registry notifications accoding 
 *			2. Wait Multiple Objects ( regkey notifiers + stop semaphore )
 *			3. Calls virtual processNotificationResult
 *		
 *	Created:	09.06.2016
 *	Author:		Alexander Pichuev.
 */

#include "..\..\Includes\UnicodeFuncs.h"
#include "..\..\Includes\cRegKey.h"
#include "..\..\Includes\cStringArray.h"
#include <time.h>

class cRegNotifier : public cRegKey, public cAutoHandle
{
public:
	cRegNotifier()
	{
	}

	bool create( HKEY hRoot, const wchar_t *lpSubKey, bool bWow64 = false )
	{
		DWORD dwWow64Key = bWow64 ? KEY_WOW64_64KEY : 0;
		LONG lErrorCode = createKey( hRoot, lpSubKey, 0, KEY_READ | dwWow64Key );

		if ( lErrorCode != ERROR_SUCCESS ) // Create key if key does not exist.
			lErrorCode = createKey( hRoot, lpSubKey, 0, KEY_ALL_ACCESS | dwWow64Key );

		if ( lErrorCode != ERROR_SUCCESS )
		{	ODS_fls( _FormatT( L"Error %d in RegOpenKeyEx( %X, %s ).", lErrorCode, hRoot, lpSubKey ) ); return false; }

		m_handle = CreateEvent( NULL, TRUE, FALSE, NULL);
		if ( m_handle == NULL )
		{ ODS_fls( L"Error in CreateEvent." ); return false; }

		DWORD dwFilter = REG_NOTIFY_CHANGE_LAST_SET;

		// Watch the registry key for a change of value.
		lErrorCode = RegNotifyChangeKeyValue( m_hKey, false, dwFilter, m_handle, TRUE );

		if (lErrorCode != ERROR_SUCCESS)
		{
			ODS_fls( L"Error in RegNotifyChangeKeyValue." );
			return false;
		}

		return true;
	}
};

class cWaitRegistryKeys
{
protected:
	const wchar_t *m_szThreadName;
	cStringArrayT m_saSubKeys;
	HKEY m_hRoot;
	bool m_bWow64;

	enum { eSig = 0x6377726B }; // cwrk
	DWORD m_sig;

	HANDLE m_hThread;
	const wchar_t *m_szStopSemaphore;
	DWORD m_timeout;

private:
	//
	//	Main function, it 
	//		create notifiers and stop semaphore
	//		waits all the objects.
	//		calls processNotificationResult.
	//
	DWORD waitFunction();

public:

	cWaitRegistryKeys()
	{
		m_hRoot = NULL;
		m_bWow64 = false;
		m_sig = eSig;
		m_szThreadName = NULL;
		m_hThread = NULL;
		m_timeout = INFINITE;

		// Uniquie name.
		m_szStopSemaphore = newString( _FormatT( L"Global\\BCWipe_WaitRegKes_StopWaiting_%X", (DWORD)clock() ) );
	}
	
	~cWaitRegistryKeys()
	{
		if ( m_hThread != NULL )
			stopWaiting();

		deleteString( &m_szStopSemaphore );
	}

	bool isValid()
	{
		if ( ( m_sig == eSig ) &&
			 ( m_hRoot != NULL ) && 
			 ( m_szStopSemaphore != NULL ) )
				return true;

		return false;
	}

	static bool checkValid( cWaitRegistryKeys *obj );
	static DWORD WINAPI waitFunction_static( void *ptr );


	//
	//	Create thread with waitFunction().
	//
	bool startWaiting();

	//
	//	Releass stop semaphore and waits thread finishes.
	//
	DWORD stopWaiting(); // return thread error code.

	//
	//	Returns number of key has been modified from table m_saSubKeys.
	//
	virtual bool processNotificationResult( DWORD res ) = 0;


	//
	//	Is called by timeout, it can stop process if return false.
	//
	virtual bool processTimeout()
	{
		return true;
	}
};

