/*******************************************************************\
*	Copyright (c) 1994-2018 Jetico, Inc., Finland
*	All rights reserved.
*
*	File: alg.cpp
*	Revision: 
*	Created:
*	Description: implementation of the interface with 
*			 Algorithm Driver
*
\*******************************************************************/

#include <windows.h>
#include "bc_ioctl.h"
#include "alg.h"
#include "UnicodeFuncs.h"
#include "kBlock.h"
#include <assert.h>
#include "..\..\..\CertificateLib\hashAlgs\hashAlgs.h"
#include "..\..\..\CertificateLib\SMemory\cPassword.h"
#include "resource.h"
#include "AutoDeallocate.h"
#include "ods.h"

static cADTable < cSymAlgInfo > symAlgs( 0 );

enum ealgid { algId_TWOFISH = 160, algId_BLOWFISH_448 = 176, algId_AES = 240, algId_Ghost = 170, algId_Dummy = 254 };

int algIdT[] = {
	algId_AES,
	algId_TWOFISH,
	algId_BLOWFISH_448
};

//
//	getAlgIdPriority
//		0 is maximal priority
//		DWORD(-1) is minimal priority
//
DWORD getAlgIdPriority( int id )
{
	int i =  maxLen( algIdT );

	while( i-- )
	{
		if ( algIdT[i] == id )
			break;
	}

	return (DWORD)i;
}


//
//	sortAlgorithms
//		Sorts algorithms, the old algorithms will be added to tail,
//		the encrypted header will check new algorithms first.
//
static bool sortAlgorithms()
{
	//
	//	before nnnoonn
	//	after  nnnnnoo
	//

	bool alreadySorted = true;

	for( int last = symAlgs.m_Q - 1; last > 0; last-- )
	{
		bool bSwapped = false;

		for( int j = 0; j < last; j++ )
		{
			if ( 
				( symAlgs.m_T[j].m_oldType && !symAlgs.m_T[j + 1].m_oldType ) ||
				( getAlgIdPriority( symAlgs.m_T[j].m_algId ) > getAlgIdPriority( symAlgs.m_T[j+1].m_algId ) )
				)
			{
				swapValues( symAlgs.m_T + j, symAlgs.m_T + j + 1 );
				bSwapped = true;
			}
		}

		if ( !bSwapped )
			break;
		else
			alreadySorted = false;
	}

	return alreadySorted;
}

bool SetSymAlgInfo( cSymAlgInfo *T, int Q )
{
	ODSFunc;

	symAlgs.Free();

	bool ok = ( symAlgs.Add( T, Q ) != NULL );

	bool isWinNT = GetVersion() < 0x80000000;

	if ( !isWinNT )
	{
		for( int i = Q; i--; )
		{
			strcat_s( symAlgs.m_T[i].m_algServiceName, ".vxd" ); 

//			ODSFE( L"SetSymAlgInfo T[%d].m_algServiceName = %s", i, getWCStr( T[i].m_algServiceName ) );
		}
	}

	sortAlgorithms();

	return ok;
}

BOOL Alg_VerifyDriver( LPCSTR AlgServiceName, UCHAR *pVMajor, UCHAR *pVMinor )
{
	char           algService[128];
	VERSION_BUFFER vBuf;
	BOOL           res = TRUE;
	DWORD          lBytesReturned;
	HANDLE         hDevice;

	strcpy_s(algService, "\\\\.\\");
	strcat_s(algService, AlgServiceName );

	memcpy( vBuf.Signature, "LOCOS97 ", maxLen( vBuf.Signature ) );

	hDevice = CreateFileA( algService,
							GENERIC_READ | GENERIC_WRITE, 0, NULL,
							OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL );

	if (hDevice == INVALID_HANDLE_VALUE) return FALSE;

	res = DeviceIoControl( hDevice, IOCTL_PRIVATE_GET_VERSION_INFO,
							&vBuf, sizeof(VERSION_BUFFER),
							&vBuf, sizeof(VERSION_BUFFER),
							&lBytesReturned, NULL);
	CloseHandle(hDevice);

	if (strncmp(vBuf.Signature, "LS06CXX ", 8) != 0)
		return FALSE;

	if ((vBuf.Status) != 0)
		return FALSE;

	if ( pVMajor != NULL )
		*pVMajor = vBuf.Major_version;

	if ( pVMinor != NULL )
		*pVMinor = vBuf.Minor_version;

	return res;
}

BOOL Alg_GetKeyHandler(
		LPCSTR AlgServiceName,
		DWORD	encModeId, // Really it does not required here, because algorithm driver always geneates LRW context here.
		UCHAR  *Key, 
		DWORD  KeyLength,      //in bites
		BYTE   *Pool,
		DWORD  *KeyHandle
	)
{
	char             algService[128];
	CREATEKEY_BUFFER ckBuf;
	BOOL             res = TRUE;
	DWORD            lBytesReturned, lengthByte;
	HANDLE           hDevice;

	strcpy_s(algService, "\\\\.\\");
	strcat_s(algService, AlgServiceName );

	memcpy( ckBuf.Signature, "LOCOS97 ", maxLen( ckBuf.Signature ) );
	ckBuf.KeyLength = KeyLength;

	lengthByte = ( KeyLength + 7 ) / 8;

	if ( lengthByte > sizeof( ckBuf.Key ) )
	{
		assert( false );
		return false;
	}

	memcpy( ckBuf.Key, Key, lengthByte );
	memcpy( ckBuf.Pool, Pool, DB_PoolSize );

	hDevice = CreateFileA( algService,
							GENERIC_READ | GENERIC_WRITE, 0, NULL,
							OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL );

	if (hDevice == INVALID_HANDLE_VALUE)
	{
		//ODSF( ::FormatOnceT( L"CreateFile( %s ) = 0x%X ", getWCStr( algService ), hDevice ) );


		return FALSE;
	}

	res = DeviceIoControl( hDevice, IOCTL_PRIVATE_CREATE_KEY_HANDLE,
							&ckBuf, sizeof(CREATEKEY_BUFFER),
							&ckBuf, sizeof(CREATEKEY_BUFFER),
							&lBytesReturned, NULL);

	CloseHandle(hDevice);

	*KeyHandle = ckBuf.KeyHandle;

	if (strncmp(ckBuf.Signature, "LS06CXX ", 8) != 0)
		return FALSE;

	if ((ckBuf.Status) != 0)
		return FALSE;

	return res;
}

BOOL Alg_FreeKeyHandler( LPCSTR AlgServiceName, DWORD KeyHandle )
{
	char           algService[64];
	TESTKEY_BUFFER tBuf;
	BOOL           res = TRUE;
	DWORD          lBytesReturned;
	HANDLE         hDevice;

	strcpy_s(algService, "\\\\.\\" );
	strcat_s(algService, AlgServiceName );

	memcpy( tBuf.Signature, "LOCOS97 ", maxLen( tBuf.Signature ) );

	tBuf.KeyHandle = KeyHandle;

	hDevice = CreateFileA( algService,
							GENERIC_READ | GENERIC_WRITE, 0, NULL,
							OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL );

	if (hDevice == INVALID_HANDLE_VALUE) return FALSE;

	res = DeviceIoControl( hDevice, IOCTL_PRIVATE_FREE_KEY_HANDLE,
							&tBuf, sizeof(TESTKEY_BUFFER),
							&tBuf, sizeof(TESTKEY_BUFFER),
							&lBytesReturned, NULL);
	CloseHandle(hDevice);

	if (!res) return FALSE;

	if (strncmp(tBuf.Signature, "LS06CXX ", 8) != 0)
		return FALSE;

	if ((tBuf.Status) != 0) return FALSE;

	return TRUE;
}

BOOL Alg_Encrypt_V7_CBC(
		LPCSTR AlgServiceName,
		DWORD  KeyHandle, 
		BYTE   *IVector, 
		BYTE   *Buffer, 
		DWORD  Length
	)
{
	char           algService[64];
	ENCRYPT_BUFFER eBuf;
	BOOL           res = TRUE;
	DWORD          lBytesReturned;
	HANDLE         hDevice;

	strcpy_s(algService, "\\\\.\\");
	strcat_s(algService, AlgServiceName );

	memcpy( eBuf.Signature, "LOCOS97 ", maxLen( eBuf.Signature ) );

	eBuf.KeyHandle = KeyHandle;
	memcpy( eBuf.IVector, IVector, 8);
	eBuf.Length    = Length;

	hDevice = CreateFileA( algService,
							GENERIC_READ | GENERIC_WRITE, 0, NULL,
							OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL );
	if (hDevice == INVALID_HANDLE_VALUE) return FALSE;

	res = DeviceIoControl( hDevice, IOCTL_PRIVATE_ENCRYPT,
							&eBuf, sizeof(ENCRYPT_BUFFER),
							(LPVOID)Buffer, Length,
							&lBytesReturned, NULL);
	CloseHandle(hDevice);

	if (!res) return FALSE;

	return TRUE;
}

BOOL Alg_Decrypt_V7_CBC(
		LPCSTR AlgServiceName,
		DWORD  KeyHandle, 
		BYTE   *IVector, 
		BYTE   *Buffer, 
		DWORD  Length
	)
{
	char           algService[64];
	ENCRYPT_BUFFER eBuf;
	BOOL           res = TRUE;
	DWORD          lBytesReturned;
	HANDLE         hDevice;

	strcpy_s(algService, "\\\\.\\");
	strcat_s(algService, AlgServiceName );

	memcpy( eBuf.Signature, "LOCOS97 ", maxLen( eBuf.Signature ) );

	eBuf.KeyHandle = KeyHandle;
	memcpy( eBuf.IVector, IVector, 8);
	eBuf.Length    = Length;

	hDevice = CreateFileA( algService,
							GENERIC_READ | GENERIC_WRITE, 0, NULL,
							OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL );
	if (hDevice == INVALID_HANDLE_VALUE) return FALSE;

	res = DeviceIoControl( hDevice, IOCTL_PRIVATE_DECRYPT,
							&eBuf, sizeof(ENCRYPT_BUFFER),
							(LPVOID)Buffer, Length,
							&lBytesReturned, NULL);
	CloseHandle(hDevice);

	if (!res) return FALSE;

	return TRUE;
}

static bool isCBCMode( DWORD encModeId )
{
	if ( encModeId == 0 )
		return true;

	if ( encModeId == EX_ENCRYPT_CBC_MODE_FLAG )
		return true;

	return false;
}

BOOL Alg_Encrypt(
		LPCSTR AlgServiceName,
		DWORD  KeyHandle, 
		DWORD  encModeId,
		BYTE   *IVector, 
		BYTE   *Buffer, 
		DWORD  Length
	)
{
	assert( KeyHandle != NULL );
	if ( KeyHandle == NULL )
		return false;

	char           algService[64];

	BOOL           res = TRUE;
	DWORD          lBytesReturned;
	HANDLE         hDevice;

	strcpy_s(algService, "\\\\.\\");
	strcat_s(algService, AlgServiceName );

	ENCRYPT_BUFFER_V8 eBuf;
	memset( &eBuf, 0, sizeof( eBuf ) );
	eBuf.Size = sizeof( eBuf );

	eBuf.EncryptOperation = true;

	eBuf.EncryptionMode = encModeId;

	memcpy( eBuf.Signature, "LOCOS97 ", maxLen( eBuf.Signature ) );
	eBuf.KeyHandle = KeyHandle;
	memcpy( eBuf.IVector, IVector, 8);
	eBuf.Length    = Length;

	hDevice = CreateFileA( algService,
							GENERIC_READ | GENERIC_WRITE, 0, NULL,
							OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL );

	if (hDevice == INVALID_HANDLE_VALUE) return FALSE;

	res = DeviceIoControl( hDevice, IOCTL_PRIVATE_ENCRYPT_DECRYPT_V8,
							&eBuf, sizeof(eBuf),
							(LPVOID)Buffer, Length,
							&lBytesReturned, NULL);
	CloseHandle(hDevice);

	if (!res)
	{
		if ( isCBCMode( eBuf.EncryptionMode ) )
			return Alg_Encrypt_V7_CBC( AlgServiceName, KeyHandle,  IVector, Buffer, Length );

		return FALSE;
	}

	return TRUE;
}

BOOL Alg_Decrypt(
		LPCSTR AlgServiceName,
		DWORD  KeyHandle, 
		DWORD  encModeId,
		BYTE   *IVector, 
		BYTE   *Buffer, 
		DWORD  Length
	)
{
	assert( KeyHandle != NULL );
	if ( KeyHandle == NULL )
		return false;

	char           algService[64];
	BOOL           res = TRUE;
	DWORD          lBytesReturned;
	HANDLE         hDevice;

	strcpy_s(algService, "\\\\.\\");
	strcat_s(algService, AlgServiceName );

	ENCRYPT_BUFFER_V8 eBuf;
	memset( &eBuf, 0, sizeof( eBuf ) );
	eBuf.Size = sizeof( eBuf );

	eBuf.EncryptOperation = false;

	eBuf.EncryptionMode = encModeId;

	memcpy( eBuf.Signature, "LOCOS97 ", maxLen( eBuf.Signature ) );

	eBuf.KeyHandle = KeyHandle;
	memcpy( eBuf.IVector, IVector, 8);
	eBuf.Length    = Length;

	hDevice = CreateFileA( algService,
							GENERIC_READ | GENERIC_WRITE, 0, NULL,
							OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL );

	if (hDevice == INVALID_HANDLE_VALUE) return FALSE;

	res = DeviceIoControl( hDevice, IOCTL_PRIVATE_ENCRYPT_DECRYPT_V8,
							&eBuf, sizeof( eBuf ),
							(LPVOID)Buffer, Length,
							&lBytesReturned, NULL);
	CloseHandle(hDevice);

	if (!res) 
	{
		if ( isCBCMode( eBuf.EncryptionMode ) )
			return Alg_Decrypt_V7_CBC( AlgServiceName, KeyHandle,  IVector, Buffer, Length );

		return FALSE;
	}

	return TRUE;
}


bool Alg_SendCommandToAlgorithmDriver(
	LPCSTR AlgServiceName,
	LPVOID InBuffer, DWORD InBufferSize, 
	LPVOID OutBuffer, DWORD OutBufferSize, 
	DWORD CtlCode, 
	bool CheckSignature )
{
	char algService[MAX_PATH];

	strcpy_s(algService, "\\\\.\\");
	strcat_s(algService, AlgServiceName );

	HANDLE         hDevice;
	DWORD          lBytesReturned = 0;
	PBUFFER_HEADER bufHead = (PBUFFER_HEADER)InBuffer;


	hDevice = CreateFileA( algService, GENERIC_READ | GENERIC_WRITE, 
							0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL );
	if (!hDevice) 
	{
		return false;
	}

	memcpy( bufHead->Signature, "LOCOS97 ", maxLen( bufHead->Signature ) );
	if (!DeviceIoControl(
				hDevice, CtlCode,
				InBuffer, InBufferSize,
				OutBuffer, OutBufferSize,
				&lBytesReturned, NULL)
		)
	{
		CloseHandle(hDevice);
		return false;
	}

	CloseHandle(hDevice);

	if (!CheckSignature) return true;

	if ( strncmp(bufHead->Signature,"LS06CXX ",8) != 0 ) 
	{
		return false;
	}

	if ( bufHead->Status != 0 ) 
	{
		return false;
	}

	return true;
}

bool Alg_SupportLRWMode( LPCSTR AlgServiceName )
{
// NOT_EXISTING_KEY_HANDLE... can be sent in
// IOCTL_PRIVATE_TEST_KEY_HANDLE. If result shows
// that the Key Handle is valid, hence,
// the Algorithm Driver supports LRW encryption mode.
//#define NOT_EXISTING_KEY_HANDLE_V8_SPECIFIC 0xABCD0123

	TESTKEY_BUFFER tBuf;

	memset( &tBuf, 0, sizeof( tBuf ) );

	tBuf.KeyHandle   = NOT_EXISTING_KEY_HANDLE_V8_SPECIFIC;
	tBuf.IsValid     = 0;

	if ( !Alg_SendCommandToAlgorithmDriver(
			AlgServiceName,
			(LPVOID)(&tBuf), sizeof(TESTKEY_BUFFER),
			(LPVOID)(&tBuf), sizeof(TESTKEY_BUFFER),
			IOCTL_PRIVATE_TEST_KEY_HANDLE,
			true ) )
	{
		return false;
	}

	return ( tBuf.IsValid == 1 );
}

#define ALGORITHMS_REGISTRY_PATH_S (TEXT("SOFTWARE\\Jetico\\BestCrypt\\Algorithms"))

//
//	readAlgsFromRegistry
//
static bool readAlgsFromRegistry()
{
	HKEY  hKey;
	DWORD length;
	FILETIME ftLastWriteTime;
	LONG res;

	if (RegOpenKeyEx( HKEY_LOCAL_MACHINE, ALGORITHMS_REGISTRY_PATH_S, 0, KEY_READ, &hKey ) != ERROR_SUCCESS)
		return false; 

	DWORD i=0;

	cSymAlgorithm alg;

	for (;;i++)
	{
		memset( &alg, 0, sizeof( alg ) );

		length = sizeof( alg.m_algServiceName ); 
		res = RegEnumKeyExA( hKey, i, alg.m_algServiceName, &length, NULL, NULL, NULL, &ftLastWriteTime );

		if (res == ERROR_NO_MORE_ITEMS)
			break;

		if (res != ERROR_SUCCESS)
			continue;

		HKEY hSubKey;
		res = RegOpenKeyExA( hKey, alg.m_algServiceName, 0, KEY_READ, &hSubKey );

		if ( res != ERROR_SUCCESS )
			continue;

		// Checks and gets algorithm ID

		DWORD buffLength = sizeof( alg.m_algId );
		DWORD dwKeyType = REG_DWORD_LITTLE_ENDIAN;

		res = RegQueryValueExA( hSubKey, "AlgorithmID", NULL, &dwKeyType, (LPBYTE)&(alg.m_algId), &buffLength );

		if ( res == ERROR_SUCCESS )
		{
			dwKeyType = REG_DWORD_LITTLE_ENDIAN;
			buffLength = sizeof(DWORD);
			DWORD dwOldType;
			res = RegQueryValueExA( hSubKey, "OldType", NULL, &dwKeyType, (LPBYTE)(&dwOldType), &buffLength );

			if ( res != ERROR_SUCCESS )
				dwOldType = 0;

			alg.m_oldType = dwOldType != 0 ;

			dwKeyType = REG_DWORD_LITTLE_ENDIAN;
			buffLength = sizeof(DWORD);
			res = RegQueryValueExA( hSubKey, "KeyLength", NULL, &dwKeyType, (LPBYTE)(&alg.m_keyLengthInBites), &buffLength );

			dwKeyType = REG_DWORD_LITTLE_ENDIAN;
			buffLength = sizeof(DWORD);
			res = RegQueryValueExA( hSubKey, "KeyLength_LRW", NULL, &dwKeyType, (LPBYTE)(&alg.m_dwLengthMS), &buffLength );
			if ( res != ERROR_SUCCESS )
				alg.m_dwLengthMS = 0;

			dwKeyType = REG_DWORD_LITTLE_ENDIAN;
			buffLength = sizeof(DWORD);
			res = RegQueryValueExA( hSubKey, "BlockSize", NULL, &dwKeyType, (LPBYTE)(&alg.m_algBlockSize), &buffLength );

			if ( res != ERROR_SUCCESS )
				alg.m_algBlockSize = 0;

			if (  alg.m_algBlockSize == 0 )
			{
				switch( alg.m_algId )
				{
					case 0xfe :
					case 0x80 : alg.m_algBlockSize =  8; break;
					case 0xA0 : alg.m_algBlockSize = 16; break;
					default :
						assert( false );
						alg.m_algBlockSize = 32;
						break;
				}
			}

			dwKeyType = REG_SZ;
			buffLength = sizeof( alg.m_dispName );
			RegQueryValueExA( hSubKey, "DisplayName", NULL, &dwKeyType, (LPBYTE)(alg.m_dispName), &buffLength );
		}

		RegCloseKey( hSubKey );

		symAlgs.Add( &alg );
	}
	
	RegCloseKey( hKey );

	return true;
}



bool Alg_GetModeInfo( LPCSTR AlgServiceName, int modeId, DWORD *pKeyBitLength )
{
	// IOCTL_PRIVATE_GET_MODE_INFO
	MODE_INFO_BUFFER bf;
	memset( &bf, 0, sizeof( bf ) );

	bf.Mode = modeId;
	if ( bf.Mode == 0 )
		bf.Mode = EX_ENCRYPT_CBC_MODE_FLAG;

	if ( !Alg_SendCommandToAlgorithmDriver( 
			AlgServiceName,
			(LPVOID)(&bf), sizeof(MODE_INFO_BUFFER),
			(LPVOID)(&bf), sizeof(MODE_INFO_BUFFER),
			IOCTL_PRIVATE_GET_MODE_INFO,
			true )
		)
	{
		return false;
	}

	if ( pKeyBitLength != NULL )
		*pKeyBitLength = bf.KeyLength; // / 8;

	return true;
}

bool Alg_GetAlgInfoByID(
		DWORD algID,
		DWORD encModeId,
		char *algServiceNameBuff,
		int algServiceNameBuffSize,
		DWORD *pKeyLength,
		DWORD *pAlgBlockSize,
		char *dispName,
		int dispNameSize )
{
	//ODSFunc;

	if ( symAlgs.m_Q == 0 )
		readAlgsFromRegistry();

	sortAlgorithms();

	for ( int i = symAlgs.m_Q; i--; )
	{
		cSymAlgInfo *alg = &(symAlgs.m_T[i]);

		if ( alg->m_algId == algID )
		{
			DWORD dwLength = alg->m_keyLengthInBites;

			//
			//	New algorithms support the function.
			//
			if ( !Alg_GetModeInfo( alg->m_algServiceName, encModeId, &(alg->m_dwLengthMS ) ) )
			{
				if ( ( encModeId != 0 ) &&
					 ( encModeId != EX_ENCRYPT_LRW_MODE_FLAG ) &&
					 ( encModeId != EX_ENCRYPT_CBC_MODE_FLAG )
					)
					return false;

				if ( encModeId == EX_ENCRYPT_LRW_MODE_FLAG )
				{
					if ( alg->m_dwLengthMS == 0 )
						return false;

					dwLength = alg->m_dwLengthMS;
				}
			}
			else
				dwLength = alg->m_dwLengthMS;

			//if ( encModeId == EX_ENCRYPT_LRW_MODE_FLAG )
			//{
			//	if ( alg->m_dwLengthMS == 0 )
			//		return false;

			//	dwLength = alg->m_dwLengthMS;
			//}

			if ( algServiceNameBuff != NULL )
				strcpy_s( algServiceNameBuff, algServiceNameBuffSize, alg->m_algServiceName );

			if ( pKeyLength != NULL )
				*pKeyLength = dwLength;

			if ( pAlgBlockSize != NULL )
				*pAlgBlockSize = alg->m_algBlockSize;

			if ( dispName != NULL )
				strcpy_s( dispName, dispNameSize, alg->m_dispName );

			return true;
		}
	}

	return false;
}

bool Alg_EnumAlgID(
		DWORD *algIdT,
		int	  *pAlgQ
	)
{
	if ( symAlgs.m_Q == 0 )
		readAlgsFromRegistry();

	sortAlgorithms();

	if ( pAlgQ == NULL )
	{
		assert( false );
		return false;
	}

	if ( symAlgs.m_Q == 0 )
	{
		assert( false );
		return false;
	}

	if ( *pAlgQ == 0 )
	{
		*pAlgQ = symAlgs.m_Q;
		return true;
	}

	if ( algIdT == NULL )
	{
		assert( false );
		return false;
	}

	for ( int i = *pAlgQ; i--; )
	{
		if ( i < symAlgs.m_Q )
			algIdT[i] = symAlgs.m_T[i].m_algId;
		else
			algIdT[i] = 0;
	}

	return symAlgs.m_Q > 0;
}

//////////////////////////////////////////////////
//
//	class cSymAlgorithm
//
cSymAlgorithm::cSymAlgorithm()
{
	m_keyHandle = NULL;
	memset( this, 0, sizeof( *this ) );
	m_algId = -1;
	m_keyLengthInBites = -1;
}

void cSymAlgorithm::Free()
{
	if ( m_keyHandle != NULL )
	{
		Alg_FreeKeyHandler( m_algServiceName, m_keyHandle );
	}

	memset( this, 0, sizeof( *this ) );
}

bool cSymAlgorithm::Init( DWORD algId, DWORD encModeId )
{
	if ( m_keyHandle != NULL )
	{
		Alg_FreeKeyHandler( m_algServiceName, m_keyHandle );
	}

	memset( this, 0, sizeof( *this ) );

	m_algId = algId;
	m_encModeId = encModeId;

	return Alg_GetAlgInfoByID( m_algId, m_encModeId, m_algServiceName, sizeof( m_algServiceName ), &m_keyLengthInBites, &m_algBlockSize, m_dispName, sizeof( m_dispName ) );
}

cEnumAlgIDs::cEnumAlgIDs( DWORD dwFirst, bool bForward )
{
	T = NULL;
	Q = 0;

	if ( !Alg_EnumAlgID( NULL, &Q )
		|| ( Q == 0 )
		)
			return;

	T = new DWORD[Q];

	if ( ( T == NULL )
		|| !Alg_EnumAlgID( T, &Q )
		)
			Q = 0;

	setFirst( dwFirst, bForward );
}

int cEnumAlgIDs::getAlgId( const char *algName )
{
	// v.2.00.3.2 - 11.02.2015
	if ( algName == NULL )
	{
		ODS_fls( L"Assertion! algName == NULL" );
		return -1;
	}

	for ( int i = 0; i < Q; i++ )
	{
		cSymAlgorithm cAlg;
		if ( !cAlg.Init( T[i], 0 ) )
			continue;

		if ( strcmp( cAlg.m_dispName, algName ) == 0 )
			return T[i];
	}

	return -1;
}

cEnumAlgIDs::~cEnumAlgIDs()
{
	if ( T != NULL )
		delete T;

	T = NULL;
	Q = 0;
}

cEnumEncModes::cEnumEncModes( DWORD dwFirst )
{
	static sEncMode stT[] = {
		IDS_encMode_XTS, EX_ENCRYPT_XTS_MODE_FLAG,
		IDS_encMode_LRW, EX_ENCRYPT_LRW_MODE_FLAG,
        IDS_encMode_CBC, EX_ENCRYPT_CBC_MODE_FLAG,
	};

	Q = maxLen( stT );
	T = new sEncMode[ Q ];
	if ( T == NULL )
	{
		Q = 0;
		return;
	}
	
	for( int i = 0; i < Q; i++ )
		T[i] = stT[i];

	if ( dwFirst == (DWORD)-1 )
		return;

	for( int i = 1; i < Q; i++ )
	{
		if ( T[i].id == dwFirst )
		{
			int ids = T[i].ids;
			T[i] = T[0];
			T[0].id = dwFirst;
			T[0].ids = ids;
			return;
		}
	}
};

int cEnumEncModes::getAlgId( const wchar_t *algName, HINSTANCE hInst )
{
	for ( int i = 0; i < Q; i++ )
	{
		wchar_t name[ 128 ];
		if ( LoadStringT( hInst, T[i].ids, name, maxLen( name ) ) == 0 )
			continue;

		if ( wcscmp( name, algName ) == 0 )
			return T[i].id;
	}

	return -1;
}

cEnumEncModes::~cEnumEncModes()
{
	if ( T != NULL )
		delete T;

	T = NULL;
	Q = 0;
}

cEnumHashAlgIDs::~cEnumHashAlgIDs()
{
	if ( T != NULL )
		delete T;

	T = NULL;
	Q = 0;
}

cEnumHashAlgIDs::cEnumHashAlgIDs( DWORD dwFirst, bool bForward )
{
	T = NULL;
	Q = 0;

	DWORD *crtT;	// Table with IDs
	int crtQ;		// IDs quantity

	if ( ha_getIDsOfAvailableAlgAlgorithms( &crtQ, (int **)&crtT ) != sae_ok )
		return;

	T = new DWORD[ crtQ ];

	if ( T == NULL )
	{ ODS_fls( L"Assertion!" ); return; }

	for( int i = 0; i < crtQ; i++ )
	{
		int size;

		if ( ha_getDigestSize( &size, crtT[i] ) != hae_ok )
		{ ODS_fls( L"Assertion!" ); continue; }

		//
		//	29.08.2014 - BC v.9 supports SHA512, KG-Ghost v.2.00
		//
		if ( size > MAXIMUM_DIGEST_SIZE_BYTES_64 )
		//if ( size > MAXIMUM_DIGEST_SIZE_BYTES )
		{
			ODS_fls( L"Warning! Digest size is too large." );
			continue;
		}

		T[Q++] = crtT[i];
	}

	setFirst( dwFirst, bForward );
}

int cEnumHashAlgIDs::getAlgId( const char *algName )
{
	ODSFunc;

	// v.2.00.3.2 - 11.02.2015
	if ( algName == NULL )
	{
		ODS_fls( L"Assertion! algName == NULL" );
		return -1;
	}

	for ( int i = 0; i < Q; i++ )
	{
		const char *name = ha_getHashAlgName( (BYTE)T[i] );

		if ( name == NULL )
			continue;

		if ( ( name == algName ) ||
			 ( strcmp( name, algName ) == 0 ) )
				return T[i];
	}

	return -1;
}

bool cSymAlgorithm::createKey( KEY_HANDLE *pKeyHandle, DWORD algId, DWORD encModeId, BYTE *keyBuff, int keyBuffSize, BYTE *pool )
{
	cSymAlgorithm cAlg;
	if ( !cAlg.createKey( algId, encModeId, keyBuff, keyBuffSize, pool ) )
		return false;

	cAlg.delegateHandle( pKeyHandle );

	return true;
}

bool cSymAlgorithm::createKey( DWORD algId, DWORD encModeId, BYTE *keyBuff, int keyBuffSize, BYTE *pool )
{
//	assert( pool != NULL );
	ODS_fls( _FormatT( L"algId 0x%X, encModeId 0x%X, *keyBuff 0x%X, keyBuffSize 0x%X, *pool 0x%X", algId, encModeId, *(DWORD*)keyBuff, keyBuffSize, pool != NULL ? *(DWORD*)pool : 0 ) );

	BYTE zeroPool[ DB_PoolSize ];

	if ( pool == NULL )
	{
		memset( zeroPool, 0, sizeof( zeroPool ) );
		pool = zeroPool;
	}

	if ( !Init( algId, encModeId ) )
	{
		//ODSE( L"cPwdAlgorithm::createKey - error in initialisation" );
		return false;
	}

	int keyLengthBytes = getByteSize( m_keyLengthInBites );

	if ( keyLengthBytes > keyBuffSize )
		return false;
	
	BOOL ok = Alg_GetKeyHandler( m_algServiceName, encModeId, keyBuff, m_keyLengthInBites, pool, &m_keyHandle );
	return ok == TRUE;

}

const wchar_t *cSymAlgorithm::getAlgName( wchar_t *bf, int bfLen )
{
	return FormatT( bf, bfLen, L"%s (%s)", getWCStr( m_dispName ), getModeName() );
}

//static
const wchar_t *cSymAlgorithm::getModeName( int encMode )
{
	switch ( encMode )
	{
		case EX_ENCRYPT_LRW_MODE_FLAG:
			return L"LRW";
		case EX_ENCRYPT_XTS_MODE_FLAG:
			return L"XTS";
		case 0:
		case EX_ENCRYPT_CBC_MODE_FLAG:
			return L"CBC";

		default:
			assert( false );
			return L"XXX";
	}
}

//static 
bool cSymAlgorithm::encrypt( BYTE *iv, BYTE *bfOut, const BYTE *bfIn, int bfSize, KEY_HANDLE keyHandle, DWORD algId, DWORD encModeId )
{
	char algServiceName[ 128 ];
	DWORD keyLengthInBites;
	DWORD algBlockSize;

	if ( !Alg_GetAlgInfoByID( algId, encModeId, algServiceName, sizeof( algServiceName ), &keyLengthInBites, &algBlockSize ) )
		return false;

	if ( ( bfSize % algBlockSize ) != 0 )
		return false;

	if ( bfOut != bfIn )
		memcpy( bfOut, bfIn, bfSize );

	return TRUE == Alg_Encrypt( algServiceName, keyHandle, encModeId, iv, bfOut, bfSize );
}

bool cSymAlgorithm::encrypt(  BYTE *iv, BYTE *bfOut, const BYTE *bfIn, int bfSize )
{
	if ( m_initError )
		return false;

	if ( ( bfSize % m_algBlockSize ) != 0 )
	{
		assert( false );
		return false;
	}

	if ( bfOut != bfIn )
		memcpy( bfOut, bfIn, bfSize );

	return TRUE == Alg_Encrypt( m_algServiceName, m_keyHandle, m_encModeId, iv, bfOut, bfSize );
}

//static 
bool cSymAlgorithm::decrypt( BYTE *iv, BYTE *bfOut, const BYTE *bfIn, int bfSize, KEY_HANDLE keyHandle, DWORD algId, DWORD encModeId )
{
	//ODSFunc;

	char algServiceName[ 128 ];
	DWORD keyLengthInBites;
	DWORD algBlockSize;

	if ( !Alg_GetAlgInfoByID( algId, encModeId, algServiceName, sizeof( algServiceName ), &keyLengthInBites, &algBlockSize ) )
		return false;

	if ( ( bfSize % algBlockSize ) != 0 )
		return false;

	BYTE IV[ IVECTOR_LENGTH_DB ];

	memcpy( IV, iv, sizeof( IV ) );
	iv = IV;

	if ( bfOut != bfIn )
	{
		memcpy( bfOut, bfIn, bfSize );
	}

	bool ok = Alg_Decrypt( algServiceName, keyHandle, encModeId, iv, (BYTE *)bfOut, bfSize ) == TRUE;

//#ifdef _ODS
//	if ( ( algId == 240 ) && ( encModeId == 0xBC000001 ) )
//	{
//		ODSF( L"Alg_Decrypt( %s, 0x%X, 0x%X, 0x%016I64X%016I64X, 0x%016I64X, %d )",
//							getWCStr( algServiceName ), keyHandle, encModeId, *((__int64 *)iv + 1),*((__int64 *)iv), *((__int64 *)bfOut), bfSize );
//
//		BYTE vMajor = 0, vMinor = 0;
//
//		Alg_VerifyDriver( algServiceName, &vMajor, &vMinor );
//
//		ODSF( L"Test driver %d.%d", (int)vMajor, (int)vMinor );
//
//		cADBlock < BYTE > adbf( bfSize );
//		for ( int i = 0; i <= sizeof( IV ); i++ )
//		{
//			memcpy( adbf.m_T, bfIn, adbf.m_Q );
// 			memset( IV + sizeof( IV ) - i, 0, i );
//			Alg_Decrypt( algServiceName, keyHandle, encModeId, IV, adbf.m_T, adbf.m_Q );
//
//			ODSF( L"%2d. ( 0x%016I64X%016I64X, 0x%016I64X, %d )",
//								 i, *((__int64 *)IV + 1 ),*((__int64 *)IV ), *((__int64 *)adbf.m_T), adbf.m_Q );
//		}
//	}
//#endif
	return ok;
}

bool cSymAlgorithm::decrypt(  BYTE *iv, BYTE *bfOut, const BYTE *bfIn, int bfSize )
{
	if ( m_initError )
		return false;

	if ( ( bfSize % m_algBlockSize ) != 0 )
		return false;

	if ( bfOut != bfIn )
		memcpy( bfOut, bfIn, bfSize );

	return TRUE == Alg_Decrypt( m_algServiceName, m_keyHandle, m_encModeId, iv, bfOut, bfSize );
}

bool cSymAlgorithm::supportLRWMode()
{
	if ( m_initError )
		return false;

	return Alg_SupportLRWMode( m_algServiceName );
}

bool cSymAlgorithm::supportMode( DWORD mode )
{
	if ( m_initError )
		return false;

	if ( Alg_GetModeInfo( m_algServiceName, mode ) )
		return true;

	if ( mode == EX_ENCRYPT_LRW_MODE_FLAG )
		return Alg_SupportLRWMode( m_algServiceName );

	if ( mode == EX_ENCRYPT_CBC_MODE_FLAG )
		return true;

	return false;
}


void cSymAlgorithm::delegateHandle( KEY_HANDLE *pKeyHandle )
{
	if ( pKeyHandle == NULL )
		return;

	*pKeyHandle = m_keyHandle;
	m_keyHandle = NULL;
}

cPwdAlgorithm::cPwdAlgorithm( DWORD algID, DWORD encModeId, cPassword *pwd, DWORD hashAlgId, BYTE *salt, int saltSize, int iteration )
{
	m_err = true;
	m_pwdKey = NULL;

	if ( algID != -1 )
		generateKey( algID, encModeId, pwd, hashAlgId, salt, saltSize, iteration );
}

int Alg_GetKeySize( DWORD algID, DWORD encModeId )
{
	// Generates key from password
	DWORD keyLengthInBites = -1;

	if ( !Alg_GetAlgInfoByID( algID, encModeId, NULL, 0, &keyLengthInBites ) )
	{
		//ODSE( L"cPwdAlgorithm::cPwdAlgorithm - Invalid ID" );
		return 0;
	}

	return ( keyLengthInBites + 7 ) / 8;
}

bool cPwdKeyV5::getKey( int keySize, cPassword *pwd, DWORD hashAlgId, BYTE *salt, int saltSize, int iterations )
{
	if ( saltSize == 8 )
		return false;
	
	if ( isSameParams( keySize, pwd, hashAlgId, salt, saltSize, iterations ) )
		return true;

	return generateKey( keySize, pwd, hashAlgId, salt, saltSize, iterations );
}

bool cPwdKeyV5::getPwdDigest( cPassword *pwd, DWORD hashAlgId, BYTE *digest )
{
	if ( pwd->GenerateUnicodeString() == 0 )
		return false;

	int pwdLen = wcslen( (wchar_t *)( pwd->unicodeBuffer ) );
	if ( pwdLen == 0 )
		return false;

	if ( ha_makeDigestEx( hashAlgId, digest, pwd->unicodeBuffer, pwdLen * sizeof(wchar_t) ) != hae_ok )
		return false;

	return true;
}

bool cPwdKeyV5::generateKey( int keySize, cPassword *pwd, DWORD hashAlgId, BYTE *salt, int saltSize, int iterations )
{
	Reset();

	int digestSize = 0;
	if ( ha_getDigestSize( &digestSize, hashAlgId ) != hae_ok )
		return false;

	int size = keySize + digestSize + saltSize;

	m_key = m_ptb.Allocate( size );

	if ( m_key == NULL )
		return false;

	if ( (BYTE)hashAlgId != hashAlgId )
	{
		assert( false );
		return false;
	}

	if ( !pwd->generateKeyFromPassword( m_key, keySize, (BYTE)hashAlgId, salt, saltSize, iterations, true ) )
		return false;

	m_keySize = keySize;
	m_hashAlgId = hashAlgId;
	m_iterations = iterations;

	if ( !getPwdDigest( pwd, hashAlgId, m_key + keySize ) )
		return false;

	memcpy( m_key + keySize + digestSize, salt, saltSize );

	return true;
}

bool cPwdKeyV5::isSameParams( int keySize, cPassword *pwd, DWORD hashAlgId, BYTE *salt, int saltSize, int iterations )
{
	if ( m_key == NULL )	return false;
	if ( m_keySize != keySize ) return false;
	if ( m_hashAlgId != hashAlgId ) return false;
	if ( m_iterations != iterations ) return false;

	int digestSize = 0;
	if ( ha_getDigestSize( &digestSize, hashAlgId ) != hae_ok )
		return false;

	if ( m_ptb.bufferSize != keySize + digestSize + saltSize ) return false;

	cPTB ptb;
	if ( ptb.Allocate( digestSize ) == NULL ) return false;

	if ( !getPwdDigest( pwd, hashAlgId, ptb.buffer ) ) return false;

	if ( memcmp( ptb.buffer, m_key + keySize, digestSize ) != 0 ) return false;
	if ( memcmp( salt, m_key + keySize + digestSize, saltSize ) != 0 ) return false;

	return true;
}


bool cPwdAlgorithm::generateKey( DWORD algID, DWORD encModeId, cPassword *pwd, DWORD hashAlgId, BYTE *salt, int saltSize, int iteration )
{
	ODSFuncS( _FormatT( L"algID 0x%X, encModeId 0x%X, pwd '%s', len %d, hashAlgId  0x%X, salt  0x%X, saltSize 0x%X, iteration 0x%X",
		algID, encModeId, cWCStr( pwd->buffer ).m_szBody, (DWORD)strlen( pwd->buffer ), hashAlgId, *(DWORD *)salt, saltSize, iteration ) );

	m_err = true;
	Free();

	int pwdKeySize = Alg_GetKeySize( algID, encModeId );

	if ( pwdKeySize <= 0 )
		return false;
	
	cPlainTextBlock ptb;

	BYTE *pwdKey = NULL;
	
	if ( ( m_pwdKey != NULL )&& m_pwdKey->getKey( pwdKeySize, pwd, hashAlgId, salt, saltSize, iteration ) )
	{
		pwdKey = m_pwdKey->m_key;
		pwdKeySize = m_pwdKey->m_keySize;
	}
	else
	{
		pwdKey = ptb.Allocate( pwdKeySize );

		if ( pwdKey == NULL )
			return false;

		if ( !pwd->generateKeyFromPassword( pwdKey, pwdKeySize, (BYTE)hashAlgId, salt, saltSize, iteration, true  ) )
		{
			//ODSE( L"cPwdAlgorithm::cPwdAlgorithm - error in generation" );
			return false;
		}
	}

	if ( !createKey( algID, encModeId, pwdKey, pwdKeySize, NULL ) )
		return false;

	m_err = false;

	return true;
}
