#include "UnicodeFuncs.h"
#include "cKeyBlockParams.h"
#include "..\..\..\CertificateLib\publicKeyAlgs\publicKeyAlgs.h"
#include "..\..\..\CertificateLib\PGPPkts\Packets.h"
#include "..\..\..\CertificateLib\PGPPkts\PGPPkts.h"
#include "..\..\..\CertificateLib\x509pkts\pkcs12_funcs.h"
#include "..\..\..\CertificateLib\RNG\cRNG.h"
#include "FileView.h"

//#include "COMMANDS.H"
#include "resource.h"

//
// Encoded block looks like:
//
//	 8 bytes - pgp key ID, if certificate was used as public key the field is filled by random
//	20 bytes - certificate public key local key id
//	 x bytes - public key encrypted key block, note the block has been additional coverd to hide MPI info.
//				the x size value copletely determinated by valid public key ( it's encrypted buffer size ).
//
static int getEncBlockSize_Public( HPKA pka, int *pBlockSize, int *pMaxSize = NULL )
{
	int blockSize;
	blockSize = pka_getBlockSize( pka );

	//
	// RSA_PublicKeyEncryptionBlock_Format generates error
	// if dataSize > outBufSize - 10
	//
	// if ( keyLength > blockSize - 10 )
	//	return -1;
	//
	if ( pMaxSize )
		*pMaxSize = blockSize - 10;

	int encBuffSize = 0;
	int res = pka_encrypt( pka, NULL, &encBuffSize, NULL, blockSize );

	if ( res != pkae_bufferSizeTooSmall )
		return -1;
	
	if ( pBlockSize )
		*pBlockSize = blockSize;

	return encBuffSize;
}

/*
static int getEncBlockSize_Secret( int keyLength, HPKA ska, int *pBlockSize )
{
	switch ( ska_getAlgID( ska ) )
	{
		//
		//	The secret key algorithm classes are based on theirs public key classes.
		//	So the trick allows to save time.
		//
		case pgppkaRSA :
		case pgppkaRSA_e:
		case pgppkaElgamal:
		case pgppkaElgamal_e:
			return getEncBlockSize_Public( keyLength, ska, pBlockSize );

		default:
			break;
	}

	HPKA pka_tmp;
	if ( ska_derivePublicKeyAlgorithm( &pka_tmp, ska ) != pkae_ok )
		return -1;

	int size = getEncBlockSize_Public( keyLength, pka_tmp, pBlockSize );

	pka_free( pka_tmp );

	return size;
}
*/

static int _encrypt( HPKA pka, BYTE *encPos, int *pEncSize, BYTE *buffer, int blockSize )
{
	int res = pka_encrypt( pka, encPos, pEncSize, buffer, blockSize );
	
	if ( res == pkae_randomBufferRequired )
	{
		cPlainTextBlock pool;
		if ( pool.Allocate( blockSize ) == NULL )
			return false;

		ShredData( pool.buffer, pool.bufferSize );
		
		cRNG::setPool( pool.buffer, pool.bufferSize );

		res = pka_encrypt( pka, encPos, pEncSize, buffer, blockSize );

		cRNG::clearPool();
	}

	return res;
}

//
//	The encrypted public key buffer is really MPI buffer.
//	The first MPI bytes with size are visible. So we have to hide them in key block.
//	The idea is : the MPI buffer will be xorred with local key ID generated by MD5 instead of SHA1
//
static bool coverBuffer( HPKA pka, BYTE *buff, int buffSize )
{
	// Generates MD5 hash instaed of usual SHA1
	int hashSize = 0;
	int res = pka_generateLocalKeyID( pka, NULL, &hashSize, pgphaMD5 );

	if ( res != x509e_bufferTooSmall )
		return false;

	cPlainTextBlock ptb;
	if ( ptb.Allocate( hashSize ) == NULL )
		return false;

	res = pka_generateLocalKeyID( pka, ptb.buffer, &hashSize, pgphaMD5 );

	if ( res != pkae_ok )
		return false;

	BYTE *cover = ptb.buffer;
	int coverSize = hashSize;

	for( int i = 0; i < buffSize; i++ )
		buff[i] ^= (BYTE)i ^ cover[ i % coverSize ];

	return true;
}

static const int localKeyIDSize = 20; // SHA1 digest size.

//
//	decodeKey
//		1. Checks buffers sizes
//		2. Discover the covered MPI buffer,
//		3. Decrypt MPI buffer by secret key
//		4. Parse formatted RSA block
//		5. Copy data from fromatted block to key buffer
//
bool cParamsPKE::decodeKey( BYTE *keyBuff, int keyBuffSize, BYTE *kbEncrypted, int kbEncryptedSize )
{
	if ( ska == NULL )
		return false;

	// pka has to be derived at secret key is opened
	if ( pka == NULL )
		return false;

	int blockSize;

	int encSize = getEncBlockSize_Public( pka, &blockSize );

	if ( encSize == -1 )
		return false;

	int size = encSize + pgp_keyIDSize + localKeyIDSize;

	if ( size > kbEncryptedSize )
		return false;

	BYTE *encPos = kbEncrypted + pgp_keyIDSize + localKeyIDSize;
	int encTailsize = kbEncryptedSize - ( pgp_keyIDSize + localKeyIDSize ); 

	cPlainTextBlock ptb;
	if ( ptb.WriteBytes( encPos, encTailsize ) == 0 )
		return false;

	if ( !coverBuffer( pka, ptb.buffer, ptb.bufferSize ) )
		return false;

	// The encrypted key size may be more then maximum allowed for the algorithm
	// so plain text key may be splited to several parts.
	// We have to decrype and concatenate all the parts.

	int pos = 0;
	cPlainTextBlock ptbKey;

	do
	{
		cPlainTextBlock ptbDec;
		if ( ptbDec.Allocate( blockSize ) == NULL )
			return false;

		if ( ska_decrypt( ska, ptbDec.buffer, &size, ptb.buffer + pos, encSize ) != pkae_ok )
			return false;

		const BYTE *dataPos;
		int dataSize;

		if ( RSA_PublicKeyEncryptionBlock_Parse( true, ptbDec.buffer, blockSize, &dataPos, &dataSize ) != pkae_ok )
			return false;

		if ( ptbKey.WriteBytes( (BYTE*)dataPos, dataSize ) != dataSize )
			return false;

		if ( ptbKey.bufferSize > keyBuffSize )
			return false;

		sKeyInfo *keyInfo = ( sKeyInfo * )ptbKey.buffer;

		if ( keyInfo->size <= ptbKey.bufferSize )
			break;

		// go to the next encrypted buffer.
		pos += encSize;
	}
	while( true );

	memcpy( keyBuff, ptbKey.buffer, ptbKey.bufferSize );
	return true;
}

//
//	encodeKey
//		1. Checks encrypted and plaintext buffer sizes
//		2. Prepare RSA formated encrypted block
//		3. Fills key ID buffers
//		4. Encrypt the block by public key
//		5. Cover the encrypted MPI by local key ID
//
//	Note, the encrypted information may be splitted to several parts, 
//	because maxExcrypted size may be less then size of encrypted key info
//
bool cParamsPKE::encodeKey( BYTE *kbEncrypted, int kbEncryptedSize, BYTE *keyBuff, int keyBuffSize )
{
	int blockSize;
	int splitSize;

	if ( pka == NULL )
		return false;

	int size = getEncBlockSize_Public( pka, &blockSize, &splitSize );

	if ( size == -1 )
		return false;

	int blockQ = ( keyBuffSize + splitSize - 1 ) / splitSize;

	if ( pgp_keyIDSize + localKeyIDSize + size * blockQ > kbEncryptedSize )
		return false;

	// Fill key ID buffers
	if ( kff == kffPGP )
	{
		assert( ptbKeyID.bufferSize == pgp_keyIDSize );
		memcpy( kbEncrypted, ptbKeyID.buffer, pgp_keyIDSize );

		int keyIDSize = 0;
		int res = pka_generateLocalKeyID( pka, NULL, &keyIDSize );

		if ( res != x509e_bufferTooSmall )
			return false;

		assert( keyIDSize == localKeyIDSize );

		keyIDSize = localKeyIDSize;

		res = pka_generateLocalKeyID( pka, kbEncrypted + pgp_keyIDSize, &keyIDSize );

		if ( res != pkae_ok )
			return false;
	}
	else // x509
	{
		// The part has to be randomized early.

//		srand( time( 0 ) ^ clock() );
//		for( int i = 0; i < pgp_keyIDSize; i++ )
//			kbEncrypted[i] = (BYTE)rand();

		if ( localKeyIDSize != ptbKeyID.bufferSize )
		{
			assert( false );
			return false;
		}

		memcpy( kbEncrypted + pgp_keyIDSize, ptbKeyID.buffer, localKeyIDSize );
	}

	int filledSize = pgp_keyIDSize + localKeyIDSize;

	BYTE *encPos = kbEncrypted + pgp_keyIDSize + localKeyIDSize;
	int encSize = size;

	for( int blockN = 0; blockN < blockQ; blockN++ )
	{
		cPlainTextBlock ptb;
		if ( ptb.Allocate( blockSize ) == NULL )
			return false;

		BYTE *curPos = keyBuff + splitSize * blockN;

		//
		//	Version 1.02.8 - 07.04.2008
		//
		//	int curSize = keyBuffSize % splitSize;
		//
		int curSize = keyBuffSize - splitSize * blockN;
		if ( curSize > splitSize )
			curSize = splitSize;

		if ( RSA_PublicKeyEncryptionBlock_Format( true, ptb.buffer, blockSize, curPos, curSize  ) != pkae_ok )
			return false;

		int res = _encrypt( pka, encPos + blockN * encSize, &encSize, ptb.buffer, blockSize );
		
		if ( res != pkae_ok )
			return false;

		filledSize += encSize;
	}

	assert( filledSize <= kbEncryptedSize );

// ??	if ( !coverBuffer( pka, encPos, kbEncryptedSize - ( pgp_keyIDSize + localKeyIDSize ) ) )
	if ( !coverBuffer( pka, encPos, filledSize - ( pgp_keyIDSize + localKeyIDSize ) ) )
		return false;

    return true;
}

int cParamsPKE::getEncodedKeySize( int keyLength )
{
	if ( pka == NULL )
		return -1;

	int maxDecodedSize;
	int encSize = getEncBlockSize_Public( pka, NULL, &maxDecodedSize );

	if ( encSize == -1 )
		return -1;

	int blockQ = ( keyLength + maxDecodedSize - 1 ) / maxDecodedSize;

	return blockQ * encSize + pgp_keyIDSize + localKeyIDSize;
}

cParamsPKE::~cParamsPKE()
{
	if ( pka != NULL )
		pka_free( pka );

	if ( ska != NULL )
		ska_free( ska );

	ska = pka = NULL;
}


static int getKeyIDTbl( cPlainTextBlock *ptbTbl, BYTE *keyPkt, int keyPktSize, bool secret = false )
{
	int size = 0;

	int res = pgp_EnumKeyIDs( NULL, &size, keyPkt, keyPktSize, secret );

	if ( res == ppe_bufferTooSmall )
	{
		if ( ptbTbl->Allocate( size ) == NULL )
			return ppe_memoryAllocation;

		res = pgp_EnumKeyIDs( ptbTbl->buffer, &size, keyPkt, keyPktSize, secret );
	}

	return res;
}

static bool ParseP12BuffAndGetSecretKey( HPKA *pSka, BYTE *buff, int buffSize, cPassword *pPwd, cUniversalString *us = NULL )
{		
	cPlainTextBlock ptbKey;

	cPlainTextBlock ptbAttrSet;

	if ( P12_PFX_PDU_parseEx( buff, buffSize, &ptbAttrSet, pPwd, &ptbKey ) != x509e_ok )
		return false;

	if ( us != NULL )
	{
		P12_attrSet set;
		if ( set.parse( ptbAttrSet.buffer, ptbAttrSet.bufferSize ) == x509e_ok )
		{
			x509_struct s;
			if ( set.getFriendlyName( &s ) == ppe_ok )
			{
				us->SetUnicode( s.B, s.L );
				us->MakeFromUnicodeString();
			}
		}
	}

	x509_privateKeyInfo pki;
	if ( pki.parse( ptbKey.buffer, ptbKey.bufferSize ) != x509e_ok )
		return false;

	if ( ska_createAlgorithm_X509( pSka, &pki ) != x509e_ok )
		return false;

	return true;
}

static wchar_t *getAlgInfo( wchar_t *wStr, int wStrMaxLen, HPKA ska )
{
	BYTE lkId[ localKeyIDSize ];

	int size = sizeof( lkId );

	ska_generateLocalKeyID( ska, lkId, &size );

	wchar_t strID[ 256 ];
	bytes2string( strID, maxLen( strID ), lkId, sizeof( lkId ) );

	FormatT(
		wStr, wStrMaxLen,
		_LoadIDS( IDS_szS_D_BitsID_S ),
		getWCStr( PKT_GetPublicKeyAlgName( ska_getAlgID( ska ) ) ),
		ska_getModuleSize( ska ) * 8,
		strID );

	return wStr;
}

//static char *getAlgInfo( char *str, HPKA ska )
//{
//	BYTE lkId[ localKeyIDSize ];
//
//	int size = sizeof( lkId );
//
//	ska_generateLocalKeyID( ska, lkId, &size );
//
//	char strID[ 256 ];
//	bytes2string( strID, sizeof( strID ), lkId, sizeof( lkId ) );
//
//	sprintf( str, " %s (%d bits) ID %s", PKT_GetPublicKeyAlgName( ska_getAlgID( ska ) ), ska_getModuleSize( ska ) * 8, strID );
//
//	return str;
//}

wchar_t curName[ 1024 ];

enum { arOK, arContinue, arError };

static int askRes( HWND hWnd, HPKA *pSKA, HPKA ska, const wchar_t *newName, bool silent )
{
	// If pSKA is NULL simple report that key with the password is found
	if ( pSKA == NULL )
	{
		wcsncpy( curName, newName, maxLen( curName ) );
		return arOK;
	}

	// First secret key found
	if ( *pSKA == NULL )
	{
		wcsncpy( curName, newName, maxLen( curName ) );
		*pSKA = ska;
		return arContinue;
	}

	// Another secret key found

	// We cannot choose key in silent mode, so report errror.
	if ( silent )
		return arError;

	cWCStr text;
	wchar_t strCur[ 256 ];
	wchar_t strNew[ 256 ];

	text.FormatT(
		_LoadIDS( IDS_OtherSKSamePasswordFound_SSSS_Q ),
		curName, getAlgInfo( strCur, maxLen( strCur ), *pSKA ),
		newName, getAlgInfo( strNew, maxLen( strNew ), ska )
	);

	int res = MessageBoxT( hWnd, text.m_szBody, _LoadIDS( IDS_KeyGen ), MB_ICONQUESTION | MB_YESNOCANCEL );

	if ( res == IDYES )
	{
		ska_free( ska );
		ska = NULL;
		return arOK;
	}
	else
	if ( res == IDNO )
	{
		ska_free( *pSKA );
		*pSKA = ska;
		wcsncpy( curName, newName, maxLen( curName ) );
		return arContinue;
	}
	else
	{
		ska_free( *pSKA );
		*pSKA = NULL;
		ska_free( ska );
		ska = NULL;
		return arError;
	}
}

//
//		Walks all p12 file and implement password
//		if pSKA is NULL simple return true it means that key with same password exists in BCDB
//		if pSKA is not NULL walks all keys and report if another key has same password
//			allows to choose the new and continue searching 
//			or use the previose and stop.
//		Returns
//			- 0 - not found;
//			- 1 - found;
//			- 2 - cancelled;
//
static int EnumP12FileAndGetSecretKey( HWND hWnd, wchar_t *folder, cPassword *pPwd, HPKA *pSKA, bool silent )
{
	if ( ( pSKA != NULL ) && ( *pSKA != NULL ) )
	{
		ska_free( *pSKA );
		*pSKA = NULL;
	}

	wchar_t path[ MAX_PATH ];
	wcscpy_s( path, folder );
	int len = (int)wcslen( path );
	if ( len == 0 )
		return false;
	if ( path[ len - 1 ] != L'\\' )
	{
		path[ len++ ] = L'\\';
		path[ len ] = L'\0';
	}

	wcsncpy( path + len, L"*.p12", MAX_PATH - len );

	WIN32_FIND_DATAW findFileData;

	HANDLE hFind = FindFirstFileT( path, &findFileData );

	if ( hFind == INVALID_HANDLE_VALUE )
		return false;

	cUniversalString friendlyNameNew;

	//bool found = false;
	int iFound = 0;

	do
	{
		wcsncpy( path + len, findFileData.cFileName, MAX_PATH - len );
		cFileView fv( path );

		if ( fv.m_err )
			continue;

		HPKA ska = NULL;

		if ( ParseP12BuffAndGetSecretKey( &ska, fv.m_ptr, (int)fv.m_size.QuadPart, pPwd, &friendlyNameNew ) )
		{
			friendlyNameNew.MakeFromUnicodeString();

			int res = askRes( hWnd, pSKA, ska, friendlyNameNew.wcBuffer, silent );

			if( res == arOK )
			{	iFound = 1; break; }
			else
			if( res == arContinue )
			{	iFound = 1; continue; }
			else // ask error result
			{	iFound = 2;	break;	}
			
		}
	}
	while( FindNextFileT( hFind, &findFileData ) );

	FindClose( hFind );

	return iFound; // *pSKA != NULL;
}

static wchar_t *setName( wchar_t *name, int nameSize, char *bf, int bfSize )
{
	int len = ( bfSize >= nameSize ? nameSize - 1 : bfSize );
	_mbstowcs( name, len, bf, -1, false, CP_ACP );
	name[ len ] = '\0';
	return name;
}

static int EnumPGPKeyRingAndGetSecretKey( HWND hWnd, wchar_t *keyRingFile, cPassword *pPwd, HPKA *pSKA, bool silent )
{
	cFileView fv( keyRingFile );

	if ( fv.m_err )
		return false;

	BYTE *skrBuff = fv.m_ptr;
	int skrBuffSize = (int)fv.m_size.QuadPart;

	cPlainTextBlock ptbTbl;

	int res = getKeyIDTbl( &ptbTbl, skrBuff, skrBuffSize, true );

	if ( res != pkae_ok )
		return false;

	int Q = ptbTbl.bufferSize / pgp_keyIDSize;

	cPlainTextBlock ptbSecKey;

	int iFound = 0;

	for( int i = 0; i < Q; i++ )
	{
		BYTE *keyID = ptbTbl.buffer + i * pgp_keyIDSize;
		
		sSecretKey *sk;
		int skSize = 0;

		BYTE *userID;
		int userIDSize;

		res = pgp_FindKeyByID( (BYTE **)&sk, &skSize, keyID, skrBuff, skrBuffSize, true, NULL, NULL, &userID, &userIDSize  );

		if ( res != pkae_ok )
			return false;

		HPKA ska = NULL;
		res = PKT_CreateSecretKeyAlgFromPacket( &ska, sk, skSize, pPwd->buffer );

		if ( res != ppe_ok )
			continue;

		wchar_t newName[ 1024 ];
		setName( newName, maxLen( newName ), (char *)userID, userIDSize );

		int res = askRes( hWnd, pSKA, ska, newName, silent );

		if( res == arOK )
		{	iFound = 1; break; }
		else
		if( res == arContinue )
		{	iFound = 1; continue; }
		else // ask error result
		{	iFound = IDCANCEL;	break;	}
	}

	return iFound;
}


//
//	FindSecretKeyInBCDB
//		Walks all p12 file and all PGP secret keys in local secret key ring
//		if pSKA is NULL simple return true it means that key with same password exists in BCDB
//		if pSKA is not NULL walks all keys and report if another key has same password
//			Both functions EnumP12FileAndGetSecretKey
//			allows to choose the new and continue searching 
//			or use the previose and stop.
//
static int FindSecretKeyInBCDB( HWND hWnd, cPassword *pPwd, HPKA *pSKA, type_kff *pKff )
{
	wchar_t path[ MAX_PATH ];

	if ( !KM_GetDataBasePath( path, maxLen( path ), KM_GetDataBasePath_Flag_Folder ) )
		return 0;

	int iFound = EnumP12FileAndGetSecretKey( hWnd, path, pPwd, pSKA, false );
	if ( iFound == IDCANCEL )
		return IDCANCEL;
	else
	if ( iFound == IDOK )
	{
		if ( pKff != NULL ) *pKff = kffX509;
		return 1;
	}
	else
	if ( !KM_GetDataBasePath( path, maxLen( path ), KM_GetDataBasePath_Flag_PGPSecretKeyRing ) )
		return 0;

	// path is secret key ring file in BCDB.

	iFound = EnumPGPKeyRingAndGetSecretKey( hWnd, path, pPwd, pSKA, false );

	if ( iFound == IDCANCEL )
		return IDCANCEL;
	else
	if ( iFound == IDOK )
	{
		if ( pKff != NULL ) *pKff = kffPGP;
		return 1;
	}

	return 0;
}

static bool ParseSecretKey( BYTE *buff, int buffSize, type_kff kff, cPassword *pwd, HPKA *pSka )
{
	if ( kff == kffX509 )
	{
		return ParseP12BuffAndGetSecretKey( pSka, buff, buffSize, pwd );
	}
	else
	if ( kff == kffPGP )
	{
		int res = PKT_CreateSecretKeyAlgFromPacket( pSka, ( sSecretKey * )buff, buffSize, pwd->buffer );

		return res == pkae_ok;
	}
	else
		return false;
}

//
//	openSecretKey
//		if keyIDSBuffer is NULL, so it walks all secret key in BCDB and implements the password
//		otherwise it opens secret key with required key ID.
//
bool cParamsPKE::openSecretKey( HWND hWnd, cPassword *pwd, BYTE *keyIDSBuffer, bool checkPwdOnly )
{
	int iFound = 0;

	if ( bCancelled )
		return false;

	if ( CKeyManDll::getStatic( true ).error )
		return false;

	cPlainTextBlock ptb;

	if ( keyIDSBuffer != NULL )
	{
		if ( ska != NULL )
			ska_free( ska );

		if ( pka != NULL )
			pka_free( pka );

		ska = pka = NULL;

		if ( KM_GetPKBlockByID( hWnd, keyIDSBuffer, pgp_keyIDSize, true, &ptb, &kff ) )
		{
			if ( !ParseSecretKey( ptb.buffer, ptb.bufferSize, kff, pwd, &ska ) )
			{
				if ( !checkPwdOnly )
					MessageBoxT( hWnd, _LoadIDS( IDS_INVALID_PWD ), _LoadIDS( IDS_OpenSecretKey ) , MB_ICONSTOP );

				return false;
			}
		}
		else
		if ( KM_GetPKBlockByID( hWnd, keyIDSBuffer + pgp_keyIDSize, localKeyIDSize, true, &ptb, &kff ) )
		{
			if ( !ParseSecretKey( ptb.buffer, ptb.bufferSize, kff, pwd, &ska ) )
			{
				if ( !checkPwdOnly )
					MessageBoxT( hWnd, _LoadIDS( IDS_INVALID_PWD ), _LoadIDS( IDS_OpenSecretKey ) , MB_ICONSTOP );

				return false;
			}
		}
		else
		{
			if ( checkPwdOnly )
				return false;

			const wchar_t *initialSKPath = L""; // = NULL;

			while( true )
			{
				ptb.Free();

				if ( !KM_BrowseForPKBlock( hWnd, initialSKPath, true, &ptb, &kff ) || ( ptb.bufferSize == 0 ) )
					return false;

				if ( ParseSecretKey( ptb.buffer, ptb.bufferSize, kff, pwd, &ska ) )
					break;

				DlgMessageBoxT( IH.m_hInst, hWnd, IDS_INVALID_PWD, MB_ICONSTOP, _LoadIDS( IDS_OpenSecretKey ) );
			}
		}
	}
	else
	if ( ( iFound = FindSecretKeyInBCDB( hWnd, pwd, ( checkPwdOnly ? NULL : &ska ), &kff ) ) != 0 )
	{
		if ( iFound == IDCANCEL )
		{
			bCancelled = true;
			return false;
		}

		if ( checkPwdOnly )
			return true;
	}
	else
	if ( !checkPwdOnly )
	{
		const wchar_t *initialSKPath = L""; // = NULL;

		while( true )
		{
			ptb.Free();

			if ( !KM_BrowseForPKBlock( hWnd, initialSKPath, true, &ptb, &kff ) || ( ptb.bufferSize == 0 ) )
			{
				bCancelled = true;
				return false;
			}

			if ( ParseSecretKey( ptb.buffer, ptb.bufferSize, kff, pwd, &ska ) )
				break;

			DlgMessageBoxT( IH.m_hInst, hWnd, IDS_INVALID_PWD, MB_ICONSTOP, _LoadIDS( IDS_OpenSecretKey ) );
		}
	}
	else
		return false;

	if ( ska_derivePublicKeyAlgorithm( &pka, ska ) != pkae_ok )
		return false;

	return true;
}