/*++

Copyright (c) 1997-1998 Jetico, Inc.

Module Name: BC_ALG.c
Abstract:    This is a common code for all encryption algorithm drivers.
Author:      Sergey A. Frolov
Environment: Kernel mode only.

Notes:

Revision History:

--*/


#include <ntddk.h>
#include <ntiologc.h>
#include <ntdddisk.h>
#include <wdm.h>

//#include <string.h>
#include <Ntstrsafe.h>

#include "bc_ioctl.h"
#include "bc_alg.h"
#include "bcstatus.h"
#include "algorithm.h"
#include "gf_mul.h"
#include "xts_mode.h"

#ifndef ALG_FOR_VC2015
#define MdlMappingNoExecute     0x40000000  // Create the mapping as noexecute
#define NTDDI_WIN8              0x06020000
#endif

// This definition is from wdm.h file from Windows DDK for Windows 8
#define MdlMappingNoExecute     0x40000000  // Create the mapping as noexecute

#if DBG
ULONG   BCDiskDebugLevel = 2;
BOOLEAN shouldBreak	= TRUE;
#else
BOOLEAN shouldBreak	= FALSE;
#endif

//#define NTDDI_WIN8 (0x06020000)

#define NonPagedPoolNx  512
#define BC_ALG_TAG ('glab')

BOOL IsWindows8orAbove = FALSE;

NTSTATUS
RegisterWipingKeyFunction();

/***************************************************************************/

#define TABLE_LENGTH 128

KEY_ELEMENT KeyTable[ TABLE_LENGTH ];

/***************************************************************************/



PVOID MyMmGetSystemAddressForMdlSafe( IN PMDL MDL, IN MM_PAGE_PRIORITY PRIORITY )
{
	PVOID  retAddress = NULL;
	CSHORT mdlFlagAdd = (MDL->MdlFlags & MDL_MAPPING_CAN_FAIL) ^ MDL_MAPPING_CAN_FAIL;

	MDL->MdlFlags |= mdlFlagAdd;

	// 11-Aug-2016. MdlMappingNoExecute flag added for Windows 8+. Driver Verifier encounters problem if the flag is not set.
	if (IsWindows8orAbove)
		retAddress = MmGetSystemAddressForMdlSafe( MDL, PRIORITY | MdlMappingNoExecute );
	else
		retAddress = MmGetSystemAddressForMdlSafe( MDL, PRIORITY );

	MDL->MdlFlags ^= mdlFlagAdd;
	return retAddress;
}


PVOID MyExAllocatePool( IN SIZE_T NumberOfBytes )
{
	if (IsWindows8orAbove)
		return ExAllocatePoolWithTag( NonPagedPoolNx, NumberOfBytes, BC_ALG_TAG );

	return ExAllocatePoolWithTag( NonPagedPool, NumberOfBytes, BC_ALG_TAG );
}


VOID MyExFreePool( IN PVOID P )
{
	ExFreePoolWithTag( P, BC_ALG_TAG );
}


NTSTATUS
DriverEntry(
    IN OUT PDRIVER_OBJECT   DriverObject,
    IN PUNICODE_STRING      RegistryPath
    )
/*++

Routine Description:
    This routine is called by the Operating System to initialize the driver.
    It fills in the dispatch entry points in the driver object.  Then
    BCalgInitializeDisk is called to create the device object and complete
    the initialization.

Arguments:
    DriverObject - a pointer to the object that represents this device
    driver.
    RegistryPath - a pointer to our Services key in the registry.

Return Value:
    STATUS_SUCCESS if this disk is initialized; an error otherwise.
--*/

{
    NTSTATUS         ntStatus;
    USHORT i;

    //if (shouldBreak) 
        //DbgBreakPoint();

    #ifdef ALG_FOR_VC2015
       IsWindows8orAbove = RtlIsNtDdiVersionAvailable( NTDDI_WIN8 );
    #endif

    //
    // Initialize the driver object with this driver's entry points.
    DriverObject->MajorFunction[IRP_MJ_CREATE] = BCalgCreateClose;
    DriverObject->MajorFunction[IRP_MJ_CLOSE]  = BCalgCreateClose;
    DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL] = BCalgDeviceControl;

    DriverObject->DriverUnload = BCalgUnloadDriver;

    // Initialize KeyTable
    for(i=0; i<TABLE_LENGTH; i++) KeyTable[i].keyExtended = NULL;
    
	ntStatus = BCalgInitializeDriver(DriverObject);

    return ntStatus;
}



/***************************************************************************/

NTSTATUS
BCalgInitializeDriver( IN PDRIVER_OBJECT   DriverObject )

/*++
Routine Description:
    This routine is called at initialization time by DriverEntry().

    It creates and initializes a device object.

Arguments:
    DriverObject - a pointer to the object that represents this device driver.
    ParamPath - a pointer to the Parameters key under our key in the Services
        section of the registry.

Return Value:
    STATUS_SUCCESS if this disk is initialized; an error otherwise.
--*/

{
    PDEVICE_OBJECT      deviceObject = NULL;    // ptr to device object
    NTSTATUS            ntStatus;
	STRING			    serviceNameString;
    UNICODE_STRING      deviceNameUString, prefix, linkPrefix, suffix;
    PBCALG_EXTENSION   gostExtension = NULL;


    // Initialize Unicode string for Service Name
    RtlInitString( &serviceNameString, SERVICE_NAME );
    ntStatus = RtlAnsiStringToUnicodeString( &suffix, &serviceNameString, TRUE );
    if (!NT_SUCCESS( ntStatus )) goto BCalgInitializeDiskExit;
    
    // Initialize deviceNameUString
	RtlInitUnicodeString (&prefix, L"\\Device\\") ; 

	deviceNameUString.Length = 0;
	deviceNameUString.MaximumLength = prefix.Length + suffix.Length + sizeof(WCHAR);
	deviceNameUString.Buffer = MyExAllocatePool( deviceNameUString.MaximumLength);

	if (!deviceNameUString.Buffer) 
	{ 
		ntStatus = STATUS_INSUFFICIENT_RESOURCES;
		goto BCalgInitializeDiskExit;
	}

    RtlAppendUnicodeStringToString( &deviceNameUString, &prefix );
    RtlAppendUnicodeStringToString( &deviceNameUString, &suffix );
             
	// Create DeviceObject
    ntStatus = IoCreateDevice( DriverObject,             // Our Driver Object
                               sizeof(BCALG_EXTENSION), // Size of device extension 
                               &deviceNameUString,       // Device name "\Device\BC_gost"
                               FILE_DEVICE_UNKNOWN,      // Device type
                               0,                        // Device characteristics
                               FALSE,                    // Exclusive device
                               &deviceObject );          // Returned ptr to Device Object

    if ( !NT_SUCCESS( ntStatus ) ) goto BCalgInitializeDiskExit;
    
    // Initialize device object
    deviceObject->Flags |= DO_DIRECT_IO;
    deviceObject->AlignmentRequirement = FILE_WORD_ALIGNMENT;
    
    gostExtension = (PBCALG_EXTENSION)deviceObject->DeviceExtension;
    gostExtension->DeviceObject = deviceObject;
  
    // Initialize gostExtension->deviceLinkUString
    RtlInitUnicodeString (&linkPrefix, L"\\DosDevices\\") ; 

	gostExtension->deviceLinkUString.Length = 0;
	gostExtension->deviceLinkUString.MaximumLength = linkPrefix.Length + suffix.Length + sizeof(WCHAR);
	gostExtension->deviceLinkUString.Buffer = MyExAllocatePool( gostExtension->deviceLinkUString.MaximumLength);

	if (!gostExtension->deviceLinkUString.Buffer) 
		{ ntStatus = STATUS_INSUFFICIENT_RESOURCES;
          goto BCalgInitializeDiskExit;
	    }

    RtlAppendUnicodeStringToString( &gostExtension->deviceLinkUString, &linkPrefix );
    RtlAppendUnicodeStringToString( &gostExtension->deviceLinkUString, &suffix );
             
	
	// Create a symbolic link between our device name "\Device\BC_gost" and
    // the Win32 name (ie "\DosDevices\BC_gost").
    ntStatus = IoCreateSymbolicLink( &gostExtension->deviceLinkUString, &deviceNameUString );


BCalgInitializeDiskExit:

    if (suffix.Buffer != NULL) 
        RtlFreeUnicodeString( &suffix );

    if (deviceNameUString.Buffer != NULL) MyExFreePool( deviceNameUString.Buffer );
    
    if ( !NT_SUCCESS( ntStatus ) )
      {
        // Delete everything that this routine has allocated.
        if ( deviceObject != NULL ) 
		  { if (gostExtension)
               if (gostExtension->deviceLinkUString.Buffer != NULL) 
			      MyExFreePool( gostExtension->deviceLinkUString.Buffer );
			      //RtlFreeUnicodeString( &gostExtension->deviceLinkUString );

            IoDeleteDevice( deviceObject );
		  }
	  }
    return ntStatus;
}



/***************************************************************************/

NTSTATUS
BCalgCreateClose(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
    )
/*++
Routine Description:
    This routine is called by the I/O system when the BCDisk is opened or closed.
    No action is performed other than completing the request successfully.

Arguments:
    DeviceObject - a pointer to the object that represents the device
    that I/O is to be done on.
    Irp - a pointer to the I/O Request Packet for this request.

Return Value:
    STATUS_INVALID_PARAMETER if parameters are invalid,
    STATUS_SUCCESS otherwise.
--*/

{
    Irp->IoStatus.Status = STATUS_SUCCESS;
    Irp->IoStatus.Information = 0;
    IoCompleteRequest( Irp, IO_NO_INCREMENT );
    return STATUS_SUCCESS;
}



/***************************************************************************/

NTSTATUS
BCalgDeviceControl(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
    )

/*++
Routine Description:
    This routine is called by the I/O system to perform a device I/O
    control function.

Arguments:
    DeviceObject - a pointer to the object that represents the device
    that I/O is to be done on.
    Irp - a pointer to the I/O Request Packet for this request.

Return Value:
    STATUS_SUCCESS if recognized I/O control code,
    STATUS_INVALID_DEVICE_REQUEST otherwise.
--*/

{   PIO_STACK_LOCATION  irpSp;
    NTSTATUS            ntStatus;

    // Set up necessary object and extension pointers.
    irpSp = IoGetCurrentIrpStackLocation( Irp );

    // Assume failure.
    Irp->IoStatus.Status = STATUS_INVALID_DEVICE_REQUEST;

    // Determine which I/O control code was specified.
    switch ( irpSp->Parameters.DeviceIoControl.IoControlCode )
    {
      case IOCTL_PRIVATE_GET_VERSION_INFO:
		if ( irpSp->Parameters.DeviceIoControl.OutputBufferLength < sizeof(VERSION_BUFFER) )
			 {	Irp->IoStatus.Status = STATUS_INVALID_PARAMETER; }
		else {	PVERSION_BUFFER inputBuffer = Irp->AssociatedIrp.SystemBuffer;
				PVERSION_BUFFER outputBuffer = Irp->AssociatedIrp.SystemBuffer;

				// Get output buffer if its passed as an MDL
				if( Irp->MdlAddress ) 
					outputBuffer = MyMmGetSystemAddressForMdlSafe( Irp->MdlAddress, NormalPagePriority );

				if ( !RtlCompareMemory( &(inputBuffer->Signature), "LOCOS97 ", 8))
					{ Irp->IoStatus.Status = STATUS_INVALID_PARAMETER; 
					  break;
					}

				RtlCopyMemory(&(outputBuffer->Signature),"LS06CXX ", 8 );
				outputBuffer->Status		= BCSTATUS_SUCCESS;
				outputBuffer->Major_version = DRIVER_MAJOR_VERSION;
				outputBuffer->Minor_version = DRIVER_MINOR_VERSION;

				Irp->IoStatus.Status = STATUS_SUCCESS;
				Irp->IoStatus.Information = sizeof(VERSION_BUFFER);
			 }
		break;

      case IOCTL_PRIVATE_GET_INFO:
		if ( irpSp->Parameters.DeviceIoControl.OutputBufferLength < sizeof(INFO_BUFFER) )
			 {	Irp->IoStatus.Status = STATUS_INVALID_PARAMETER; }
		else {	PINFO_BUFFER inputBuffer = Irp->AssociatedIrp.SystemBuffer;
				PINFO_BUFFER outputBuffer = Irp->AssociatedIrp.SystemBuffer;
				char dName[] = DISPLAY_NAME;
				int i;

				// Get output buffer if its passed as an MDL
				if( Irp->MdlAddress ) 
					outputBuffer = MyMmGetSystemAddressForMdlSafe( Irp->MdlAddress, NormalPagePriority );

				if ( !RtlCompareMemory( &(inputBuffer->Signature), "LOCOS97 ", 8))
					{ Irp->IoStatus.Status = STATUS_INVALID_PARAMETER; 
					  break;
					}

				RtlCopyMemory(&(outputBuffer->Signature),"LS06CXX ", 8 );
				outputBuffer->Status		= BCSTATUS_SUCCESS;

				outputBuffer->AlgorithmID = ALGORITHM_ID;
				outputBuffer->KeyLength   = KEY_LENGTH;
				for(i=0; (i<DISPLAY_NAME_LENGTH) && (dName[i] != '\0'); i++)
				  { outputBuffer->DisplayName[i] = dName[i]; }
				dName[ DISPLAY_NAME_LENGTH - 1 ] = '\0';

				Irp->IoStatus.Status = STATUS_SUCCESS;
				Irp->IoStatus.Information = sizeof(INFO_BUFFER);
			 }
		break;

      case IOCTL_PRIVATE_GET_MODE_INFO:
		if ( irpSp->Parameters.DeviceIoControl.OutputBufferLength < sizeof(MODE_INFO_BUFFER) )
		{	
			Irp->IoStatus.Status = STATUS_INVALID_PARAMETER; 
		}
		else 
		{	
			PMODE_INFO_BUFFER inputBuffer  = Irp->AssociatedIrp.SystemBuffer;
			PMODE_INFO_BUFFER outputBuffer = Irp->AssociatedIrp.SystemBuffer;

			// Get output buffer if its passed as an MDL
			if( Irp->MdlAddress )
			{
				outputBuffer = MyMmGetSystemAddressForMdlSafe( Irp->MdlAddress, NormalPagePriority );
			}
                        
			if ( !RtlCompareMemory( &(inputBuffer->Signature), "LOCOS97 ", 8))
			{ 
				Irp->IoStatus.Status = STATUS_INVALID_PARAMETER; 
				break;
			}

			RtlCopyMemory(&(outputBuffer->Signature),"LS06CXX ", 8 );

			outputBuffer->Status = BCSTATUS_SUCCESS;

			switch( inputBuffer->Mode )
			{
				case EX_ENCRYPT_ECB_MODE_FLAG:
					RtlStringCbCopyA( outputBuffer->ModeName, 32, ECB_MODE_NAME );
					outputBuffer->KeyLength = KEY_LENGTH;
					break;
	
				case EX_ENCRYPT_CBC_MODE_FLAG:
					RtlStringCbCopyA( outputBuffer->ModeName, 32, CBC_MODE_NAME );
					outputBuffer->KeyLength = KEY_LENGTH;
					break;
				case EX_ENCRYPT_LRW_MODE_FLAG:
					RtlStringCbCopyA( outputBuffer->ModeName, 32, LRW_MODE_NAME );
					outputBuffer->KeyLength = MAX_KEY_LENGTH;
					break;

				case EX_ENCRYPT_XTS_MODE_FLAG:
					RtlStringCbCopyA( outputBuffer->ModeName, 32, XTS_MODE_NAME );
					if (ALGORITHM_BLOCK_SIZE_BYTES == 16)
					{
						outputBuffer->KeyLength = MAX_KEY_LENGTH;
					}
					else
					{
						outputBuffer->KeyLength = 0;
						outputBuffer->Status = BCSTATUS_ERR_ALGORITHM_NOT_SUPPORTS_XTS;
					}
					break;
			}

			outputBuffer->Reserved_1 = 0;
			outputBuffer->Reserved_2 = 0;
			outputBuffer->Reserved_3 = 0;
			outputBuffer->Reserved_4 = 0;
			
			Irp->IoStatus.Status = STATUS_SUCCESS;
			Irp->IoStatus.Information = sizeof(MODE_INFO_BUFFER);
		}

		break;

      case IOCTL_PRIVATE_CREATE_KEY_HANDLE:
		if ( irpSp->Parameters.DeviceIoControl.OutputBufferLength < sizeof(CREATEKEY_BUFFER) )
			 {	Irp->IoStatus.Status = STATUS_INVALID_PARAMETER; }
		else {	PCREATEKEY_BUFFER inputBuffer = Irp->AssociatedIrp.SystemBuffer;
				PCREATEKEY_BUFFER outputBuffer = Irp->AssociatedIrp.SystemBuffer;

				// Get output buffer if its passed as an MDL
				if( Irp->MdlAddress ) 
					outputBuffer = MyMmGetSystemAddressForMdlSafe( Irp->MdlAddress, NormalPagePriority );

				if ( !RtlCompareMemory( &(inputBuffer->Signature), "LOCOS97 ", 8))
					{ Irp->IoStatus.Status = STATUS_INVALID_PARAMETER; 
					  break;
					}

				RtlCopyMemory(&(outputBuffer->Signature),"LS06CXX ", 8 );

				ntStatus = BCalgCreateKey(outputBuffer);

				RegisterWipingKeyFunction();

				//outputBuffer->Status		= BCSTATUS_SUCCESS;
				Irp->IoStatus.Status = ntStatus;
				Irp->IoStatus.Information = sizeof( CREATEKEY_BUFFER );
			 }
		break;

      case IOCTL_PRIVATE_TEST_KEY_HANDLE:
		if ( irpSp->Parameters.DeviceIoControl.OutputBufferLength < sizeof(TESTKEY_BUFFER) )
			 {	Irp->IoStatus.Status = STATUS_INVALID_PARAMETER; }
		else {	PTESTKEY_BUFFER inputBuffer = Irp->AssociatedIrp.SystemBuffer;
				PTESTKEY_BUFFER outputBuffer = Irp->AssociatedIrp.SystemBuffer;

				if( Irp->MdlAddress ) 
					outputBuffer = MyMmGetSystemAddressForMdlSafe( Irp->MdlAddress, NormalPagePriority );

				if ( !RtlCompareMemory( &(inputBuffer->Signature), "LOCOS97 ", 8))
					{ Irp->IoStatus.Status = STATUS_INVALID_PARAMETER; 
					  break;
					}

				RtlCopyMemory(&(outputBuffer->Signature),"LS06CXX ", 8 );

				ntStatus = BC_TestKey( inputBuffer->KeyHandle );
              
				if (!NT_SUCCESS( ntStatus ))
				  { outputBuffer->Status = BCSTATUS_ERR_INVALID_KEY;
				    outputBuffer->IsValid = FALSE;
				  }
				else 
				  {	outputBuffer->Status = BCSTATUS_SUCCESS;
				    outputBuffer->IsValid = TRUE;
				  }

				Irp->IoStatus.Status = ntStatus;
				Irp->IoStatus.Information = sizeof( TESTKEY_BUFFER );
			 }
		break;

      case IOCTL_PRIVATE_FREE_KEY_HANDLE:
		if ( irpSp->Parameters.DeviceIoControl.OutputBufferLength < sizeof(TESTKEY_BUFFER) )
			 {	Irp->IoStatus.Status = STATUS_INVALID_PARAMETER; }
		else {	PTESTKEY_BUFFER inputBuffer = Irp->AssociatedIrp.SystemBuffer;
				PTESTKEY_BUFFER outputBuffer = Irp->AssociatedIrp.SystemBuffer;

				if( Irp->MdlAddress ) 
					outputBuffer = MyMmGetSystemAddressForMdlSafe( Irp->MdlAddress, NormalPagePriority );

				if ( !RtlCompareMemory( &(inputBuffer->Signature), "LOCOS97 ", 8))
					{ Irp->IoStatus.Status = STATUS_INVALID_PARAMETER; 
					  break;
					}

				RtlCopyMemory(&(outputBuffer->Signature),"LS06CXX ", 8 );

				ntStatus = BC_FreeKey( inputBuffer->KeyHandle );
              
				if (!NT_SUCCESS( ntStatus ))
				  { outputBuffer->Status = BCSTATUS_ERR_INVALID_KEY;
				    outputBuffer->IsValid = FALSE;
				  }
				else 
				  {	outputBuffer->Status = BCSTATUS_SUCCESS;
				    outputBuffer->IsValid = TRUE;
				  }

				Irp->IoStatus.Status = ntStatus;
				Irp->IoStatus.Information = sizeof( TESTKEY_BUFFER );
			 }
		break;


      case IOCTL_PRIVATE_ENCRYPT:
		if ( irpSp->Parameters.DeviceIoControl.InputBufferLength < sizeof(ENCRYPT_BUFFER) )
		{	
			Irp->IoStatus.Status = STATUS_INVALID_PARAMETER; 
		}
		else 
		{	
			PENCRYPT_BUFFER inputBuffer = Irp->AssociatedIrp.SystemBuffer;
			PUCHAR          outputBuffer = Irp->AssociatedIrp.SystemBuffer;

			// Get output buffer if its passed as an MDL (METHOD_OUT_DIRECT used)
			if( Irp->MdlAddress ) 
				outputBuffer = MyMmGetSystemAddressForMdlSafe( Irp->MdlAddress, NormalPagePriority );

			if ( !RtlCompareMemory( &(inputBuffer->Signature), "LOCOS97 ", 8))
			{ 
				Irp->IoStatus.Status = STATUS_INVALID_PARAMETER; 
				break;
			}

			//RtlCopyMemory(&(inputBuffer->Signature),"LS06CXX ", 8 );
			
			ntStatus = BC_Encrypt(inputBuffer->KeyHandle,
			                      inputBuffer->IVector,
			                      outputBuffer, outputBuffer,
			                      inputBuffer->Length );

			//outputBuffer->Status	  = BCSTATUS_SUCCESS;
			Irp->IoStatus.Status      = ntStatus;
			Irp->IoStatus.Information = sizeof( ENCRYPT_BUFFER );
		}
		break;

      case IOCTL_PRIVATE_ENCRYPT_DECRYPT_V8:
		if ( irpSp->Parameters.DeviceIoControl.InputBufferLength < sizeof(ENCRYPT_BUFFER_V8) )
		{	
			Irp->IoStatus.Status = STATUS_INVALID_PARAMETER; 
		}
		else 
		{	
			PENCRYPT_BUFFER_V8 inputBuffer = Irp->AssociatedIrp.SystemBuffer;
			PUCHAR             outputBuffer = Irp->AssociatedIrp.SystemBuffer;

			// Get output buffer if its passed as an MDL (METHOD_OUT_DIRECT used)
			if( Irp->MdlAddress ) 
				outputBuffer = MyMmGetSystemAddressForMdlSafe( Irp->MdlAddress, NormalPagePriority );

			if ( !RtlCompareMemory( &(inputBuffer->Signature), "LOCOS97 ", 8))
			{ 
				Irp->IoStatus.Status = STATUS_INVALID_PARAMETER; 
				break;
			}

			//RtlCopyMemory(&(inputBuffer->Signature),"LS06CXX ", 8 );
			
			if (inputBuffer->EncryptOperation)
			{
				ntStatus = BC_EncryptEx( 
                                              inputBuffer->KeyHandle,
			                      inputBuffer->IVector,
			                      outputBuffer, outputBuffer,
			                      inputBuffer->Length,
                                              inputBuffer->EncryptionMode );
			}
			else
			{
				ntStatus = BC_DecryptEx( 
                                              inputBuffer->KeyHandle,
			                      inputBuffer->IVector,
			                      outputBuffer, outputBuffer,
			                      inputBuffer->Length,
                                              inputBuffer->EncryptionMode );
			}

			//outputBuffer->Status      = BCSTATUS_SUCCESS;
			Irp->IoStatus.Status      = ntStatus;
			Irp->IoStatus.Information = sizeof( ENCRYPT_BUFFER );
		}
		break;

      case IOCTL_PRIVATE_DECRYPT:
		if ( irpSp->Parameters.DeviceIoControl.InputBufferLength < sizeof(ENCRYPT_BUFFER) )
			 {	Irp->IoStatus.Status = STATUS_INVALID_PARAMETER; }
		else {	PENCRYPT_BUFFER inputBuffer = Irp->AssociatedIrp.SystemBuffer;
				PUCHAR          outputBuffer = Irp->AssociatedIrp.SystemBuffer;

				// Get output buffer if its passed as an MDL (METHOD_OUT_DIRECT used)
				if( Irp->MdlAddress ) 
					outputBuffer = MyMmGetSystemAddressForMdlSafe( Irp->MdlAddress, NormalPagePriority );

				if ( !RtlCompareMemory( &(inputBuffer->Signature), "LOCOS97 ", 8))
					{ Irp->IoStatus.Status = STATUS_INVALID_PARAMETER; 
					  break;
					}

				//RtlCopyMemory(&(inputBuffer->Signature),"LS06CXX ", 8 );
				ntStatus = BC_Decrypt(inputBuffer->KeyHandle,
                                      inputBuffer->IVector,
                                      outputBuffer, outputBuffer,
                                      inputBuffer->Length );

				//outputBuffer->Status		= BCSTATUS_SUCCESS;
				Irp->IoStatus.Status = ntStatus;
				Irp->IoStatus.Information = sizeof( ENCRYPT_BUFFER );
			 }
		break;

      case IOCTL_PRIVATE_GETFUNCTIONS:
		if ( irpSp->Parameters.DeviceIoControl.InputBufferLength < sizeof(GETFUNC_BUFFER) )
		{	
			Irp->IoStatus.Status = STATUS_INVALID_PARAMETER; 
		}
		else 
		{
			PGETFUNC_BUFFER inputBuffer  = Irp->AssociatedIrp.SystemBuffer;
			PGETFUNC_BUFFER outputBuffer = Irp->AssociatedIrp.SystemBuffer;

			// Get output buffer if its passed as an MDL
			if( Irp->MdlAddress ) 
				outputBuffer = MyMmGetSystemAddressForMdlSafe( Irp->MdlAddress, NormalPagePriority );

			if ( !RtlCompareMemory( &(inputBuffer->Signature), "LOCOS97 ", 8))
			{ 
				Irp->IoStatus.Status = STATUS_INVALID_PARAMETER; 
				  break;
			}

			RtlCopyMemory(&(inputBuffer->Signature),"LS06CXX ", 8 );
                
			outputBuffer->FuncTestKey = (PVOID)(&BC_TestKey);
			outputBuffer->FuncFreeKey = (PVOID)(&BC_FreeKey);
			outputBuffer->FuncEncrypt = (PVOID)(&BC_Encrypt);
			outputBuffer->FuncDecrypt = (PVOID)(&BC_Decrypt);

			//outputBuffer->Status = BCSTATUS_SUCCESS;
			Irp->IoStatus.Status = STATUS_SUCCESS;
			Irp->IoStatus.Information = sizeof(GETFUNC_BUFFER);
		}
		break;

      case IOCTL_PRIVATE_GETFUNCTIONS_EX:
		if ( irpSp->Parameters.DeviceIoControl.InputBufferLength < sizeof(GETFUNC_BUFFER) )
		{	
			Irp->IoStatus.Status = STATUS_INVALID_PARAMETER; 
		}
		else 
		{
			PGETFUNC_BUFFER inputBuffer  = Irp->AssociatedIrp.SystemBuffer;
			PGETFUNC_BUFFER outputBuffer = Irp->AssociatedIrp.SystemBuffer;

			// Get output buffer if its passed as an MDL
			if( Irp->MdlAddress ) 
				outputBuffer = MyMmGetSystemAddressForMdlSafe( Irp->MdlAddress, NormalPagePriority );

			if ( !RtlCompareMemory( &(inputBuffer->Signature), "LOCOS97 ", 8))
			{ 
				Irp->IoStatus.Status = STATUS_INVALID_PARAMETER; 
				  break;
			}

			RtlCopyMemory(&(inputBuffer->Signature),"LS06CXX ", 8 );
                
			outputBuffer->FuncTestKey = (PVOID)(&BC_TestKey);
			outputBuffer->FuncFreeKey = (PVOID)(&BC_FreeKey);
			outputBuffer->FuncEncrypt = (PVOID)(&BC_EncryptEx);
			outputBuffer->FuncDecrypt = (PVOID)(&BC_DecryptEx);

			//outputBuffer->Status = BCSTATUS_SUCCESS;
			Irp->IoStatus.Status = STATUS_SUCCESS;
			Irp->IoStatus.Information = sizeof(GETFUNC_BUFFER);
		}
		break;

      case IOCTL_PRIVATE_SUPPORTS_SYNCRO_64BITS:
		if ( irpSp->Parameters.DeviceIoControl.OutputBufferLength < sizeof(SYNCRO64BITS_BUFFER) )
			 {	Irp->IoStatus.Status = STATUS_INVALID_PARAMETER; }
		else {	PSYNCRO64BITS_BUFFER inputBuffer = Irp->AssociatedIrp.SystemBuffer;
		        PSYNCRO64BITS_BUFFER outputBuffer = Irp->AssociatedIrp.SystemBuffer;

				// Get output buffer if its passed as an MDL
				if( Irp->MdlAddress ) 
					outputBuffer = MyMmGetSystemAddressForMdlSafe( Irp->MdlAddress, NormalPagePriority );

				if ( !RtlCompareMemory( &(inputBuffer->Signature), "LOCOS97 ", 8))
					{ Irp->IoStatus.Status = STATUS_INVALID_PARAMETER; 
					  break;
					}

				RtlCopyMemory(&(outputBuffer->Signature),"LS06CXX ", 8 );
				outputBuffer->Status        = BCSTATUS_SUCCESS;
				outputBuffer->SupportStatus = TRUE;

				Irp->IoStatus.Status = STATUS_SUCCESS;
				Irp->IoStatus.Information = sizeof(VERSION_BUFFER);
			 }
		break;

	default:

        // The specified I/O control code is unrecognized by this driver.
        // The I/O status field in the IRP has already been set so just
        // terminate the switch.
        break;
    }

    // Finish the I/O operation by simply completing the packet and returning
    // the same status as in the packet itself.
    ntStatus = Irp->IoStatus.Status;
    IoCompleteRequest( Irp, IO_NO_INCREMENT );
    return ntStatus;
}

/***************************************************************************/

VOID
BCalgUnloadDriver(
    IN PDRIVER_OBJECT DriverObject
    )
/*++
Routine Description:
    This routine is called by the I/O system to unload the driver.
    Any resources previously allocated must be freed.

Arguments:
    DriverObject - a pointer to the object that represents our driver.

Return Value:
    None
--*/

{
    PDEVICE_OBJECT deviceObject = DriverObject->DeviceObject;
    KEY_HANDLE     keyHandle;   
    NTSTATUS       ntStatus;

    // We must free memory allocated for all key handles
    for(keyHandle = 0; keyHandle < TABLE_LENGTH ; keyHandle++)
    {
         if (KeyTable[keyHandle].keyExtended != NULL)
         {
             ntStatus = BC_FreeKey( keyHandle );
         }
    }

    if ( deviceObject != NULL )
    {
        PBCALG_EXTENSION gostExtension = deviceObject->DeviceExtension;

        if (gostExtension != NULL)
        { if (gostExtension->deviceLinkUString.Buffer != NULL)
           { IoDeleteSymbolicLink( &gostExtension->deviceLinkUString );
             MyExFreePool( gostExtension->deviceLinkUString.Buffer );
           }
        }
        IoDeleteDevice( deviceObject );
    }
}


/***************************************************************************/

NTSTATUS
BCalgCreateKey(IN PCREATEKEY_BUFFER CkBuffer)
/*++
Routine Description:
 
Arguments:
 
Return Value:
--*/
{ KEY_HANDLE   keyHandle;
  UCHAR *keyExtended = NULL;
  UCHAR *pool        = NULL;    
  UCHAR *lrwContext  = NULL;
  UCHAR *xtsContext  = NULL;
	
  if ((CkBuffer->KeyLength) < KEY_LENGTH)
    { CkBuffer->Status = BCSTATUS_ERR_INVALID_KEY_LENGTH;
	  goto errorExit;
    }

  // Define an empty element in Key Table
  for(keyHandle = 0; keyHandle < TABLE_LENGTH ; keyHandle++)
    if (KeyTable[keyHandle].keyExtended == NULL) break;

  if (keyHandle >= TABLE_LENGTH) 
    { CkBuffer->Status = BCSTATUS_ERR_NOSYSTEM_RESOURCES;
      goto errorExit;
    }
  
  // Allocate memory space for Key structure
  keyExtended = (UCHAR *)MyExAllocatePool( KEY_LENGTH_EXTENDED / 8 );
  pool        = (UCHAR *)MyExAllocatePool( POOL_SIZE_BYTES );
  lrwContext  = (UCHAR *)MyExAllocatePool( sizeof(GfmulCtx) );
  xtsContext  = (UCHAR *)MyExAllocatePool( KEY_LENGTH_EXTENDED / 8 );

  if ((keyExtended == NULL) || (pool == NULL) || (lrwContext == NULL) || (xtsContext == NULL))
  {
      CkBuffer->Status = BCSTATUS_ERR_NOSYSTEM_RESOURCES;
      goto errorExit;
  }

  RtlZeroMemory( keyExtended, KEY_LENGTH_EXTENDED / 8 );
  RtlZeroMemory( lrwContext,  sizeof(GfmulCtx) );
  RtlZeroMemory( xtsContext,  KEY_LENGTH_EXTENDED / 8 );

  // Initialize pool of random data
  RtlCopyBytes( pool, CkBuffer->Pool, POOL_SIZE_BYTES );
  
  // Initialize Key structure
  if (!BC_ExtendKey(CkBuffer->Key, keyExtended))
  {
	  CkBuffer->Status = BCSTATUS_ERR_ALGORITHM_FAILED_SELF_TEST;
	  goto errorExit;
  }

  // Initialize context for LRW mode
  if (ALGORITHM_BLOCK_SIZE_BYTES == 8)
  {
      GFMUL_InitCtxBlock8( &(CkBuffer->Key[(CkBuffer->KeyLength / 8) - 8]), (GfmulCtx *)lrwContext );
  }
  else
  {
      GFMUL_InitCtxBlock16( &(CkBuffer->Key[(CkBuffer->KeyLength / 8) - 16]), (GfmulCtx *)lrwContext );
  }  

  // Initialize XTS context structure
  if (!BC_ExtendKey( &(CkBuffer->Key[XTS_KEY2_OFFSET]), xtsContext ))
  {
	  CkBuffer->Status = BCSTATUS_ERR_ALGORITHM_FAILED_SELF_TEST;
	  goto errorExit;
  }

  KeyTable[keyHandle].pool        = pool;
  KeyTable[keyHandle].keyExtended = keyExtended;
  KeyTable[keyHandle].lrwContext  = lrwContext;
  KeyTable[keyHandle].xtsContext  = xtsContext;

  CkBuffer->KeyHandle = keyHandle + 1; // 0 is invalid Key HAndle
  CkBuffer->Status = BCSTATUS_SUCCESS;
  return STATUS_SUCCESS;

errorExit:
  if (keyExtended != NULL)
  { MyExFreePool(keyExtended);
  }
  if (pool != NULL)
  { MyExFreePool(pool);
  }
  if (lrwContext != NULL)
  { MyExFreePool(lrwContext);
  }
  if (xtsContext != NULL)
  { MyExFreePool(xtsContext);
  }
  CkBuffer->KeyHandle = 0;
  return STATUS_SUCCESS;
}


/***************************************************************************/

BOOL 
BC_ExtendKey( PUCHAR KeyFrom, PUCHAR KeyTo )
/*++
Routine Description:
 
Arguments:
 
Return Value:
--*/
{
  if ((!KeyFrom) || (!KeyTo)) return FALSE;
  
  return KeyExtend( KeyFrom, (PDWORD) KeyTo );
}


/**************************************************************************/

VOID
Transform_IV( DWORD *IVector, DWORD *Pool)
{
	DWORD i;
	i = (IVector[0] & 0x3f) << 1;
	IVector[0] ^= Pool[i];
	IVector[1] ^= Pool[i+1];
}


VOID Increment_IV( PUCHAR Syncro )
{
  PWORD          pdwSyncro;
  UINT           i;
  PLARGE_INTEGER pliSyncro;

  pdwSyncro = (PWORD)Syncro;

  if ( (pdwSyncro[0] == pdwSyncro[1]) && // for compatibility with older versions of BestCrypt
       (pdwSyncro[1] == pdwSyncro[2]) && 
       (pdwSyncro[2] == pdwSyncro[3]) )
  {
     
     for( i=0; i<4; i++) (pdwSyncro[i])++;
  }
  else
  {
     pliSyncro = (PLARGE_INTEGER)Syncro;
     (pliSyncro->QuadPart)++;
  }
}



VOID
Transform_Buf(	DWORD *Pool, 
                DWORD *InBuffer, 
                DWORD *OutBuffer, 
                DWORD Length )    // Length in DWORDs
{
	DWORD i;
	for ( i=0; i<Length; i++ ) OutBuffer[i] = InBuffer[i] ^ Pool[i];
}


/***************************************************************************/

NTSTATUS 
BC_Encrypt( KEY_HANDLE KeyHandle,
            PUCHAR     IVector,
            PUCHAR     InBuffer,
            PUCHAR     OutBuffer,
            DWORD      Length )       // in bytes
/*++
Routine Description:
 
Arguments:
 
Return Value:
--*/
{ KEY_HANDLE hKey  = KeyHandle - 1;
  DWORD length = Length;
  PWORD s;
  WORD  ivector[4], ivectorS[4];
  UINT  i, offset;
  
  if (!Length) return STATUS_SUCCESS;

  if (!NT_SUCCESS( BC_TestKey( KeyHandle ) ))
     return STATUS_INSUFFICIENT_RESOURCES;

  s = (PWORD)IVector;
  for( i=0; i<4; i++) ivector[i] = ivectorS[i] = s[i];
  
  for( offset = 0; ; offset += 512)
  { 
	Transform_IV(	(PDWORD)ivector, (PDWORD)KeyTable[hKey].pool );
	if (length > 512) 
	  { length = length - 512;
		Transform_Buf(	(PDWORD)KeyTable[hKey].pool,
						(PDWORD)(&(InBuffer[offset] )),
						(PDWORD)(&(OutBuffer[offset])), 128  );
        Encrypt( (PDWORD)ivector,  (PDWORD)KeyTable[hKey].keyExtended,
                 (PDWORD)(&(OutBuffer[offset] )),
				 (PDWORD)(&(OutBuffer[offset])), 512 );
	  }
	else
	  {	Transform_Buf(	(PDWORD)KeyTable[hKey].pool,
						(PDWORD)(&(InBuffer[offset] )),
						(PDWORD)(&(OutBuffer[offset])), length >> 2 );
		Encrypt( (PDWORD)ivector,  (PDWORD)KeyTable[hKey].keyExtended,
                 (PDWORD)(&(OutBuffer[offset] )),
				 (PDWORD)(&(OutBuffer[offset])), length );
	    break;
	  }
    // We need to update IVector every 512 bytes
    Increment_IV( (PUCHAR)ivectorS );
    for( i=0; i<4; i++) ivector[i] = ivectorS[i]; 
  }
  return STATUS_SUCCESS;
}


/***************************************************************************/

NTSTATUS 
BC_Decrypt( KEY_HANDLE KeyHandle,
            PUCHAR     IVector,
            PUCHAR     InBuffer,
            PUCHAR     OutBuffer,
            DWORD      Length )       // in bytes
/*++
Routine Description:
 
Arguments:
 
Return Value:
--*/
{ KEY_HANDLE hKey  = KeyHandle - 1;
  DWORD length = Length;
  PWORD s;
  WORD  ivector[4], ivectorS[4];
  UINT  i, offset;

  if (!Length) return STATUS_SUCCESS;

  if (!NT_SUCCESS( BC_TestKey( KeyHandle ) ))
     return STATUS_INSUFFICIENT_RESOURCES;

  s = (PWORD)IVector;
  for( i=0; i<4; i++) ivector[i] = ivectorS[i] = s[i];
  
  for( offset = 0; ; offset += 512)
  {
	Transform_IV(	(PDWORD)ivector, (PDWORD)KeyTable[hKey].pool );
	if (length > 512) 
	  { length = length - 512;
        Decrypt( (PDWORD)ivector,  (PDWORD)KeyTable[hKey].keyExtended,
                 (PDWORD)(&(InBuffer[offset] )),
				 (PDWORD)(&(OutBuffer[offset])), 512 );
		Transform_Buf(	(PDWORD)KeyTable[hKey].pool,
						(PDWORD)(&(OutBuffer[offset] )),
						(PDWORD)(&(OutBuffer[offset])), 128 );
	  }
	else
	  { Decrypt( (PDWORD)ivector,  (PDWORD)KeyTable[hKey].keyExtended,
                 (PDWORD)(&(InBuffer[offset] )),
				 (PDWORD)(&(OutBuffer[offset])), length );
		Transform_Buf(	(PDWORD)KeyTable[hKey].pool,
						(PDWORD)(&(OutBuffer[offset] )),
						(PDWORD)(&(OutBuffer[offset])), length >> 2 );
	    break;
	  }
    // We need to update IVector every 512 bytes
    Increment_IV( (PUCHAR)ivectorS );
    for( i=0; i<4; i++) ivector[i] = ivectorS[i]; 
  }
  return STATUS_SUCCESS;
}



/***************************************************************************/

NTSTATUS 
BC_TestKey( KEY_HANDLE KeyHandle )
/*++
Routine Description:
 
Arguments:
 
Return Value:
--*/
{ KEY_HANDLE hKey = KeyHandle - 1;

  if (KeyHandle == (KEY_HANDLE)NOT_EXISTING_KEY_HANDLE_V8_SPECIFIC)
         return STATUS_SUCCESS; // read explanations in bc_ioctl.h
  
  if (hKey >= TABLE_LENGTH) return STATUS_INSUFFICIENT_RESOURCES;
  
  if (!(KeyTable[hKey].keyExtended)) return STATUS_INSUFFICIENT_RESOURCES;

  return STATUS_SUCCESS;
}


/***************************************************************************/

BC_ALGID_BLOCK sig = { EYE_CATCH_STRING, ALGORITHM_ID, KEY_LENGTH, DISPLAY_NAME, SERVICE_NAME, EYE_CATCH_STRING_1, sizeof(BC_ALGID_BLOCK), ALGORITHM_BLOCK_SIZE_BYTES, 512 };

NTSTATUS 
BC_FreeKey( KEY_HANDLE KeyHandle )
/*++
Routine Description:
 
Arguments:
 
Return Value:
--*/
{ KEY_HANDLE hKey = KeyHandle - 1;
  DWORD      i;

  if (!NT_SUCCESS( BC_TestKey( KeyHandle ) ))
     return STATUS_INSUFFICIENT_RESOURCES;

  /* Erase key data and pool from memory */

  RtlSecureZeroMemory( KeyTable[hKey].keyExtended, KEY_LENGTH_EXTENDED / 8);
  RtlSecureZeroMemory( KeyTable[hKey].pool, POOL_SIZE_BYTES );
  RtlSecureZeroMemory( KeyTable[hKey].lrwContext, sizeof(GfmulCtx) );
  RtlSecureZeroMemory( KeyTable[hKey].xtsContext, KEY_LENGTH_EXTENDED / 8 );

  MyExFreePool( KeyTable[hKey].keyExtended );
  MyExFreePool( KeyTable[hKey].pool );
  MyExFreePool( KeyTable[hKey].lrwContext );
  MyExFreePool( KeyTable[hKey].xtsContext );

  KeyTable[hKey].keyExtended = NULL;
  KeyTable[hKey].pool = NULL;
  KeyTable[hKey].lrwContext = NULL;
  KeyTable[hKey].xtsContext = NULL;

  return STATUS_SUCCESS;
}


/***************************************************************************/


int Encrypt_Decrypt_LRW( BOOL          IsEncryptOperation,
                         unsigned char *InBuffer,
                         unsigned char *OutBuffer,
                         unsigned long Length,         // in bytes
                         unsigned char *InitialVector, // 8 bytes - it is address of 512-byte sector !
                         unsigned char *ExtendedKey,
                         unsigned char *LrwContext )
{
	int            blockSize_words = 0, length_words, j;
	unsigned short *in_w, *out_w;
	unsigned short ivLocal[8], t_lrw[8];
	LARGE_INTEGER  *liByteAddress;

	RtlSecureZeroMemory( ivLocal, sizeof(unsigned short) * 8 );

	liByteAddress = (LARGE_INTEGER *)ivLocal;

	*liByteAddress = *( (PLARGE_INTEGER)InitialVector );

	// make the address from sector number to byte offset
	liByteAddress->QuadPart = liByteAddress->QuadPart << 9;

	blockSize_words = ALGORITHM_BLOCK_SIZE_BYTES / sizeof(unsigned short);
	length_words    = (int)Length     / sizeof(unsigned short);

	in_w  = (unsigned short *)InBuffer;
	out_w = (unsigned short *)OutBuffer;

	for( j = 0; j < length_words; j += blockSize_words)
	{
		// calculate t and then IV for next block
		if ( ALGORITHM_BLOCK_SIZE_BYTES == 8)
		{
			GFMUL_MulBlock8( (unsigned char *)ivLocal, (unsigned char *)t_lrw, (GfmulCtx *)LrwContext );

			// add 8 (size of block) to ivLocal
			liByteAddress->QuadPart = liByteAddress->QuadPart + 8;

			out_w[0] = in_w[0] ^ t_lrw[0];
			out_w[1] = in_w[1] ^ t_lrw[1];
			out_w[2] = in_w[2] ^ t_lrw[2];
			out_w[3] = in_w[3] ^ t_lrw[3];

			if (IsEncryptOperation)
				Encrypt_ECB( (unsigned char *)out_w, (unsigned char *)out_w, ExtendedKey );
			else
				Decrypt_ECB( (unsigned char *)out_w, (unsigned char *)out_w, ExtendedKey );

			out_w[0] ^= t_lrw[0];
			out_w[1] ^= t_lrw[1];
			out_w[2] ^= t_lrw[2];
			out_w[3] ^= t_lrw[3];

			in_w  = &(in_w[4]);
			out_w = &(out_w[4]);
		}
		else
		{
			GFMUL_MulBlock16( (unsigned char *)ivLocal, (unsigned char *)t_lrw, (GfmulCtx *)LrwContext );

			// add 16 (size of block) to ivLocal
			liByteAddress->QuadPart = liByteAddress->QuadPart + 16;

			out_w[0] = in_w[0] ^ t_lrw[0];
			out_w[1] = in_w[1] ^ t_lrw[1];
			out_w[2] = in_w[2] ^ t_lrw[2];
			out_w[3] = in_w[3] ^ t_lrw[3];
			out_w[4] = in_w[4] ^ t_lrw[4];
			out_w[5] = in_w[5] ^ t_lrw[5];
			out_w[6] = in_w[6] ^ t_lrw[6];
			out_w[7] = in_w[7] ^ t_lrw[7];

			if (IsEncryptOperation)
				Encrypt_ECB( (unsigned char *)out_w, (unsigned char *)out_w, ExtendedKey );
			else
				Decrypt_ECB( (unsigned char *)out_w, (unsigned char *)out_w, ExtendedKey );

			out_w[0] ^= t_lrw[0];
			out_w[1] ^= t_lrw[1];
			out_w[2] ^= t_lrw[2];
			out_w[3] ^= t_lrw[3];
			out_w[4] ^= t_lrw[4];
			out_w[5] ^= t_lrw[5];
			out_w[6] ^= t_lrw[6];
			out_w[7] ^= t_lrw[7];

			in_w  = &(in_w[8]);
			out_w = &(out_w[8]);
		}

        }
	return 1;
}



/***************************************************************************/


// EX_ENCRYPT flags are defined in bc_ioctl.h
// #define EX_ENCRYPT_LRW_MODE_FLAG  0x00000001
// #define EX_ENCRYPT_CBC_MODE_FLAG  0x00000002
// #define EX_ENCRYPT_XTS_MODE_FLAG  0xBC000004


NTSTATUS 
BC_EncryptEx( KEY_HANDLE KeyHandle,
              PUCHAR     IVector,
              PUCHAR     InBuffer,
              PUCHAR     OutBuffer,
              DWORD      Length,       // in bytes
              DWORD      ExEncryptFlags)
{
	KEY_HANDLE hKey  = KeyHandle - 1;

	if (KeyHandle == 0)
		return STATUS_UNSUCCESSFUL;
 
	if (ExEncryptFlags == EX_ENCRYPT_CBC_MODE_FLAG)
	{
		return BC_Encrypt( KeyHandle, IVector, InBuffer, OutBuffer, Length );
	}
	else if (ExEncryptFlags == EX_ENCRYPT_LRW_MODE_FLAG)
	{
		Encrypt_Decrypt_LRW( TRUE, InBuffer, OutBuffer, Length, IVector, 
	                     KeyTable[hKey].keyExtended, 
	                     KeyTable[hKey].lrwContext );
	}
	else if (ExEncryptFlags == EX_ENCRYPT_XTS_MODE_FLAG)
	{
		if (sig.blockSize != 16)
			return BCSTATUS_ERR_ALGORITHM_NOT_SUPPORTS_XTS;

		XTS_EncryptDecrypt( Encrypt_ECB, Encrypt_ECB, *((ULONGLONG *)IVector), 
		                    InBuffer, OutBuffer, Length, 
		                    KeyTable[hKey].keyExtended, KeyTable[hKey].xtsContext );
	}	

	return STATUS_SUCCESS;
}

/*
BOOL DbgCheck = FALSE;
ULONGLONG DbgCheckValue = 0;
*/

NTSTATUS 
BC_DecryptEx( KEY_HANDLE KeyHandle,
              PUCHAR     IVector,
              PUCHAR     InBuffer,
              PUCHAR     OutBuffer,
              DWORD      Length,       // in bytes
              DWORD      ExEncryptFlags)
{
	KEY_HANDLE hKey  = KeyHandle - 1;
 
	if (KeyHandle == 0)
		return STATUS_UNSUCCESSFUL;
 
	if (ExEncryptFlags == EX_ENCRYPT_CBC_MODE_FLAG)
	{
		return BC_Decrypt( KeyHandle, IVector, InBuffer, OutBuffer, Length );
	}
	else if (ExEncryptFlags == EX_ENCRYPT_LRW_MODE_FLAG)
	{
		Encrypt_Decrypt_LRW( FALSE, InBuffer, OutBuffer, Length, IVector, 
	                     KeyTable[hKey].keyExtended, 
	                     KeyTable[hKey].lrwContext );
	}	
	else if (ExEncryptFlags == EX_ENCRYPT_XTS_MODE_FLAG)
	{
		if (sig.blockSize != 16)
			return BCSTATUS_ERR_ALGORITHM_NOT_SUPPORTS_XTS;

		XTS_EncryptDecrypt( Encrypt_ECB, Decrypt_ECB, *((ULONGLONG *)IVector), 
		                    InBuffer, OutBuffer, Length, 
		                    KeyTable[hKey].keyExtended, KeyTable[hKey].xtsContext );

		/*
		if (DbgCheck)
		{
			DWORD i;

			for (i=0; i < Length / sizeof(ULONGLONG); i++)
			{
				if ( ((ULONGLONG *)OutBuffer)[i] != DbgCheckValue)
				{
					DbgPrint("Decrypted value is not 0!\n");
				}
			}
		}
		*/
	}	

	return STATUS_SUCCESS;
}


/*******************************************************************/

#define BCFNT_NAME "\\Device\\bcfnt"

char Locos03[] = "LOCOS03 ";
char Ls06cxx[] = "LS06CXX ";

BOOL WipeFunctionIsRegistered = FALSE;

BOOL BC_WipeAllKeys()
{
	KEY_HANDLE keyHandle;

	for( keyHandle = 1; keyHandle <= TABLE_LENGTH; keyHandle++ )
	{
		BC_FreeKey( keyHandle );
	}

	return TRUE;
}


NTSTATUS
RegisterWipingKeyFunction()
/*++
Routine Description:
 
Arguments:
 
Return Value:
--*/
{
  UNICODE_STRING  usBcfntName;
  ANSI_STRING     ansiBcfntName;
  PFILE_OBJECT    fileObject = NULL;
  PDEVICE_OBJECT  deviceObject = NULL;
  PIRP            pIRP = NULL;
  IO_STATUS_BLOCK IoStatusBlock;
  KEVENT          event;
  NTSTATUS        ntStatus = 0xFF;

  REGISTER_WIPE_FUNCTION_BUFFER rwBuffer;
 
  if (WipeFunctionIsRegistered) return STATUS_SUCCESS;

  WipeFunctionIsRegistered = TRUE;
 
  RtlInitAnsiString( &ansiBcfntName, BCFNT_NAME );

  if (!NT_SUCCESS( RtlAnsiStringToUnicodeString( &usBcfntName, &ansiBcfntName, TRUE)))
  { return STATUS_SUCCESS;
  }

  // define PDEVICE_OBJECT and PFILE_OBJECT for BCFNT driver
  // because we are going to ask the driver wipe encryption keys 
  // after finishing hibernate
  ntStatus = IoGetDeviceObjectPointer( &usBcfntName, FILE_ALL_ACCESS, 
                                       &fileObject, &deviceObject);

  if (!NT_SUCCESS(ntStatus))
     { goto exit;
     }

  RtlZeroMemory( &rwBuffer, sizeof(REGISTER_WIPE_FUNCTION_BUFFER) );
  RtlCopyMemory( rwBuffer.Signature, Locos03, 8 );
  rwBuffer.Status = 0xFF;
  rwBuffer.Size = sizeof(REGISTER_WIPE_FUNCTION_BUFFER);
  rwBuffer.Register = TRUE;
  rwBuffer.WipeKeyProc = (PVOID)BC_WipeAllKeys;

  KeInitializeEvent(&event, NotificationEvent, FALSE);

  pIRP = IoBuildDeviceIoControlRequest( 
                IOCTL_PRIVATE_REGISTER_WIPE_FUNCTION, deviceObject,
                &rwBuffer, sizeof(REGISTER_WIPE_FUNCTION_BUFFER),
                &rwBuffer, sizeof(REGISTER_WIPE_FUNCTION_BUFFER),
		FALSE, &event,
		&IoStatusBlock );

  if (!pIRP) goto exit;

  ntStatus = IoCallDriver( deviceObject, pIRP );

  if (!NT_SUCCESS(ntStatus)) goto exit;

  KeWaitForSingleObject(&event, Executive, KernelMode, FALSE, NULL);

  if (RtlCompareMemory( rwBuffer.Signature, Ls06cxx, 8 ) != 8)
  {
    ntStatus = 0xF0;
  }

  ntStatus = (NTSTATUS)rwBuffer.Status;

exit:
  if (fileObject) ObDereferenceObject(fileObject);
  RtlFreeUnicodeString( &usBcfntName );
  return ntStatus;
}

