/////////////////////////////////////////////////////////////////////////////////////
//
//      Copyright (c) 2017 Jetico Inc. Oy, Finland
//      All rights reserved.
//
//      File: bc_camellia.h
//      Revision:
//      Created:	22.07.2017
//      Description:
//            RFC 3713: "Camellia Encryption Algorithm", April 2004
//      Implementation:
//            Based on code from camellia.h/camellia.c
//
/////////////////////////////////////////////////////////////////////////////////////


#include <ntddk.h>          // for DbgBreakPoint()

#include "bc_camellia.h"
#include "camellia.h"


int Camellia_SelfTest();  // returns 0 if self tests returned successfully


//************ Camellia ECB mode decryption - Export Variant **************

void Encrypt_ECB( BYTE  *Input,        /* 16 bytes of input data              */
                  BYTE  *Output,       /* 16 bytes of output data             */
                  BYTE  *ExtendedKey)
{
	Camellia_EncryptBlock(KEY_LENGTH, (const unsigned char *)Input, (unsigned int *)ExtendedKey, (unsigned char *)Output);
}


void Decrypt_ECB( BYTE  *Input,        /* 8 bytes of input data              */
                  BYTE  *Output,       /* 8 bytes of output data             */
                  BYTE  *ExtendedKey)
{
	Camellia_DecryptBlock(KEY_LENGTH, (const unsigned char *)Input, (unsigned int *)ExtendedKey, (unsigned char *)Output);
}



//************ Camellia CBC mode encryption **************

VOID
Encrypt( DWORD *IVector, 
         DWORD *KeyAddress,
         DWORD *SrcBuffer,
         DWORD *DstBuffer,
         DWORD Length )    // in bytes
{
	DWORD i,j;
	DWORD ivLocal[4];

	for(i=0; i<4; i++) 
		ivLocal[i] = IVector[i&1];

	for(i=0; Length > 0; Length -= 16, i=i+4)
	{
		// Use CFB mode to encrypt data of length less than a full block
		if (Length < 16)
		{
			Encrypt_ECB( (BYTE *)ivLocal, (BYTE *)ivLocal, (BYTE *)KeyAddress);

			for (j = 0; j < Length; j++)
				((BYTE *)DstBuffer)[(i*4)+j] = ((BYTE *)SrcBuffer)[(i*4)+j] ^ ((BYTE *)ivLocal)[j];

			return;
		}

		ivLocal[0] ^= SrcBuffer[i  ];
		ivLocal[1] ^= SrcBuffer[i+1];
		ivLocal[2] ^= SrcBuffer[i+2];
		ivLocal[3] ^= SrcBuffer[i+3];

		// Encrypt data with Serpent
		Encrypt_ECB( (BYTE *)ivLocal, (BYTE *)ivLocal, (BYTE *)KeyAddress );

		DstBuffer[i  ] = ivLocal[0];
		DstBuffer[i+1] = ivLocal[1];
		DstBuffer[i+2] = ivLocal[2];
		DstBuffer[i+3] = ivLocal[3];
	}
}


//************ Camellia CBC mode decryption **************


VOID
Decrypt( DWORD *IVector, 
         DWORD *KeyAddress,
         DWORD *SrcBuffer,
         DWORD *DstBuffer,
         DWORD Length )    // in bytes
{
	DWORD i,j;
	DWORD ivLocal[4], ivNew[4], buf[4];

	for(i=0; i<4; i++) 
		ivLocal[i] = IVector[i&1];

	for(i=0; Length > 0; Length -= 16, i=i+4)
	{
		// Use CFB mode to decrypt data of length less than a full block
		if (Length < 16)
		{
			Encrypt_ECB( (BYTE *)ivLocal, (BYTE *)ivLocal, (BYTE *)KeyAddress);

			for (j = 0; j < Length; j++) {
				((BYTE *)DstBuffer)[(i*4)+j] = ((BYTE *)SrcBuffer)[(i*4)+j] ^ ((BYTE *)ivLocal)[j];
			}

			return;
		}

		for(j=0; j<4; j++) 
			buf[j] = ivNew[j] = SrcBuffer[i+j];

		Decrypt_ECB( (BYTE *)buf, (BYTE *)buf, (BYTE *)KeyAddress );

		for(j=0; j<4; j++)
		{
			DstBuffer[i+j] = ivLocal[j] ^ buf[j];
			ivLocal[j] = ivNew[j];
		}
	}
}



//*********** Camellia Key Extension **************

BOOL
KeyExtend( PUCHAR KeySource,         // KEY_LENGTH bits
           PDWORD KeyDestination )   // KEY_LENGTH_EXTENDED bits
{
	static BOOL bSelfTestWasRun = FALSE;
	static BOOL bSelfTestResult = TRUE;

	if ((!KeySource) || (!KeyDestination)) 
		return FALSE;

	if (!bSelfTestWasRun)
	{
		bSelfTestResult = (Camellia_SelfTest() == 0);
		bSelfTestWasRun = TRUE;
	}

	Camellia_Ekeygen( KEY_LENGTH, (const unsigned char *)KeySource, (unsigned int *)KeyDestination);

	return bSelfTestResult;
}


//*********** Camellia_SelfTest ******************

static const struct {
	int keylen;
	unsigned char key[32], pt[16], ct[16];
} tests[] = {
	{  /* CAMELLIA-256-ECB:key=603DEB1015CA71BE2B73AEF0857D77811F352C073B6108D72D9810A30914DFF4:plaintext=F69F2445DF4F9B17AD2B417BE66C3710:ciphertext=7960109FB6DC42947FCFE59EA3C5EB6B */
		32,
		{ 0x60, 0x3D, 0xEB, 0x10, 0x15, 0xCA, 0x71, 0xBE, 0x2B, 0x73, 0xAE, 0xF0, 0x85, 0x7D, 0x77, 0x81, 0x1F, 0x35, 0x2C, 0x07, 0x3B, 0x61, 0x08, 0xD7, 0x2D, 0x98, 0x10, 0xA3, 0x09, 0x14, 0xDF, 0xF4 },
		{ 0xF6, 0x9F, 0x24, 0x45, 0xDF, 0x4F, 0x9B, 0x17, 0xAD, 0x2B, 0x41, 0x7B, 0xE6, 0x6C, 0x37, 0x10 },
		{ 0x79, 0x60, 0x10, 0x9F, 0xB6, 0xDC, 0x42, 0x94, 0x7F, 0xCF, 0xE5, 0x9E, 0xA3, 0xC5, 0xEB, 0x6B },
	},
	{
		16,
		{ 0x80, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
		{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
		{ 0x6C, 0x22, 0x7F, 0x74, 0x93, 0x19, 0xA3, 0xAA, 0x7D, 0xA2, 0x35, 0xA9, 0xBB, 0xA0, 0x5A, 0x2C },
	},
	// Camellia tests from RFC3713
	{ //CAMELLIA-128-ECB:0123456789abcdeffedcba9876543210::0123456789abcdeffedcba9876543210:67673138549669730857065648eabe43
		128 / 8,
		{ 0x01, 0x23, 0x45, 0x67, 0x89, 0xab, 0xcd, 0xef, 0xfe, 0xdc, 0xba, 0x98, 0x76, 0x54, 0x32, 0x10 },
		{ 0x01, 0x23, 0x45, 0x67, 0x89, 0xab, 0xcd, 0xef, 0xfe, 0xdc, 0xba, 0x98, 0x76, 0x54, 0x32, 0x10 },
		{ 0x67, 0x67, 0x31, 0x38, 0x54, 0x96, 0x69, 0x73, 0x08, 0x57, 0x06, 0x56, 0x48, 0xea, 0xbe, 0x43 }
	},
	{ // CAMELLIA-192-ECB:0123456789abcdeffedcba98765432100011223344556677::0123456789abcdeffedcba9876543210:b4993401b3e996f84ee5cee7d79b09b9
		192 / 8,
		{ 0x01, 0x23, 0x45, 0x67, 0x89, 0xab, 0xcd, 0xef, 0xfe, 0xdc, 0xba, 0x98, 0x76, 0x54, 0x32, 0x10, 0x00, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77 },
		{ 0x01, 0x23, 0x45, 0x67, 0x89, 0xab, 0xcd, 0xef, 0xfe, 0xdc, 0xba, 0x98, 0x76, 0x54, 0x32, 0x10 },
		{ 0xb4, 0x99, 0x34, 0x01, 0xb3, 0xe9, 0x96, 0xf8, 0x4e, 0xe5, 0xce, 0xe7, 0xd7, 0x9b, 0x09, 0xb9 },
	},
	{ // CAMELLIA-256-ECB:0123456789abcdeffedcba987654321000112233445566778899aabbccddeeff::0123456789abcdeffedcba9876543210:9acc237dff16d76c20ef7c919e3a7509
		256 / 8,
		{ 0x01, 0x23, 0x45, 0x67, 0x89, 0xab, 0xcd, 0xef, 0xfe, 0xdc, 0xba, 0x98, 0x76, 0x54, 0x32, 0x10, 0x00, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77, 0x88, 0x99, 0xaa, 0xbb, 0xcc, 0xdd, 0xee, 0xff },
		{ 0x01, 0x23, 0x45, 0x67, 0x89, 0xab, 0xcd, 0xef, 0xfe, 0xdc, 0xba, 0x98, 0x76, 0x54, 0x32, 0x10 },
		{ 0x9a, 0xcc, 0x23, 0x7d, 0xff, 0x16, 0xd7, 0x6c, 0x20, 0xef, 0x7c, 0x91, 0x9e, 0x3a, 0x75, 0x09 }
	},
};


unsigned int TestKeyTable[ CAMELLIA_TABLE_WORD_LEN ];

int Camellia_SelfTest() // returns 0 if self tests returned successfully
{
	unsigned char enc[16] = { 0 };
	unsigned char dec[16] = { 0 };

	int res = -1;

	int i;

	//DbgBreakPoint();

	for (i = sizeof(tests) / sizeof(tests[0]); i--; )
	{
		Camellia_Ekeygen(tests[i].keylen * 8, tests[i].key, TestKeyTable);

		Camellia_EncryptBlock(tests[i].keylen * 8, (const unsigned char *)tests[i].pt, TestKeyTable, enc );

		if (memcmp(enc, tests[i].ct, sizeof(enc)) != 0)
			return 1;

		//	Camellia_Decrypt_ECB(&TestKeyCtx, dec, enc);
		Camellia_DecryptBlock(tests[i].keylen * 8, enc, TestKeyTable, dec );

		if (memcmp(dec, tests[i].pt, sizeof(dec)) != 0)
			return 2;
	}

	return 0;
}

