/********************************************************************************************
 * File: bc_rc6.h
 *  
 * Purpose: Use the RC6 algorithm in CBC (or CFB) mode with 16 byte blocks and a 256-bit key.
 *
 * The RC6 algorithm's authors: Ronald L. Rivest and RSA Labs.
 * More information about RC6 at: http://www.rsa.com/rsalabs/rc6/index.html
 *
 * The code is based Dr B. R. Gladman's <brg@gladman.uk.net> algorithm implementation.
 *
 * BestCrypt module by Michael Oestergaard Pedersen <michael@daimi.au.dk>
 *
 ********************************************************************************************/

#ifndef __BC_RC6
#define __BC_RC6

// Comment this line to use CFB instead of CBC mode
#define __CBCMODE

#ifdef __CBCMODE
#define SERVICE_NAME	"BC_RC6"
#define DISPLAY_NAME    "RC6"
#define ALGORITHM_ID     0xE6
#else
#define SERVICE_NAME	"BC_RC6_CFB"
#define DISPLAY_NAME    "RC6_CFB"
#define ALGORITHM_ID     0xE7
#endif

#define DRIVER_MAJOR_VERSION	4
#define DRIVER_MINOR_VERSION	2

// Key length in bits - required for external definitions of 
// BestCrypt Algorithm modules
#define KEY_LENGTH 256

// number of bits per block
#define	BLOCK_SIZE 128

#define ALGORITHM_BLOCK_SIZE_BYTES 16

// We have to reserve space both for encryption and decryption keys,
// The following structure is the same as defined in aes.c, 
// excluding the fields that are not used in our case, when we 
// use a definite block size (16 bytes)

#define BC_BLOCK_SIZE 16

// The value is in bits (44 DWORDS = 1408 bits)
#define KEY_LENGTH_EXTENDED 1408

// Circular rotate of 32 bit values
#ifdef _MSC_VER

#include <stdlib.h>
#pragma intrinsic(_lrotr,_lrotl)
#define rotr(x,n) _lrotr(x,n)
#define rotl(x,n) _lrotl(x,n)

#else

#define rotr(x,n)   (((x) >> ((int)(n))) | ((x) << (32 - (int)(n))))
#define rotl(x,n)   (((x) << ((int)(n))) | ((x) >> (32 - (int)(n))))

#endif

/******* Exported functions *********/

#define IN

BOOL
KeyExtend( PUCHAR KeySource,
           PDWORD KeyDestination );

/* 
   Encrypts or decrypts a 128 bit block of data in ECB mode. Used only for testing
*/

VOID
RC6_ECB_Mode( DWORD *KeyAddress,
              DWORD *SrcBuffer,
              DWORD *DstBuffer,
              BOOL Operation );	// TRUE Encrypt, FALSE Decrypt

/* 
   Encryption and decryption functions
*/

VOID
Encrypt( DWORD *IVector, 
         DWORD *KeyAdress,
         DWORD *SrcBuffer,
         DWORD *DstBuffer,
         DWORD Length );    // in bytes

VOID
Decrypt( DWORD *IVector, 
         DWORD *KeyAdress,
         DWORD *SrcBuffer,
         DWORD *DstBuffer,
         DWORD Length );    // in bytes

void Encrypt_ECB( BYTE  *Input,        /* 8 bytes of input data              */
                  BYTE  *Output,       /* 8 bytes of output data             */
                  BYTE  *ExtendedKey); 


void Decrypt_ECB( BYTE  *Input,        /* 8 bytes of input data              */
                  BYTE  *Output,       /* 8 bytes of output data             */
                  BYTE  *ExtendedKey); 


#endif
