/***************************************************************************************
 * File: bc_rijn.c
 *  
 * Purpose: Using the Rijndael algorithm in CBC and CFB mode
 *          We use it with 16-bytes blocks and 256-bit key
 *
 *  Rijndael algorithm's authors:
 *		Joan Daemen
 *		Vincent Rijmen
 *
 * The code is based on Dr B. R. Gladman <brg@gladman.uk.net> algorithm implementation.
 *
 ***************************************************************************************/


#include <windows.h>   /* the string MUST be commented if */
                       /* we compile Windows 95 or NT drivers */

#ifndef __WIN95__
#include <windef.h>
#else
#include <basedef.h>
#endif

#include "bc_rijn.h"
#include "aes.h"

VOID
NTAPI
DbgBreakPoint(
    VOID
    );

/************ RIJNDAEL 32-bit CFB mode encryption *****************/

VOID
encrypt_CFB( DWORD *IVector, 
             DWORD *KeyAddress,
             DWORD *SrcBuffer,
             DWORD *DstBuffer,
             DWORD Length )    /* in bytes */
{
  DWORD i,j,dwLength;
  DWORD shReg[4], shRegEnc[4];

  for(i=0; i<(BLOCK_SIZE/32); i++) shReg[i] = IVector[i];

  dwLength = Length>>2; /* Length of buffer in 32-bit DWORDs */

  for(i=0; i<dwLength; i++)
  {
    for(j=0; j<(BLOCK_SIZE/32); j++) shRegEnc[j] = shReg[j];

	encrypt((const byte *)shRegEnc, (byte *)shRegEnc, (const aes *)KeyAddress );

	DstBuffer[i] = SrcBuffer[i] ^ shRegEnc[ (BLOCK_SIZE/32 - 1) ];

    for(j=BLOCK_SIZE/32; (j--)>0 ;)
	{  shReg[ j ] = shReg[ j - 1];
    }

	shReg[0] = DstBuffer[i];
  }
}


/************ RIJNDAEL 32-bit CFB mode decryption *****************/

VOID
decrypt_CFB( DWORD *IVector, 
             DWORD *KeyAddress,
             DWORD *SrcBuffer,
             DWORD *DstBuffer,
             DWORD Length )    /* in bytes */
{
  DWORD i,j,dwLength;
  DWORD shReg[4], shRegEnc[4], srcSave;

  for(i=0; i<(BLOCK_SIZE/32); i++) shReg[i] = IVector[i];

  dwLength = Length>>2; /* Length of buffer in 32-bit DWORDs */

  for(i=0; i<dwLength; i++)
  {
	for(j=0; j<(BLOCK_SIZE/32); j++) shRegEnc[j] = shReg[j];

	srcSave = SrcBuffer[i];
	  
	encrypt((const byte *)shRegEnc, (byte *)shRegEnc, (const aes *)KeyAddress );

	DstBuffer[i] = SrcBuffer[i] ^ shRegEnc[ (BLOCK_SIZE/32 - 1) ];

    for(j=BLOCK_SIZE/32; (j--)>0 ;)
	{  shReg[ j ] = shReg[ j - 1];
    }
  
	shReg[0] = srcSave;
  }
}




/************ RIJNDAEL CBC mode encryption *****************/
 
VOID
Encrypt( DWORD *IVector, 
         DWORD *KeyAddress,
         DWORD *SrcBuffer,
         DWORD *DstBuffer,
         DWORD Length )    /* in bytes */
{ DWORD i,l;
  DWORD ivLocal[4];

  for(i=0; i<4; i++) ivLocal[i] = IVector[i&1];

  if (Length < (BLOCK_SIZE/8))
  {
    encrypt_CFB(ivLocal,KeyAddress,SrcBuffer,DstBuffer,Length);
    return;
  }

  /* CBC encryption */
  for(i=0, l=0; l<(Length>>4); l++, i=i+4)
    { 
	  // do ECB encryption of (Initial_Vector XOR Data)
      ivLocal[0] ^= SrcBuffer[i  ];
      ivLocal[1] ^= SrcBuffer[i+1];
      ivLocal[2] ^= SrcBuffer[i+2];
      ivLocal[3] ^= SrcBuffer[i+3];

      encrypt((const byte *)ivLocal, (byte *)ivLocal, (const aes *)KeyAddress );
            
      DstBuffer[i  ] = ivLocal[0];
      DstBuffer[i+1] = ivLocal[1];
      DstBuffer[i+2] = ivLocal[2];
      DstBuffer[i+3] = ivLocal[3];
    }
}


/************** RIJNDAEL CBC mode decryption ****************/

VOID
Decrypt( DWORD *IVector, 
         DWORD *KeyAddress,
         DWORD *SrcBuffer,
         DWORD *DstBuffer,
         DWORD Length )    // in bytes
{ DWORD i,j,l;
  DWORD ivOld[8], ivNew[8], buf[8]; // we reserved by 8 elements, but needed in 4 - just for testing purposes

  // for(i=0; i<4; i++) ivOld[i] = IVector[i&1];

  ivOld[0] = IVector[0];
  ivOld[1] = IVector[1];
  ivOld[2] = IVector[0];
  ivOld[3] = IVector[1];

  if (Length < (BLOCK_SIZE/8))
  {
    decrypt_CFB(ivOld,KeyAddress,SrcBuffer,DstBuffer,Length);
    return;
  }

  /* CBC decryption */
  for(i=0, l=0; l<(Length>>4) ; l++, i = i+4)
  { 
      // Encrypted Data -> new IV, 
      // then do ECB decryption of Encrypted Data, 
      // then XOR decrypted data with old IV
/*
      for(j=0; j<4; j++) buf[j] = ivNew[j] = SrcBuffer[i+j];

      decrypt((const byte *)buf, (byte *)buf, (const aes *)KeyAddress );

      for(j=0; j<4; j++) 
      { 
         DstBuffer[i+j] = ivOld[j] ^ buf[j];
	 ivOld[j] = ivNew[j];
      }
*/
      buf[0] = ivNew[0] = SrcBuffer[i];
      buf[1] = ivNew[1] = SrcBuffer[i+1];
      buf[2] = ivNew[2] = SrcBuffer[i+2];
      buf[3] = ivNew[3] = SrcBuffer[i+3];

      decrypt((const byte *)buf, (byte *)buf, (const aes *)KeyAddress );

      DstBuffer[i]   = ivOld[0] ^ buf[0];
      DstBuffer[i+1] = ivOld[1] ^ buf[1];
      DstBuffer[i+2] = ivOld[2] ^ buf[2];
      DstBuffer[i+3] = ivOld[3] ^ buf[3];

      ivOld[0] = ivNew[0];
      ivOld[1] = ivNew[1];
      ivOld[2] = ivNew[2];
      ivOld[3] = ivNew[3];
  }
}


////////////////////////////////////////////////////////

//BOOL WasBreakpoint = FALSE;


BOOL
KeyExtend( PUCHAR KeySource,		//32 8-bits bytes - 256-bit key
           PDWORD KeyDestination )
{

	/*
	if (!WasBreakpoint)
	{
		DbgBreakPoint();
		WasBreakpoint = TRUE;
	
	}
	*/

	if (set_key((const byte*)KeySource, KEY_LENGTH / 8, both, (aes *)KeyDestination ) != aes_good)
		return FALSE;

	return TRUE;
}


////////////////////////////////////////////////////////


void Encrypt_ECB( BYTE  *Input,       /* 8 bytes of input data              */
                  BYTE  *Output,      /* 8 bytes of output data             */
                  BYTE  *ExtendedKey) /* Three [16][2] arrays of DWORDs     */ 
{
	encrypt( Input, Output, (const aes *)ExtendedKey );
}



void Decrypt_ECB( BYTE  *Input,       /* 8 bytes of input data              */
                  BYTE  *Output,      /* 8 bytes of output data             */
                  BYTE  *ExtendedKey) /* Three [16][2] arrays of DWORDs     */ 
{
	decrypt( Input, Output, (const aes *)ExtendedKey );
}


////////////////////////////////////////////////////////

